@@ -1,86 +1,14 @@
 ##############################################################################
 #
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-#
-# Copyright (c) Digital Creations.  All rights reserved.
-#
-# This license has been certified as Open Source(tm).
-#
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-#
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-#
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-#
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-#
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-#
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-#
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-#
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-#
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-#
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-#
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-#
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-#
-#
-# Disclaimer
-#
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-#
-#
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
-#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
+# 
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
+# 
 ##############################################################################
 
 """
@@ -99,22 +27,26 @@ __version__ = "1.0"
 
 #----------------------------------------------------------------------
 
-from medusa import asynchat, asyncore, logger
+import asynchat, asyncore
+from medusa import logger
 from medusa.counter import counter
 from medusa.http_server import compute_timezone_for_log
 
-from ZServer import CONNECTION_LIMIT
+from ZServer import CONNECTION_LIMIT, requestCloseOnExec
 
 from PubCore import handle
 from PubCore.ZEvent import Wakeup
 from ZPublisher.HTTPResponse import HTTPResponse
 from ZPublisher.HTTPRequest import HTTPRequest
-from Producers import ShutdownProducer, LoggingProducer
+from Producers import ShutdownProducer, LoggingProducer, file_part_producer, file_close_producer
+
+import DebugLogger
 
 from cStringIO import StringIO
 from tempfile import TemporaryFile
 import socket, string, os, sys, time
 from types import StringType
+import thread
 
 tz_for_log = compute_timezone_for_log()
 
@@ -338,11 +270,13 @@ class FCGIChannel(asynchat.async_chat):
     great loss.
     """
     closed=0
-
+    using_temp_stdin=None
+    
     def __init__(self, server, sock, addr):
         self.server = server
         self.addr = addr
         asynchat.async_chat.__init__(self, sock)
+        requestCloseOnExec(sock)
         self.setInitialState()
         self.remainingRecs = 1  # We have to read at least one
         self.env = {}
@@ -424,11 +358,23 @@ class FCGIChannel(asynchat.async_chat):
             elif rec.role == FCGI_RESPONDER:  self.remainingRecs = 2
             elif rec.role == FCGI_FILTER:     self.remainingRecs = 3
 
-
         # Read some name-value pairs (the CGI environment)
         elif rec.recType == FCGI_PARAMS:
             if rec.contentLength == 0:  # end of the stream
+            
+                if self.env.has_key('REQUEST_METHOD'):
+                    method=self.env['REQUEST_METHOD']
+                else:
+                    method='GET'
+                if self.env.has_key('PATH_INFO'):
+                    path=self.env['PATH_INFO']
+                else:
+                    path=''
+                DebugLogger.log('B', id(self), '%s %s' % (method, path))       
+     
                 self.remainingRecs = self.remainingRecs - 1
+                self.content_length=string.atoi(self.env.get(
+                    'CONTENT_LENGTH','0'))
             else:
                 self.env.update(rec.values)
 
@@ -437,7 +383,16 @@ class FCGIChannel(asynchat.async_chat):
             if rec.contentLength == 0:  # end of the stream
                 self.remainingRecs = self.remainingRecs - 1
             else:
+                # see if stdin is getting too big, and 
+                # replace it with a tempfile if necessary
+                if len(rec.content) + self.stdin.tell() > 1048576 and \
+                        not self.using_temp_stdin:
+                    t=TemporaryFile()
+                    t.write(self.stdin.getvalue())
+                    self.stdin=t
+                    self.using_temp_stdin=1
                 self.stdin.write(rec.content)
+ 
 
         # read some filter data
         elif rec.recType == FCGI_DATA:
@@ -448,13 +403,15 @@ class FCGIChannel(asynchat.async_chat):
 
 
         # We've processed the record.  Now what do we do?
-        if self.remainingRecs:
+        if self.remainingRecs > 0:
             # prepare to get the next record
             self.setInitialState()
 
         else:
             # We've got them all.  Let ZPublisher do its thang.
 
+            DebugLogger.log('I', id(self), self.stdin.tell())
+
             # But first, fixup the auth header if using newest mod_fastcgi.
             if self.env.has_key('Authorization'):
                 self.env['HTTP_AUTHORIZATION'] = self.env['Authorization']
@@ -470,18 +427,29 @@ class FCGIChannel(asynchat.async_chat):
         to ZPublisher for processing.
         """
         response = FCGIResponse(stdout = FCGIPipe(self, FCGI_STDOUT),
-                                stderr = FCGIPipe(self, FCGI_STDERR))
+                                stderr = StringIO())
         response.setChannel(self)
         request  = HTTPRequest(self.stdin, self.env, response)
         handle(self.server.module, request, response)
 
 
     def log_request(self, bytes):
-        # XXX need to add reply code logging
-        if self.env.has_key('PATH_INFO'):
-            path='%s%s' % (self.server.module, self.env['PATH_INFO'])
+        
+        DebugLogger.log('E', id(self))
+
+        if self.env.has_key('HTTP_USER_AGENT'):
+            user_agent=self.env['HTTP_USER_AGENT']
         else:
-            path='%s/' % self.server.module
+            user_agent=''
+        if self.env.has_key('HTTP_REFERER'):
+            referer=self.env['HTTP_REFERER']
+        else:
+            referer=''
+       
+        if self.env.has_key('PATH_INFO'):
+            path=self.env['PATH_INFO']
+        else:
+            path=''
         if self.env.has_key('REQUEST_METHOD'):
             method=self.env['REQUEST_METHOD']
         else:
@@ -489,24 +457,26 @@ class FCGIChannel(asynchat.async_chat):
         if self.addr:
             self.server.logger.log (
                 self.addr[0],
-                '%d - - [%s] "%s %s" %d' % (
+                '%s - - [%s] "%s %s" %d %d "%s" "%s"' % (
                     self.addr[1],
                     time.strftime (
                     '%d/%b/%Y:%H:%M:%S ',
-                    time.gmtime(time.time())
+                    time.localtime(time.time())
                     ) + tz_for_log,
-                    method, path, bytes
+                    method, path, self.reply_code, bytes,
+                    referer, user_agent
                     )
                 )
         else:
             self.server.logger.log (
-                '127.0.0.1',
-                '- - [%s] "%s %s" %d' % (
+                '127.0.0.1 ',
+                '- - [%s] "%s %s" %d %d "%s" "%s"' % (
                     time.strftime (
                     '%d/%b/%Y:%H:%M:%S ',
-                    time.gmtime(time.time())
+                    time.localtime(time.time())
                     ) + tz_for_log,
-                    method, path, bytes
+                    method, path, self.reply_code, bytes,
+                    referer, user_agent
                     )
                 )
 
@@ -530,12 +500,31 @@ class FCGIChannel(asynchat.async_chat):
         rec.recType = recType
         rec.reqId   = self.requestId
         # Can't send more than 64K minus header size.  8K seems about right.
-        while data:
-            chunk = data[:8192]
-            data = data[8192:]
-            rec.content = chunk
-            self.push(rec.getRecordAsString(), 0)
+        if type(data)==type(''):
+            # send some string data
+            while data:
+                chunk = data[:8192]
+                data = data[8192:]
+                rec.content = chunk
+                self.push(rec.getRecordAsString(), 0)
+        else:
+            # send a producer
+            p, cLen=data
+            eLen = (cLen + 7) & (0xFFFF - 7)    # align to an 8-byte boundary
+            padLen = eLen - cLen
 
+            hdr = [ rec.version,
+                    rec.recType,
+                    rec.reqId >> 8,
+                    rec.reqId & 0xFF,
+                    cLen >> 8,
+                    cLen & 0xFF,
+                    padLen,
+                    0]
+            hdr = string.join(map(chr, hdr), '')
+            self.push(hdr, 0)
+            self.push(p, 0)
+            self.push(padLen * '\000', 0)            
 
     def sendStreamTerminator(self, recType):
         rec = FCGIRecord()
@@ -544,7 +533,6 @@ class FCGIChannel(asynchat.async_chat):
         rec.content = ""
         self.push(rec.getRecordAsString(), 0)
 
-
     def sendEndRecord(self, appStatus=0):
         rec = FCGIRecord()
         rec.recType        = FCGI_END_REQUEST
@@ -601,6 +589,7 @@ class FCGIServer(asyncore.dispatcher):
                  logger_object=None):
 
         self.ip = ip
+        self.count=counter()
         asyncore.dispatcher.__init__(self)
         if not logger_object:
             logger_object = logger.file_logger(sys.stdout)
@@ -642,6 +631,7 @@ class FCGIServer(asyncore.dispatcher):
 
 
     def handle_accept(self):
+        self.count.increment()
         try:
             conn, addr = self.accept()
         except socket.error:
@@ -657,6 +647,9 @@ class FCGIServer(asyncore.dispatcher):
     def writable (self):
         return 0
 
+    def create_socket(self, family, type):
+        asyncore.dispatcher.create_socket(self, family, type)
+        requestCloseOnExec(self.socket)
 
     def listen(self, num):
         # override asyncore limits for nt's listen queue size
@@ -667,44 +660,78 @@ class FCGIServer(asyncore.dispatcher):
 
 class FCGIResponse(HTTPResponse):
   
+    _tempfile=None
+    _templock=None
+    _tempstart=0
+    
     def setChannel(self, channel):
         self.channel = channel
 
     def write(self, data):
+        stdout=self.stdout
+        
         if not self._wrote:
-            self.stdout.write(str(self))
+            l=self.headers.get('content-length', None)
+            if l is not None:
+                try:
+                    if type(l) is type(''): l=string.atoi(l)
+                    if l > 128000:
+                        self._tempfile=TemporaryFile()
+                        self._templock=thread.allocate_lock()
+                except: pass
+                    
+            stdout.write(str(self))
             self._wrote=1
-        self.stdout.write(data)
+
+        if not data: return
+
+        t=self._tempfile
+        if t is None:
+            stdout.write(data)
+        else:
+            while data:
+                # write file producers
+                # each producer holds 32K data
+                chunk=data[:32768]
+                data=data[32768:]
+                l=len(chunk)
+                b=self._tempstart
+                e=b+l
+                self._templock.acquire()
+                try:
+                    t.seek(b)
+                    t.write(chunk)
+                finally:
+                    self._templock.release()
+                self._tempstart=e
+                stdout.write((file_part_producer(t,self._templock,b,e), l))
 
     def _finish(self):
-        self.channel.sendStreamTerminator(FCGI_STDERR)
+        self.channel.reply_code=self.status
+    
+        DebugLogger.log('A', id(self.channel), '%d %d' % (
+                self.status, self.stdout.length))
+    
+        t=self._tempfile
+        if t is not None:
+            self.stdout.write((file_close_producer(t), 0))
+        self._tempfile=None
+
         self.channel.sendStreamTerminator(FCGI_STDOUT)
         self.channel.sendEndRecord()
         self.stdout.close()
         self.stderr.close()
 
-        # The following was adapted from PCGIPipe.finish and PCGIPipe.close
-        # I don't really understand it enough to know if I got it right...
-        shutdown = 0
-        if self.headers.get('bobo-exception-type','') == \
-                'exceptions.SystemExit':
-            r = self.headers.get('bobo-exception-value','0')
-            try: r=string.atoi(r)
-            except: r = r and 1 or 0
-            shutdown = r
-
         if not self.channel.closed:
             self.channel.push_with_producer(LoggingProducer(self.channel,
                                                             self.stdout.length,
-                                                            'log_request'))
-        if shutdown:
-            sys.ZServerExitCode = shutdown
-            self.channel.push_with_producer(ShutdownProducer())
+                                                            'log_request'), 0)
+        if self._shutdownRequested():
+            self.channel.push(ShutdownProducer(), 0)
             Wakeup(lambda: asyncore.close_all())
         else:
+            self.channel.push(None,0)
             Wakeup()
-
-        self.channel.close_when_done()
         self.channel=None
 
 
@@ -722,10 +749,12 @@ class FCGIPipe:
         self.recType = recType
         self.length  = 0
 
-
     def write(self, data):
-        datalen = len(data)
-        if datalen:
+        if type(data)==type(''):
+            datalen = len(data)
+        else:
+            p, datalen = data
+        if data:
             self.channel.sendDataRecord(data, self.recType)
             self.length = self.length + datalen
 
