@@ -1,85 +1,13 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 
@@ -138,14 +66,15 @@ FTP Authorization
 
 from PubCore import handle
 from medusa.ftp_server import ftp_channel, ftp_server, recv_channel
-from medusa import asyncore, asynchat, filesys
+import asyncore, asynchat
+from medusa import filesys
+
 from FTPResponse import make_response
 from FTPRequest import FTPRequest
 
-from ZServer import CONNECTION_LIMIT
+from ZServer import CONNECTION_LIMIT, requestCloseOnExec
 
 from cStringIO import StringIO
-import string
 import os
 from mimetypes import guess_type
 import marshal
@@ -161,6 +90,7 @@ class zope_ftp_channel(ftp_channel):
     
     def __init__ (self, server, conn, addr, module):
         ftp_channel.__init__(self,server,conn,addr)
+        requestCloseOnExec(conn)
         self.module=module
         self.userid=''
         self.password=''
@@ -171,7 +101,7 @@ class zope_ftp_channel(ftp_channel):
         path=apply(os.path.join,args)
         path=os.path.normpath(path)
         if os.sep != '/':
-            path=string.replace(path,os.sep,'/')
+            path=path.replace(os.sep,'/')
         return path
     
     # Overriden async_chat methods
@@ -201,28 +131,45 @@ class zope_ftp_channel(ftp_channel):
         self.get_dir_list(line,1)
     
     def get_dir_list(self, line, long=0):
+        self.globbing = None
+        self.recursive = 0
         # we need to scan the command line for arguments to '/bin/ls'...
         # XXX clean this up, maybe with getopts
+
         if len(line) > 1:
-            args = string.split(line[1])
+            args = line[1].split()
         else:
             args =[]
         path_args = []
+
+        # Extract globbing information 
+
+       
+        for i in range(len(args)):
+            x = args[i]
+            if x.find('*')!=-1 or x.find('?')!=-1:
+                self.globbing = x
+                args[i] = '.'     
+
         for arg in args:
             if arg[0] != '-':
                 path_args.append (arg)
             else:
                 if 'l' in arg:
                     long=1
+                if 'R' in arg:
+                    self.recursive = 1
+
         if len(path_args) < 1:
             dir = '.'
         else:
             dir = path_args[0]
+
         self.listdir(dir, long)
     
     def listdir (self, path, long=0):
         response=make_response(self, self.listdir_completion, long)
-        request=FTPRequest(path, 'LST', self, response)
+        request=FTPRequest(path, 'LST', self, response,globbing=self.globbing,recursive=self.recursive)
         handle(self.module, request, response)         
         
     def listdir_completion(self, long, response):
@@ -298,7 +245,7 @@ class zope_ftp_channel(ftp_channel):
     def cmd_mdtm(self, line):
         'show last modification time of file'
         if len (line) != 2:
-            self.command.not_understood (string.join (line))
+            self.command.not_understood (' '.join(line))
             return
         response=make_response(self, self.mdtm_completion)
         request=FTPRequest(line[1],'MDTM',self,response)
@@ -325,7 +272,7 @@ class zope_ftp_channel(ftp_channel):
     def cmd_size(self, line):
         'return size of file'
         if len (line) != 2:
-            self.command.not_understood (string.join (line))
+            self.command.not_understood (' '.join(line))
             return
         response=make_response(self, self.size_completion)
         request=FTPRequest(line[1],'SIZE',self,response)
@@ -344,10 +291,14 @@ class zope_ftp_channel(ftp_channel):
 
     def cmd_retr(self,line):
         if len(line) < 2:
-            self.command_not_understood (string.join (line))
+            self.command_not_understood (' '.join(line))
             return
         response=make_response(self, self.retr_completion, line[1])
+        self._response_producers = response.stdout._producers
         request=FTPRequest(line[1],'RETR',self,response)
+        # Support download restarts if possible.
+        if self.restart_position > 0:
+            request.environ['HTTP_RANGE'] = 'bytes=%d-' % self.restart_position
         handle(self.module,request,response) 
 
     def retr_completion(self, file, response):        
@@ -357,8 +308,9 @@ class zope_ftp_channel(ftp_channel):
             if not response._wrote:
                 self.client_dc.push(response.body)
             else:
-                for producer in response.stdout._producers:
+                for producer in self._response_producers:
                     self.client_dc.push_with_producer(producer)
+            self._response_producers = None
             self.client_dc.close_when_done()
             self.respond(
                     "150 Opening %s mode data connection for file '%s'" % (
@@ -373,7 +325,7 @@ class zope_ftp_channel(ftp_channel):
     def cmd_stor (self, line, mode='wb'):
         'store a file'
         if len (line) < 2:
-            self.command_not_understood (string.join (line))
+            self.command_not_understood (' '.join(line))
             return
         elif self.restart_position:
             restart_position = 0
@@ -400,18 +352,48 @@ class zope_ftp_channel(ftp_channel):
         handle(self.module,request,response)       
            
     def stor_completion(self,response):
-        status=response.getStatus()        
+        status=response.getStatus()
+        message = response.getMessage()
+
         if status in (200,201,204,302):
-            self.client_dc.channel.respond('226 Transfer complete.')
+            self.client_dc.channel.respond('226 ' + (
+                message or 'Transfer complete.'))
         elif status==401:
-            self.client_dc.channel.respond('426 Unauthorized.')
+            self.client_dc.channel.respond('426 ' + (
+                message or 'Unauthorized.'))
         else:
-            self.client_dc.channel.respond('426 Error creating file.')       
+            self.client_dc.channel.respond('426 ' + (
+                message or 'Error creating file.'))
         self.client_dc.close()
-        
+
+    def cmd_rnfr (self, line):
+        'rename from'
+        if len (line) != 2:
+            self.command_not_understood (' '.join(line))
+        else:
+            self.fromfile = line[1]
+            self.respond ('350 RNFR command successful.')
+
+    def cmd_rnto (self, line):
+        if len (line) != 2:
+            self.command_not_understood (' '.join(line))
+            return
+        pathf,idf=os.path.split(self.fromfile)
+        patht,idt=os.path.split(line[1])
+        response=make_response(self, self.rnto_completion)
+        request=FTPRequest(pathf,('RNTO',idf,idt),self,response)
+        handle(self.module,request,response)       
+
+    def rnto_completion(self,response):   
+        status=response.getStatus()
+        if status==200:
+            self.respond ('250 RNTO command successful.')
+        else:
+            self.respond ('550 error renaming file.')
+
     def cmd_dele(self, line):
         if len (line) != 2:
-            self.command.not_understood (string.join (line))
+            self.command.not_understood (' '.join (line))
             return
         path,id=os.path.split(line[1])
         response=make_response(self, self.dele_completion)
@@ -420,7 +402,7 @@ class zope_ftp_channel(ftp_channel):
         
     def dele_completion(self,response):   
         status=response.getStatus()
-        if status==200 and string.find(response.body,'Not Deletable')==-1:
+        if status==200 and response.body.find('Not Deletable')==-1:
             self.respond('250 DELE command successful.')
         elif status==401:
             self.respond('530 Unauthorized.') 
@@ -429,7 +411,7 @@ class zope_ftp_channel(ftp_channel):
 
     def cmd_mkd(self, line):
         if len (line) != 2:
-            self.command.not_understood (string.join (line))
+            self.command.not_understood (' '.join (line))
             return
         path,id=os.path.split(line[1])
         response=make_response(self, self.mkd_completion)
@@ -451,7 +433,7 @@ class zope_ftp_channel(ftp_channel):
         # XXX should object be checked to see if it's folderish
         #     before we allow it to be RMD'd?
         if len (line) != 2:
-            self.command.not_understood (string.join (line))
+            self.command.not_understood (' '.join (line))
             return
         path,id=os.path.split(line[1])
         response=make_response(self, self.rmd_completion)
@@ -462,7 +444,7 @@ class zope_ftp_channel(ftp_channel):
 
     def rmd_completion(self,response):
         status=response.getStatus()
-        if status==200 and string.find(response.body,'Not Deletable')==-1:
+        if status==200 and response.body.find('Not Deletable')==-1:
             self.respond('250 RMD command successful.')
         elif status==401:
             self.respond('530 Unauthorized.') 
@@ -475,7 +457,7 @@ class zope_ftp_channel(ftp_channel):
             self.userid = line[1]
             self.respond('331 Password required.')
         else:
-            self.command_not_understood (string.join (line))
+            self.command_not_understood (' '.join (line))
 
     def cmd_pass(self, line):
         'specify password'
@@ -484,7 +466,7 @@ class zope_ftp_channel(ftp_channel):
         else:
             pw = line[1]
         self.password=pw
-        i=string.find(self.userid,'@')
+        i=self.userid.find('@')
         if i ==-1:
             if self.server.limiter.check_limit(self):
                 self.respond ('230 Login successful.')
@@ -529,6 +511,7 @@ class zope_ftp_channel(ftp_channel):
         self.respond('502 Command not implemented.')
 
 
+
 # Override ftp server receive channel reponse mechanism 
 # XXX hack alert, this should probably be redone in a more OO way.
 
@@ -586,7 +569,7 @@ class FTPLimiter:
         total=0
         class_total=0
         if channel.anonymous:
-            for existing_channel in asyncore.socket_map.keys():
+            for existing_channel in asyncore.socket_map.values():
                 if (hasattr(existing_channel,'server') and
                         existing_channel.server is channel.server):
                     total=total+1
@@ -595,7 +578,7 @@ class FTPLimiter:
             if class_total > self.anon_limit:
                 return None
         else:                
-            for existing_channel in asyncore.socket_map.keys():
+            for existing_channel in asyncore.socket_map.values():
                 if (hasattr(existing_channel,'server') and
                         existing_channel.server is channel.server):
                     total=total+1
@@ -612,13 +595,35 @@ class FTPServer(ftp_server):
     
     ftp_channel_class = zope_ftp_channel
     limiter=FTPLimiter(10,1)
+    shutup=0
 
     def __init__(self,module,*args,**kw):
+        self.shutup=1
         apply(ftp_server.__init__, (self, None) + args, kw)
+        self.shutup=0
         self.module=module
-        
+        self.log_info('FTP server started at %s\n'
+                      '\tHostname: %s\n\tPort: %d' % (
+			time.ctime(time.time()),
+			self.hostname,
+			self.port
+			))
+
+    def log_info(self, message, type='info'):
+        if self.shutup: return
+        asyncore.dispatcher.log_info(self, message, type)
+
+    def create_socket(self, family, type):
+        asyncore.dispatcher.create_socket(self, family, type)
+        requestCloseOnExec(self.socket)
+
     def handle_accept (self):
-        conn, addr = self.accept()
+        try:
+            conn, addr = self.accept()
+        except TypeError:
+            # unpack non-sequence as result of accept
+            # returning None (in case of EWOULDBLOCK)
+            return
         self.total_sessions.increment()
         self.log_info('Incoming connection from %s:%d' % (addr[0], addr[1]))
         self.ftp_channel_class (self, conn, addr, self.module)  
