@@ -1,85 +1,13 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 '''Sequence insertion
@@ -121,12 +49,32 @@
 
     Attributes
 
-      sort -- Define the sort order for sequence items.  If an item in
-      the sequence does not define 
+      sort -- Define the sort order for sequence items.  Parameter to the
+      attribute is either a sort option, or list of sort options separated
+      by comma.  Every sort option consists of variable name, optional
+      comparison function name (default is cmp) and optional sort order
+      (default is asc).
+          Examples: sort="date" or sort="date,time" or
+      sort="title/locale,date/cmp/desc". If you want to specify sort order,
+      you cannot omit the function; use cmp for standard comparison.
+          Few predefined comparison functions available: standard cmp,
+      nocase (ignore string case), strcoll (alias "locale"),
+      strcoll_nocase (alias "locale_nocase"). Locale functions are
+      available only if module locale is already imported (you started Zope
+      with -L locale).
+
+      sort_expr -- The "sort" attribute accepts only static list of
+      sort options. This calculated parameter allows you to calculate the
+      list of sort options on the fly.
+
+      sort_expr -- This allows an expression to control sort order.
 
       reverse -- Reverse the sequence (may be combined with sort).  Note
       that this can cause a huge memory use in lazy activation instances. 
 
+      reverse_expr -- This calculated parameter allows you to calculate the
+      need of reversing on the fly.
+
       Within an 'in' block, variables are substituted from the
       elements of the iteration.  The elements may be either
       instance or mapping objects.  In addition, the variables:
@@ -186,7 +134,7 @@
     Batch sequence insertion
 
       When displaying a large number of objects, it is sometimes
-      desireable to display just a sub-sequence of the data.
+      desirable to display just a sub-sequence of the data.
       An 'in' command may have optional parameters,
       as in::
 
@@ -208,7 +156,7 @@
 
           'orphan'  -- The desired minimum number of objects to be
                        displayed.  The default value for this
-                       parameter is 3.
+                       parameter is 0.
 
           'overlap' -- The desired overlap between batches. The
                        default is no overlap.     
@@ -240,14 +188,14 @@
                  <!--#/in-->
 
              If the original URL is: 'foo/bar?x=1&y=2', then the
-             rendered text (after row data are displated) will be::
+             rendered text (after row data are displayed) will be::
 
                       <a href="foo/bar?x=1&y=2&batch_start=20">
                       (Next 20 results)
                       </a>
 
              If the original URL is: 'foo/bar?batch_start=10&x=1&y=2',
-             then the rendered text (after row data are displated)
+             then the rendered text (after row data are displayed)
              will be::
 
                       <a href="foo/bar?x=1&y=2&batch_start=30">
@@ -353,16 +301,16 @@
         mean -- The mean of numeric values values.
 
         variance -- The variance of numeric values computed with a
-          degrees of freedom qeual to the count - 1.
+          degrees of freedom equal to the count - 1.
 
         variance-n -- The variance of numeric values computed with a
-          degrees of freedom qeual to the count.
+          degrees of freedom equal to the count.
 
         standard-deviation -- The standard deviation of numeric values
-          computed with a degrees of freedom qeual to the count - 1.
+          computed with a degrees of freedom equal to the count - 1.
 
         standard-deviation-n -- The standard deviation of numeric
-          values computed with a degrees of freedom qeual to the count.
+          values computed with a degrees of freedom equal to the count.
 
       Missing values are either 'None' or the attribute 'Value'
       of the module 'Missing', if present.
@@ -382,15 +330,17 @@
 
 ''' #'
 
-__rcs_id__='$Id: DT_In.py,v 1.38 1999/08/27 14:56:27 petrilli Exp $'
-__version__='$Revision: 1.38 $'[11:-2]
+__rcs_id__='$Id: DT_In.py,v 1.57 2001/11/28 15:50:54 matt Exp $'
+__version__='$Revision: 1.57 $'[11:-2]
 
+import sys
 from DT_Util import ParseError, parse_params, name_param, str
-from DT_Util import render_blocks, InstanceDict, ValidationError
-from string import find, atoi, join
-import ts_regex
+from DT_Util import render_blocks, InstanceDict, ValidationError, Eval
+from DT_Util import simple_name, add_with_prefix
+import re
 from DT_InSV import sequence_variables, opt
 TupleType=type(())
+StringTypes = (type(''), type(u''))
 
 class InFactory:
     blockContinuations=('else',)
@@ -408,14 +358,16 @@ class InClass:
     expr=sort=batch=mapping=None
     start_name_re=None
     reverse=None
+    sort_expr=reverse_expr=None
     
     def __init__(self, blocks):
         tname, args, section = blocks[0]
         args=parse_params(args, name='', start='1',end='-1',size='10',
-                          orphan='3',overlap='1',mapping=1,
+                          orphan='0',overlap='1',mapping=1,
                           skip_unauthorized=1,
                           previous=1, next=1, expr='', sort='',
-                          reverse=1)
+                          reverse=1, sort_expr='', reverse_expr='',
+                          prefix='')
         self.args=args
         has_key=args.has_key
 
@@ -423,6 +375,12 @@ class InClass:
             self.sort=sort=args['sort']
             if sort=='sequence-item': self.sort=''
 
+        if has_key('sort_expr'):
+            self.sort_expr=Eval(args['sort_expr'])
+
+        if has_key('reverse_expr'):
+            self.reverse_expr=Eval(args['reverse_expr'])
+
         if has_key('reverse'):
             self.reverse=args['reverse']
             
@@ -430,6 +388,11 @@ class InClass:
         for n in 'start', 'size', 'end':
             if has_key(n): self.batch=1
 
+        prefix = args.get('prefix')
+        if prefix and not simple_name(prefix):
+            raise ParseError, _tm(
+                'prefix is not a simple name', 'in')
+        
         for n in 'orphan','overlap','previous','next':
             if has_key(n) and not self.batch:
                 raise ParseError, (
@@ -442,13 +405,14 @@ class InClass:
         if has_key('start'):
             v=args['start']
             if type(v)==type(''):
-                try: atoi(v)
+                try: int(v)
                 except:
-                    self.start_name_re=ts_regex.compile(
+
+                    self.start_name_re=re.compile(
                         '&+'+
-                        join(map(lambda c: "[%s]" % c, v),'')+
+                        ''.join(["[%s]" % c for c in v])+
                         '=[0-9]+&+')
-                    
+                   
         name,expr=name_param(args,'in',1)
         if expr is not None: expr=expr.eval
         self.__name__, self.expr = name, expr
@@ -484,15 +448,21 @@ class InClass:
             raise 'InError', (
                 'Strings are not allowed as input to the in tag.')
 
+
         section=self.section
         params=self.args
         
         mapping=self.mapping
 
-        if self.sort is not None:
-            sequence=self.sort_sequence(sequence)
+        if self.sort_expr is not None:
+            self.sort=self.sort_expr.eval(md)
+            sequence=self.sort_sequence(sequence, md)
+        elif self.sort is not None:
+            sequence=self.sort_sequence(sequence, md)
 
-        if self.reverse is not None:
+        if self.reverse_expr is not None and self.reverse_expr.eval(md):
+            sequence=self.reverse_sequence(sequence)
+        elif self.reverse is not None:
             sequence=self.reverse_sequence(sequence)
             
         next=previous=0
@@ -501,7 +471,7 @@ class InClass:
         end=int_param(params,md,'end',0)
         size=int_param(params,md,'size',0)
         overlap=int_param(params,md,'overlap',0)
-        orphan=int_param(params,md,'orphan','3')
+        orphan=int_param(params,md,'orphan','0')
         start,end,sz=opt(start,end,size,orphan,sequence)
         if params.has_key('next'): next=1
         if params.has_key('previous'): previous=1
@@ -511,17 +481,22 @@ class InClass:
 
         try: query_string=md['QUERY_STRING']
         except: query_string=''
-
-        vars=sequence_variables(sequence,'?'+query_string,self.start_name_re)
+        prefix = params.get('prefix')
+        vars = sequence_variables(sequence, '?'+query_string,
+                                  self.start_name_re, prefix)
         kw=vars.data
+        pkw = add_with_prefix(kw, 'sequence', prefix)
+        for k, v in kw.items():
+            pkw[k] = v
+        pkw['sequence-step-size']=sz
+        pkw['sequence-step-overlap']=overlap
+        pkw['sequence-step-start']=start
+        pkw['sequence-step-end']=end
+        pkw['sequence-step-start-index']=start-1
+        pkw['sequence-step-end-index']=end-1
+        pkw['sequence-step-orphan']=orphan
+
         kw['mapping']=mapping
-        kw['sequence-step-size']=sz
-        kw['sequence-step-overlap']=overlap
-        kw['sequence-step-start']=start
-        kw['sequence-step-end']=end
-        kw['sequence-step-start-index']=start-1
-        kw['sequence-step-end-index']=end-1
-        kw['sequence-step-orphan']=orphan
 
         push=md._push
         pop=md._pop
@@ -534,10 +509,10 @@ class InClass:
                 if first > 0:
                     pstart,pend,psize=opt(0,first+overlap,
                                           sz,orphan,sequence)
-                    kw['previous-sequence']=1
-                    kw['previous-sequence-start-index']=pstart-1
-                    kw['previous-sequence-end-index']=pend-1
-                    kw['previous-sequence-size']=pend+1-pstart
+                    pkw['previous-sequence']=1
+                    pkw['previous-sequence-start-index']=pstart-1
+                    pkw['previous-sequence-end-index']=pend-1
+                    pkw['previous-sequence-size']=pend+1-pstart
                     result=render(section,md)
 
                 elif self.elses: result=render(self.elses, md)
@@ -548,73 +523,85 @@ class InClass:
                     # there are more items, without actually
                     # computing a length:
                     sequence[end]
-                    pstart,pend,psize=opt(end+1-overlap,0,
-                                          sz,orphan,sequence)
-                    kw['next-sequence']=1
-                    kw['next-sequence-start-index']=pstart-1
-                    kw['next-sequence-end-index']=pend-1
-                    kw['next-sequence-size']=pend+1-pstart
-                    result=render(section,md)
-                except:
+                except IndexError:
                     if self.elses: result=render(self.elses, md)
                     else: result=''
+                else:
+                    pstart,pend,psize=opt(end+1-overlap,0,
+                                          sz,orphan,sequence)
+                    pkw['next-sequence']=1
+                    pkw['next-sequence-start-index']=pstart-1
+                    pkw['next-sequence-end-index']=pend-1
+                    pkw['next-sequence-size']=pend+1-pstart
+                    result=render(section,md)
             else:
                 result = []
                 append=result.append
-                validate=md.validate
+                guarded_getitem = getattr(md, 'guarded_getitem', None)
                 for index in range(first,end):
-                    if index==first and index > 0:
-                        pstart,pend,psize=opt(0,index+overlap,
-                                              sz,orphan,sequence)
-                        kw['previous-sequence']=1
-                        kw['previous-sequence-start-index']=pstart-1
-                        kw['previous-sequence-end-index']=pend-1
-                        kw['previous-sequence-size']=pend+1-pstart
-                    else:
-                        kw['previous-sequence']=0
-                        if index==last:
-                            try:
-                                # The following line is a sneaky way to
-                                # test whether there are more items,
-                                # without actually computing a length:
-                                sequence[end]
-                                pstart,pend,psize=opt(end+1-overlap,0,
-                                                      sz,orphan,sequence)
-                                kw['previous-sequence']=0
-                                kw['next-sequence']=1
-                                kw['next-sequence-start-index']=pstart-1
-                                kw['next-sequence-end-index']=pend-1
-                                kw['next-sequence-size']=pend+1-pstart
-                            except: pass
+                    # preset
+                    pkw['previous-sequence']= 0
+                    pkw['next-sequence']= 0 # now more often defined then previously
+                    #
+                    if index==first or index==last:
+                        # provide batching information
+                        if first > 0:
+                            pstart,pend,psize=opt(0,first+overlap,
+                                                  sz,orphan,sequence)
+                            if index==first: pkw['previous-sequence']=1
+                            pkw['previous-sequence-start-index']=pstart-1
+                            pkw['previous-sequence-end-index']=pend-1
+                            pkw['previous-sequence-size']=pend+1-pstart
+                        try:
+                            # The following line is a sneaky way to
+                            # test whether there are more items,
+                            # without actually computing a length:
+                            sequence[end]
+                            pstart,pend,psize=opt(end+1-overlap,0,
+                                                  sz,orphan,sequence)
+                            if index==last: pkw['next-sequence']=1
+                            pkw['next-sequence-start-index']=pstart-1
+                            pkw['next-sequence-end-index']=pend-1
+                            pkw['next-sequence-size']=pend+1-pstart
+                        except: pass
         
-                    if index==last: kw['sequence-end']=1
+                    if index==last: pkw['sequence-end']=1
 
-                    client=sequence[index]
-
-                    if validate is not None:
-                        try: vv=validate(sequence,sequence,index,client,md)
-                        except: vv=0
-                        if not vv:
+                    if guarded_getitem is not None:
+                        try: client = guarded_getitem(sequence, index)
+                        except ValidationError, vv:
                             if (params.has_key('skip_unauthorized') and
                                 params['skip_unauthorized']):
-                                if index==first: kw['sequence-start']=0
+                                if index==first: pkw['sequence-start']=0
                                 continue
-                            raise ValidationError, index
+                            raise ValidationError, '(item %s): %s' % (
+                                index, vv), sys.exc_info()[2]
+                    else:
+                        client = sequence[index]
 
-                    kw['sequence-index']=index
-                    if type(client)==TupleType and len(client)==2:
+                    pkw['sequence-index']=index
+                    t = type(client)
+                    if t is TupleType and len(client)==2:
                         client=client[1]
 
-                    if mapping: push(client)
-                    else: push(InstanceDict(client, md))
+                    if mapping:
+                        pushed = 1
+                        push(client)
+                    elif t in StringTypes:
+                        pushed = 0
+                    else:
+                        pushed = 1
+                        push(InstanceDict(client, md))
 
                     try: append(render(section, md))
-                    finally: pop(1)
+                    finally:
+                        if pushed:
+                            pop()
 
-                    if index==first: kw['sequence-start']=0
+                    if index==first: pkw['sequence-start']=0
 
 
-                result=join(result, '')
+                result = ''.join(result)
 
         finally:
             if cache: pop()
@@ -644,14 +631,24 @@ class InClass:
         section=self.section        
         mapping=self.mapping
 
-        if self.sort is not None:
-            sequence=self.sort_sequence(sequence)
 
-        if self.reverse is not None:
+        if self.sort_expr is not None:
+            self.sort=self.sort_expr.eval(md)
+            sequence=self.sort_sequence(sequence, md)
+        elif self.sort is not None:
+            sequence=self.sort_sequence(sequence, md)
+
+        if self.reverse_expr is not None and self.reverse_expr.eval(md):
             sequence=self.reverse_sequence(sequence)
-            
-        vars=sequence_variables(sequence)
+        elif self.reverse is not None:
+            sequence=self.reverse_sequence(sequence)
+
+        prefix = self.args.get('prefix')
+        vars=sequence_variables(sequence, alt_prefix=prefix)
         kw=vars.data
+        pkw = add_with_prefix(kw, 'sequence', prefix)
+        for k, v in kw.items():
+            pkw[k] = v
         kw['mapping']=mapping
 
         l=len(sequence)
@@ -666,33 +663,42 @@ class InClass:
         try:
                 result = []
                 append=result.append
-                validate=md.validate
+                guarded_getitem = getattr(md, 'guarded_getitem', None)
                 for index in range(l):
-                    if index==last: kw['sequence-end']=1
-                    client=sequence[index]
-
-                    if validate is not None:
-                        try: vv=validate(sequence,sequence,index,client,md)
-                        except: vv=0
-                        if not vv:
+                    if index==last: pkw['sequence-end']=1
+                    if guarded_getitem is not None:
+                        try: client = guarded_getitem(sequence, index)
+                        except ValidationError, vv:
                             if (self.args.has_key('skip_unauthorized') and
                                 self.args['skip_unauthorized']):
-                                if index==1: kw['sequence-start']=0
+                                if index==1: pkw['sequence-start']=0
                                 continue
-                            raise ValidationError, index
+                            raise ValidationError, '(item %s): %s' % (
+                                index, vv), sys.exc_info()[2]
+                    else:
+                        client = sequence[index]
 
-                    kw['sequence-index']=index
-                    if type(client)==TupleType and len(client)==2:
+                    pkw['sequence-index']=index
+                    t = type(client)
+                    if t is TupleType and len(client)==2:
                         client=client[1]
 
-                    if mapping: push(client)
-                    else: push(InstanceDict(client, md))
+                    if mapping:
+                        pushed = 1
+                        push(client)
+                    elif t in StringTypes:
+                        pushed = 0
+                    else:
+                        pushed = 1
+                        push(InstanceDict(client, md))
 
                     try: append(render(section, md))
-                    finally: pop()
-                    if index==0: kw['sequence-start']=0
+                    finally:
+                        if pushed:
+                            pop()
+                    if index==0: pkw['sequence-start']=0
 
-                result=join(result, '')
+                result = ''.join(result)
 
         finally:
             if cache: pop()
@@ -700,14 +706,38 @@ class InClass:
 
         return result
 
-    def sort_sequence(self, sequence):
+    def sort_sequence(self, sequence, md):
+
+        # Modified with multiple sort fields by Ross Lazarus
+        # April 7 2000 rossl@med.usyd.edu.au
+        # eg <dtml-in "foo" sort="akey,anotherkey">
+        
+        # Modified with advanced sort functions by
+        # Oleg Broytmann <phd@phd.pp.ru> 30 Mar 2001
+        # eg <dtml-in "foo" sort="akey/nocase,anotherkey/cmp/desc">
 
         sort=self.sort
+        need_sortfunc = sort.find('/') >= 0
+
+        sortfields = sort.split(',')   # multi sort = key1,key2 
+        multsort = len(sortfields) > 1 # flag: is multiple sort
+
+        if need_sortfunc:
+            # prepare the list of functions and sort order multipliers
+            sf_list = make_sortfunctions(sortfields, md)
+
+            # clean the mess a bit
+            if multsort: # More than one sort key.
+                sortfields = map(lambda x: x[0], sf_list)
+            else:
+                sort = sf_list[0][0]
+
         mapping=self.mapping
         isort=not sort
-        k=None
+
         s=[]
         for client in sequence:
+            k = None
             if type(client)==TupleType and len(client)==2:
                 if isort: k=client[0]
                 v=client[1]
@@ -716,37 +746,145 @@ class InClass:
                 v=client
 
             if sort:
-                if mapping: k=v[sort]
-                else: k=getattr(v, sort)
-                if not basic_type(k):           
-                    try: k=k()
-                    except: pass
+                 if multsort: # More than one sort key.
+                     k = []
+                     for sk in sortfields:
+                         try:
+                             if mapping: akey = v[sk]
+                             else: akey = getattr(v, sk)
+                         except AttributeError, KeyError: akey = None
+                         if not basic_type(akey):
+                             try: akey = akey()
+                             except: pass
+                         k.append(akey)
+                 else: # One sort key.
+                     try:
+                         if mapping: k = v[sort]
+                         else: k = getattr(v, sort)
+                     except AttributeError, KeyError: k = None
+                     if not basic_type(type(k)):           
+                         try: k = k()
+                         except: pass
 
             s.append((k,client))
 
-        s.sort()
+        if need_sortfunc:
+            by = SortBy(multsort, sf_list)
+            s.sort(by)
+        else:
+            s.sort()
 
         sequence=[]
-        for k, client in s: sequence.append(client)
-
+        for k, client in s:
+             sequence.append(client)
         return sequence
 
-
     def reverse_sequence(self, sequence):
         s=list(sequence)
         s.reverse()
         return s
 
 
-basic_type={type(''): 1, type(0): 1, type(0.0): 1, type(()): 1, type([]): 1
-            }.has_key
+basic_type={type(''): 1, type(0): 1, type(0.0): 1, type(()): 1, type([]): 1,
+            type(None) : 1 }.has_key
 
 def int_param(params,md,name,default=0, st=type('')):
     try: v=params[name]
     except: v=default
     if v:
-        try: v=atoi(v)
+        try: v=int(v)
         except:
             v=md[v]
-            if type(v) is st: v=atoi(v)
+            if type(v) is st: v=int(v)
     return v
+
+
+# phd: Advanced sort support
+
+def nocase(str1, str2):
+    return cmp(str1.lower(), str2.lower())
+
+if sys.modules.has_key("locale"): # only if locale is already imported
+    from locale import strcoll
+
+    def strcoll_nocase(str1, str2):
+        return strcoll(str1.lower(), str2.lower())
+
+
+def make_sortfunctions(sortfields, md):
+    """Accepts a list of sort fields; splits every field, finds comparison
+    function. Returns a list of 3-tuples (field, cmp_function, asc_multplier)"""
+
+    sf_list = []
+    for field in sortfields:
+        f = field.split('/')
+        l = len(f)
+
+        if l == 1:
+            f.append("cmp")
+            f.append("asc")
+        elif l == 2:
+            f.append("asc")
+        elif l == 3:
+            pass
+        else:
+            raise SyntaxError, "sort option must contain no more than 2 slashes"
+
+        f_name = f[1]
+
+        # predefined function?
+        if f_name == "cmp":
+            func = cmp # builtin
+        elif f_name == "nocase":
+            func = nocase
+        elif f_name in ("locale", "strcoll"):
+            func = strcoll
+        elif f_name in ("locale_nocase", "strcoll_nocase"):
+            func = strcoll_nocase
+        else: # no - look it up in the namespace
+            func = md.getitem(f_name, 0)
+
+        sort_order = f[2].lower()
+
+        if sort_order == "asc":
+            multiplier = +1
+        elif sort_order == "desc":
+            multiplier = -1
+        else:
+            raise SyntaxError, "sort oder must be either ASC or DESC"
+
+        sf_list.append((f[0], func, multiplier))
+
+    return sf_list
+
+
+class SortBy:
+    def __init__(self, multsort, sf_list):
+        self.multsort = multsort
+        self.sf_list = sf_list
+
+    def __call__(self, o1, o2):
+        multsort = self.multsort
+        if multsort:
+            o1 = o1[0] # if multsort - take the first element (key list)
+            o2 = o2[0]
+
+        sf_list = self.sf_list
+        l = len(sf_list)
+
+        # assert that o1 and o2 are tuples of apropriate length
+        assert len(o1) == l + 1 - multsort, "%s, %d" % (o1, l + multsort)
+        assert len(o2) == l + 1 - multsort, "%s, %d" % (o2, l + multsort)
+
+        # now run through the list of functions in sf_list and
+        # compare every object in o1 and o2
+        for i in range(l):
+            # if multsort - we already extracted the key list
+            # if not multsort - i is 0, and the 0th element is the key
+            c1, c2 = o1[i], o2[i]
+            func, multiplier = sf_list[i][1:3]
+            n = func(c1, c2)
+            if n: return n*multiplier
+
+        # all functions returned 0 - identical sequences
+        return 0
