@@ -1,90 +1,18 @@
 /*****************************************************************************
+
+  Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
   
-  Zope Public License (ZPL) Version 1.0
-  -------------------------------------
-  
-  Copyright (c) Digital Creations.  All rights reserved.
-  
-  This license has been certified as Open Source(tm).
-  
-  Redistribution and use in source and binary forms, with or without
-  modification, are permitted provided that the following conditions are
-  met:
-  
-  1. Redistributions in source code must retain the above copyright
-     notice, this list of conditions, and the following disclaimer.
-  
-  2. Redistributions in binary form must reproduce the above copyright
-     notice, this list of conditions, and the following disclaimer in
-     the documentation and/or other materials provided with the
-     distribution.
-  
-  3. Digital Creations requests that attribution be given to Zope
-     in any manner possible. Zope includes a "Powered by Zope"
-     button that is installed by default. While it is not a license
-     violation to remove this button, it is requested that the
-     attribution remain. A significant investment has been put
-     into Zope, and this effort will continue if the Zope community
-     continues to grow. This is one way to assure that growth.
-  
-  4. All advertising materials and documentation mentioning
-     features derived from or use of this software must display
-     the following acknowledgement:
-  
-       "This product includes software developed by Digital Creations
-       for use in the Z Object Publishing Environment
-       (http://www.zope.org/)."
-  
-     In the event that the product being advertised includes an
-     intact Zope distribution (with copyright and license included)
-     then this clause is waived.
-  
-  5. Names associated with Zope or Digital Creations must not be used to
-     endorse or promote products derived from this software without
-     prior written permission from Digital Creations.
-  
-  6. Modified redistributions of any form whatsoever must retain
-     the following acknowledgment:
-  
-       "This product includes software developed by Digital Creations
-       for use in the Z Object Publishing Environment
-       (http://www.zope.org/)."
-  
-     Intact (re-)distributions of any official Zope release do not
-     require an external acknowledgement.
-  
-  7. Modifications are encouraged but must be packaged separately as
-     patches to official Zope releases.  Distributions that do not
-     clearly separate the patches from the original work must be clearly
-     labeled as unofficial distributions.  Modifications which do not
-     carry the name Zope may be packaged in any form, as long as they
-     conform to all of the clauses above.
-  
-  
-  Disclaimer
-  
-    THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-    EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-    PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-    USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-    SUCH DAMAGE.
-  
-  
-  This software consists of contributions made by Digital Creations and
-  many individuals on behalf of Digital Creations.  Specific
-  attributions are listed in the accompanying credits file.
+  This software is subject to the provisions of the Zope Public License,
+  Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+  WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+  WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+  FOR A PARTICULAR PURPOSE
   
  ****************************************************************************/
 static char cDocumentTemplate_module_documentation[] = 
 ""
-"\n$Id: cDocumentTemplate.c,v 1.29.4.1 2000/01/04 17:12:15 jim Exp $"
+"\n$Id: cDocumentTemplate.c,v 1.42 2001/11/28 15:50:55 matt Exp $"
 ;
 
 #include "ExtensionClass.h"
@@ -92,8 +20,9 @@ static char cDocumentTemplate_module_documentation[] =
 static PyObject *py_isDocTemp=0, *py_blocks=0, *py_=0, *join=0, *py_acquire;
 static PyObject *py___call__, *py___roles__, *py_AUTHENTICATED_USER;
 static PyObject *py_hasRole, *py__proxy_roles, *py_Unauthorized;
-static PyObject *py_Unauthorized_fmt, *py_validate;
-static PyObject *py__push, *py__pop, *py_aq_base;
+static PyObject *py_Unauthorized_fmt, *py_guarded_getattr;
+static PyObject *py__push, *py__pop, *py_aq_base, *py_renderNS;
+static PyObject *py___class__, *html_quote;
 
 /* ----------------------------------------------------- */
 
@@ -108,7 +37,7 @@ typedef struct {
   PyObject *inst;
   PyObject *cache;
   PyObject *namespace;
-  PyObject *validate;
+  PyObject *guarded_getattr;
 } InstanceDictobject;
 
 staticforward PyExtensionClass InstanceDictType;
@@ -116,18 +45,19 @@ staticforward PyExtensionClass InstanceDictType;
 static PyObject *
 InstanceDict___init__(InstanceDictobject *self, PyObject *args)
 {
-  self->validate=NULL;
+  self->guarded_getattr=NULL;
   UNLESS(PyArg_ParseTuple(args, "OO|O",
 			  &(self->inst),
 			  &(self->namespace),
-			  &(self->validate)))
+			  &(self->guarded_getattr)))
     return NULL;
   Py_INCREF(self->inst);
   Py_INCREF(self->namespace);
-  if (self->validate)
-    Py_INCREF(self->validate);
+  if (self->guarded_getattr)
+    Py_INCREF(self->guarded_getattr);
   else
-    UNLESS(self->validate=PyObject_GetAttr(self->namespace, py_validate))
+    UNLESS(self->guarded_getattr=PyObject_GetAttr(self->namespace,
+                                                  py_guarded_getattr))
        return NULL;
     
   UNLESS(self->cache=PyDict_New()) return NULL;
@@ -150,7 +80,7 @@ InstanceDict_dealloc(InstanceDictobject *self)
   Py_XDECREF(self->inst);
   Py_XDECREF(self->cache);
   Py_XDECREF(self->namespace);
-  Py_XDECREF(self->validate);
+  Py_XDECREF(self->guarded_getattr);
   Py_DECREF(self->ob_type);
   PyMem_DEL(self);
 }
@@ -182,7 +112,7 @@ InstanceDict_subscript( InstanceDictobject *self, PyObject *key)
   char *name;
   
   /* Try to get value from the cache */
-  if (r=PyObject_GetItem(self->cache, key)) return r;
+  if ((r=PyObject_GetItem(self->cache, key))) return r;
   PyErr_Clear();
   
   /* Check for __str__ */
@@ -193,56 +123,28 @@ InstanceDict_subscript( InstanceDictobject *self, PyObject *key)
       return PyObject_Str(self->inst);
     }
   
-  /* Do explicit acquisition with "roles" rule */
-  if (r=PyObject_GetAttr(self->inst, py_acquire))
-    {
-      /* Sanity check in case of explicit Aq */
-      if (v=PyObject_GetAttr(self->inst, key)) Py_DECREF(v);  
-      else 
-	{
-	  Py_DECREF(r);
-	  goto KeyError;
-	}
+  if (self->guarded_getattr != Py_None) {
+    r = PyObject_CallFunction(self->guarded_getattr, "OO", self->inst, key);
+  }
+  else {
+    r = PyObject_GetAttr(self->inst, key);
+  }
 
-      if (self->validate != Py_None)
-	{
-	  UNLESS_ASSIGN(r,PyObject_CallFunction(
-		 r, "OOO", key, self->validate, self->namespace))
-	    {
-	      PyObject *tb;
+  if (!r) {
+    PyObject *tb;
 
-	      PyErr_Fetch(&r, &v, &tb);
-	      if (r != PyExc_AttributeError || PyObject_Compare(v,key))
-		{
-		  PyErr_Restore(r,v,tb);
-		  return NULL;
-		}
-	      Py_XDECREF(r);
-	      Py_XDECREF(v);
-	      Py_XDECREF(tb);
-	      
-	      goto KeyError;
-	    }
-	}
-      else
-	UNLESS_ASSIGN(r, PyObject_GetAttr(self->inst, key)) goto KeyError;
-    }  
-  else
-    {
-      PyErr_Clear();
+    PyErr_Fetch(&r, &v, &tb);
+    if (r != PyExc_AttributeError) /* || PyObject_Compare(v,key)) */
+      {
+	PyErr_Restore(r,v,tb);
+	return NULL;
+      }
+    Py_XDECREF(r);
+    Py_XDECREF(v);
+    Py_XDECREF(tb);
 
-      /* OK, use getattr */
-      UNLESS(r=PyObject_GetAttr(self->inst, key)) goto KeyError;
-
-      if (self->validate != Py_None)
-	{
-	  UNLESS(v=PyObject_CallFunction(
-	    self->validate,"OOOOO",
-	    self->inst, self->inst, key, r, self->namespace))
-	    return NULL;
-	  Py_DECREF(v);
-	}
-    }
+    goto KeyError;
+  }
   
   if (r && PyObject_SetItem(self->cache, key, r) < 0) PyErr_Clear();
   
@@ -312,9 +214,7 @@ typedef struct {
 staticforward PyExtensionClass MMtype;
 
 static PyObject *
-MM_push(self, args)
-	MM *self;
-	PyObject *args;
+MM_push(MM *self, PyObject *args)
 {
   PyObject *src;
   UNLESS(PyArg_Parse(args, "O", &src)) return NULL;
@@ -324,9 +224,7 @@ MM_push(self, args)
 }
 
 static PyObject *
-MM_pop(self, args)
-	MM *self;
-	PyObject *args;
+MM_pop(MM *self, PyObject *args)
 {
   int i=1, l;
   PyObject *r;
@@ -343,9 +241,7 @@ err:
 }
 
 static PyObject *
-MM__init__(self, args)
-     MM *self;
-     PyObject *args;
+MM__init__(MM *self, PyObject *args)
 {
   UNLESS(PyArg_Parse(args, "")) return NULL;
   UNLESS(self->data=PyList_New(0)) return NULL;
@@ -355,83 +251,138 @@ MM__init__(self, args)
   return Py_None;
 }
 
+static int
+safe_PyCallable_Check(PyObject *x)
+{
+  PyObject *klass;
+
+  if (x == NULL)
+    return 0;
+  klass = PyObject_GetAttr(x, py___class__);
+  if (klass) {
+    PyObject *call = PyObject_GetAttr(x, py___call__);
+    if (call) {
+      Py_DECREF(klass);
+      Py_DECREF(call);
+      return 1;
+    }
+    else {
+      PyErr_Clear();
+      Py_DECREF(klass);
+      if (PyClass_Check(x) || PyExtensionClass_Check(x))
+        return 1;
+      else
+        return 0;
+    }
+  }
+  else {
+    PyErr_Clear();
+    return PyCallable_Check(x);
+  }
+}
+
+static int 
+dtObjectIsCallable(PyObject *ob) {
+  PyObject *base=0;
+  int result=0;
+
+  /* Ensure that an object is really callable by unwrapping it */
+  UNLESS(base=PyObject_GetAttr(ob, py_aq_base)) {
+    PyErr_Clear();
+    return safe_PyCallable_Check(ob);
+  }
+  result=safe_PyCallable_Check(base);
+  Py_DECREF(base);
+  return result;
+}
+
+static int
+dtObjectIsDocTemp(PyObject *ob) {
+  PyObject *base=0;
+  PyObject *value=0;
+  int result=0;
+
+  /* Ensure that 'isDocTemp' is not acquired */
+  UNLESS(base=PyObject_GetAttr(ob, py_aq_base)) {
+    PyErr_Clear();
+    base = ob;
+    Py_INCREF(base);
+  }
+
+  if ( (value = PyObject_GetAttr(base, py_isDocTemp)) ) {
+    if (PyObject_IsTrue(value)) {
+      result = 1;
+    }
+    Py_DECREF(value);
+  }
+  else PyErr_Clear();
+
+  Py_DECREF(base);
+  return result;
+}
+
+
 static PyObject *
 MM_cget(MM *self, PyObject *key, int call)
 {
   long i;
-  int dt=0;
-  PyObject *e, *t, *rr, *tb;
+  PyObject *e, *rr;
 
   UNLESS(-1 != (i=PyList_Size(self->data))) return NULL;
   while (--i >= 0)
     {
-      e=PyList_GetItem(self->data,i);
-      if (e=PyObject_GetItem(e,key))
+      e=PyList_GET_ITEM(self->data,i);
+      if (PyDict_Check(e))
+        {
+          e=PyDict_GetItem(e, key);
+          Py_XINCREF(e);
+        }
+      else
+        {
+          UNLESS (e=PyObject_GetItem(e,key)) 
+            {
+              if (PyErr_Occurred() == PyExc_KeyError)
+                PyErr_Clear();
+              else
+                return NULL;
+            }
+        }
+
+      if (e)
 	{
-	  dt=0;
+          if (!call) return e;
 
-	  if (PyCallable_Check(e))
+          /* Try calling __render_with_namespace__ */
+          if ((rr = PyObject_GetAttr(e, py_renderNS))) 
+            {
+              Py_DECREF(e);
+              UNLESS_ASSIGN(rr, PyObject_CallFunction(rr, "O", self))
+                return NULL;
+              return rr;
+            }
+          else PyErr_Clear();
+
+	  if (dtObjectIsCallable(e))
 	    {
-	      /* Decide whether we have a document template */
-	      if (rr=PyObject_GetAttr(e,py_isDocTemp))
-		{
-		  if (PyObject_IsTrue(rr)) dt=1;
-		  Py_DECREF(rr);
-		}
-	      else PyErr_Clear();
-
 	      /* Try calling the object */
-	      if (call)
-		{
-		  if (dt)
-		    {
-		      ASSIGN(e,PyObject_CallFunction(e,"OO", Py_None, self));
-		      UNLESS(e)
-			return NULL;
-		    }
-		  else
-		    {
-		      rr=PyObject_CallObject(e,NULL);
-		      if (rr) ASSIGN(e,rr);
-		      else
-			{
-			  PyErr_Fetch(&t, &rr, &tb);
-			  if (t!=PyExc_AttributeError ||
-			     PyObject_Compare(rr,py___call__) != 0)
-			    {
-			      PyErr_Restore(t,rr,tb);
-			      Py_DECREF(e);
-			      return NULL;
-			    }
-			  /* 
-                            Added by Brian on 08/30/99. We need to be sure
-                            to DECREF the exception in the event of an 
-                            AttributeError to avoid leaking.
-                          */
-			  else {
-			    Py_XDECREF(t);
-			    Py_XDECREF(rr);
-			    Py_XDECREF(tb);
-			  }
-
-			}
-		    }
-		}
-	    }
+              if (dtObjectIsDocTemp(e))
+                {
+                  ASSIGN(e,PyObject_CallFunction(e,"OO", Py_None, self));
+                  UNLESS(e) return NULL;
+                  return e;
+                }
 
+              rr=PyObject_CallObject(e,NULL);
+              if (rr) ASSIGN(e,rr);
+              else {
+                Py_DECREF(e);
+                return NULL;
+              }
+            }
 	  return e;
 	}
-      PyErr_Fetch(&e, &rr, &tb);
-      if (dt || e != PyExc_KeyError)
-	{
-	  PyErr_Restore(e,rr,tb);
-	  return NULL;
-	}
-      Py_XDECREF(e);
-      Py_XDECREF(rr);
-      Py_XDECREF(tb);
     }
-  PyErr_SetObject(PyExc_KeyError,key);
+  PyErr_SetObject(PyExc_KeyError, key);
   return NULL;
 }
 
@@ -479,8 +430,7 @@ static struct PyMethodDef MM_methods[] = {
 };
 
 static void
-MM_dealloc(self)
-     MM *self;
+MM_dealloc(MM *self)
 {
   Py_XDECREF(self->data);
   Py_XDECREF(self->dict);
@@ -501,7 +451,7 @@ MM_getattro(MM *self, PyObject *name)
     {
       PyObject *v;
 
-      if (v=PyDict_GetItem(self->dict, name))
+      if ((v=PyDict_GetItem(self->dict, name)))
 	{
 	  Py_INCREF(v);
 	  return v;
@@ -531,8 +481,7 @@ MM_setattro(MM *self, PyObject *name, PyObject *v)
 }
 
 static int
-MM_length(self)
-	MM *self;
+MM_length(MM *self)
 {
   long l=0, el, i;
   PyObject *e=0;
@@ -611,7 +560,7 @@ newDictInstance(PyObject *data)
 static PyObject *
 MM_call(MM *self, PyObject *args, PyObject *kw)
 {
-  PyObject *r;
+  PyObject *r, *t;
   int i, l=0;
 
   if (args && (l=PyTuple_Size(args)) < 0) return NULL;
@@ -635,7 +584,9 @@ MM_call(MM *self, PyObject *args, PyObject *kw)
     }
 
   ASSIGN(r, OBJECT(newDictInstance(r)));
-  return r;
+  UNLESS(t=PyTuple_New(1)) goto err;
+  PyTuple_SET_ITEM(t, 0, r);
+  return t;
 
 err:
   Py_XDECREF(r);
@@ -682,11 +633,6 @@ static PyExtensionClass MMtype = {
 	METHOD_CHAIN(MM_methods)
 };
 
-static struct PyMethodDef TemplateDict_methods[] = {
-  {NULL,		NULL}		/* sentinel */
-};
-
-
 
 /* List of methods defined in the module */
 
@@ -715,7 +661,7 @@ static int
 render_blocks_(PyObject *blocks, PyObject *rendered,
 	       PyObject *md, PyObject *mda)
 {
-  PyObject *block;
+  PyObject *block, *t;
   int l, i, k=0, append;
 
   if ((l=PyList_Size(blocks)) < 0) return -1;
@@ -724,95 +670,123 @@ render_blocks_(PyObject *blocks, PyObject *rendered,
       block=PyList_GET_ITEM(((PyListObject*)blocks), i);
       append=1;
 
-      if (PyTuple_Check(block))
+      if (PyTuple_Check(block) 
+          && PyTuple_GET_SIZE(block) > 1 
+          && PyTuple_GET_ITEM(block, 0)
+          && PyString_Check(PyTuple_GET_ITEM(block, 0)))
 	{
-	  int bs;
+          switch (PyString_AS_STRING(PyTuple_GET_ITEM(block, 0))[0])
+            {
+            case 'v': /* var  */
+	      t=PyTuple_GET_ITEM(block,1);
 
-	  bs=((PyTupleObject*)block)->ob_size;
-	  
-	  if (bs==1)
-	    {
-	      /* Simple var */
-	      block=PyTuple_GET_ITEM(block,0);
-	      if (PyString_Check(block)) block=PyObject_GetItem(md,block);
-	      else block=PyObject_CallObject(block,mda);
-	      if (block) ASSIGN(block, PyObject_Str(block));
-	      UNLESS(block) return -1;
-	    }
-	  else
-	    {
-	      /* if */
-	      int icond, m;
-	      PyObject *cond, *n, *cache;
+	      if (t == NULL) return -1;
 
-	      UNLESS(cache=PyDict_New()) return -1;
-	      cond=PyObject_GetAttr(md,py__push);
-	      if (cond) ASSIGN(cond, PyObject_CallFunction(cond,"O",cache));
-	      Py_DECREF(cache);
-	      if (cond) Py_DECREF(cond);
-	      else return -1;
+	      if (PyString_Check(t)) t=PyObject_GetItem(md, t);
+	      else t=PyObject_CallObject(t, mda);
+
+              if (t == NULL || (! PyString_Check(t)))
+                {
+                  if (t) ASSIGN(t, PyObject_Str(t));
+                  UNLESS(t) return -1;
+                }
+
+              if (PyString_Check(t) 
+                  && PyTuple_GET_SIZE(block) == 3) /* html_quote */
+                {
+                  if (strchr(PyString_AS_STRING(t), '&')
+                      || strchr(PyString_AS_STRING(t), '<')
+                      || strchr(PyString_AS_STRING(t), '>')
+                      || strchr(PyString_AS_STRING(t), '"')
+                      )
+                    ASSIGN(t, PyObject_CallFunction(html_quote, "O", t));
+		    if (t == NULL) return -1;
+                }
+                  
+              block = t;
+              break;
+            case 'i': /* if */
+              {
+                int icond, m, bs;
+                PyObject *cond, *n, *cache;
+
+                bs = PyTuple_GET_SIZE(block) - 1; /* subtract code */
+
+                UNLESS(cache=PyDict_New()) return -1;
+                cond=PyObject_GetAttr(md,py__push);
+                if (cond) ASSIGN(cond, PyObject_CallFunction(cond,"O",cache));
+                Py_DECREF(cache);
+                if (cond) Py_DECREF(cond);
+                else return -1;
 	      
-	      append=0;
-	      m=bs-1;
-	      for (icond=0; icond < m; icond += 2)
-		{
-		  cond=PyTuple_GET_ITEM(block,icond);
-		  if (PyString_Check(cond))
-		    {
-		      /* We have to be careful to handle key errors here */
-		      n=cond;
-		      if (cond=PyObject_GetItem(md,cond))
-			{
-			  if (PyDict_SetItem(cache, n, cond) < 0)
-			    {
-			      Py_DECREF(cond);
-			      return if_finally(md,1);
-			    }
-			}
-		      else
-			{
-			  PyObject *t, *v, *tb;
-
-			  PyErr_Fetch(&t, &v, &tb);
-			  if (t != PyExc_KeyError || PyObject_Compare(v,n))
-			    {
-			      PyErr_Restore(t,v,tb);
-			      return if_finally(md,1);
-			    }
-			  Py_XDECREF(t);
-			  Py_XDECREF(v);
-			  Py_XDECREF(tb);
-			  cond=Py_None;
-			  Py_INCREF(cond);
-			}
-		    }
-		  else
-		    UNLESS(cond=PyObject_CallObject(cond,mda))
-		       return if_finally(md,1);
-
-		  if (PyObject_IsTrue(cond))
-		    {
-		      Py_DECREF(cond);
-		      block=PyTuple_GET_ITEM(block,icond+1);
-		      if (block!=Py_None &&
-			 render_blocks_(block, rendered, md, mda) < 0)
-			return if_finally(md,1);
-		      m=-1;
-		      break;
-		    }
-		  else Py_DECREF(cond);
-		}
+                append=0;
+                m=bs-1;
+                for (icond=0; icond < m; icond += 2)
+                  {
+                    cond=PyTuple_GET_ITEM(block,icond+1);
+                    if (PyString_Check(cond))
+                      {
+                        /* We have to be careful to handle key errors here */
+                        n=cond;
+                        if ((cond=PyObject_GetItem(md,cond)))
+                          {
+                            if (PyDict_SetItem(cache, n, cond) < 0)
+                              {
+                                Py_DECREF(cond);
+                                return if_finally(md,1);
+                              }
+                          }
+                        else
+                          {
+                            PyObject *t, *v, *tb;
+                            
+                            PyErr_Fetch(&t, &v, &tb);
+                            if (t != PyExc_KeyError || PyObject_Compare(v,n))
+                              {
+                                PyErr_Restore(t,v,tb);
+                                return if_finally(md,1);
+                              }
+                            Py_XDECREF(t);
+                            Py_XDECREF(v);
+                            Py_XDECREF(tb);
+                            cond=Py_None;
+                            Py_INCREF(cond);
+                          }
+                      }
+                    else
+                      UNLESS(cond=PyObject_CallObject(cond,mda))
+                        return if_finally(md,1);
+                    
+                    if (PyObject_IsTrue(cond))
+                      {
+                        Py_DECREF(cond);
+                        block=PyTuple_GET_ITEM(block,icond+1+1);
+                        if (block!=Py_None &&
+                            render_blocks_(block, rendered, md, mda) < 0)
+                          return if_finally(md,1);
+                        m=-1;
+                        break;
+                      }
+                    else Py_DECREF(cond);
+                  }
 		if (icond==m)
 		  {
-		    block=PyTuple_GET_ITEM(block,icond);
+		    block=PyTuple_GET_ITEM(block,icond+1);
 		    if (block!=Py_None &&
-		       render_blocks_(block, rendered, md, mda) < 0)
+                        render_blocks_(block, rendered, md, mda) < 0)
 		      return if_finally(md,1);
 		  }
-
+                
 		if (if_finally(md,0) == -2) return -1;
-	    }
-	}
+              }
+              break;
+            default:
+              PyErr_Format(PyExc_ValueError,
+                           "Invalid DTML command code, %s",
+                           PyString_AS_STRING(PyTuple_GET_ITEM(block, 0)));
+              return -1;
+            }
+        }
       else if (PyString_Check(block))
 	{
 	  Py_INCREF(block);
@@ -866,222 +840,57 @@ err:
   return NULL;
 }  
   
+static PyObject *
+safe_callable(PyObject *self, PyObject *args)
+{
+  PyObject *ob;
+  int res;
+
+  UNLESS(PyArg_ParseTuple(args,"O", &ob)) return NULL;
+  res = safe_PyCallable_Check(ob);
+  if (res)
+    return PyInt_FromLong(1);
+  else
+    return PyInt_FromLong(0);
+}  
+
 static struct PyMethodDef Module_Level__methods[] = {
   {"render_blocks", (PyCFunction)render_blocks,	METH_VARARGS,
    ""},
+  {"safe_callable", (PyCFunction)safe_callable,	METH_VARARGS,
+   "callable() with a workaround for a problem with ExtensionClasses\n"
+   "and __call__()."},
   {NULL, (PyCFunction)NULL, 0, NULL}		/* sentinel */
 };
 
-static PyObject *
-validate(PyObject *self, PyObject *args)
-{
-  PyObject *inst, *parent, *name, *value, *md, *__roles__, *i, *p;
-  char *cname="";
-  long ir;
-
-  /* def validate(self, inst, parent, name, value, md): */
-  UNLESS(PyArg_ParseTuple(args,"OOOOO",&inst,&parent,&name,&value,&md))
-    return NULL;
-
-  if (PyString_Check(name))
-    {
-      UNLESS(cname=PyString_AsString(name)) return NULL;
-      if (*cname=='a' && cname[1]=='q' && cname[2]=='_'
-	  && strcmp(cname, "aq_explicit") && strcmp(cname, "aq_parent"))
-	/* We disallow names beginning with "aq_" unless they are
-	   aq_parent or aq_explicit */
-	return PyInt_FromLong(0);
-    }
-
-  /*
-        if hasattr(value, '__roles__'): roles=value.__roles__
-	else:
-            
-            if (name[:5]=='func_' and
-	        type(parent) in (
-                types.MethodType, types.FunctionType,
-                type(DTMLMethod.validate) )
-                ):
-                # we don't want any function attributes
-                return 0
-
-	    if hasattr(parent,'__roles__'): roles=parent.__roles__
-            elif hasattr(parent, 'aq_acquire'):
-                try: roles=parent.aq_acquire('__roles__')
-                except AttributeError:
-                    if hasattr(inst, 'aq_base'): inst=inst.aq_base
-                    if hasattr(parent, 'aq_base'): parent=parent.aq_base
-                    return inst is parent
-            else:
-                if hasattr(inst, 'aq_base'): inst=inst.aq_base
-                if hasattr(parent, 'aq_base'): parent=parent.aq_base
-                return inst is parent
-            value=parent
-   */
-  UNLESS(__roles__=PyObject_GetAttr(value,py___roles__))
-    {
-      PyErr_Clear();
-
-      if (
-	  (*cname=='f' && 
-	   cname[1]=='u' && cname[2]=='n' && cname[3]=='c' && cname[4]=='_')
-	  && 
-	  (PyECMethod_Check(parent) 
-	   || PyFunction_Check(parent)
-	   || PyMethod_Check(parent)
-	   )
-	  )
-	return PyInt_FromLong(0);
-
-      UNLESS(__roles__=PyObject_GetAttr(parent,py___roles__))
-	{
-	  PyErr_Clear();
-	  if (__roles__=PyObject_GetAttr(parent,py_acquire))
-	    {
-	      ASSIGN(__roles__, 
-		     PyObject_CallFunction(__roles__, "O", py___roles__));
-	    }
-	}
-      value=parent;
-    }
-  if (! __roles__) 
-    {
-      PyErr_Clear();
-
-      /* Waaaa, check for wrapped objects, waaaaa! */
-      UNLESS(i=PyObject_GetAttr(inst, py_aq_base))
-	{
-	  PyErr_Clear();
-	  Py_INCREF(inst);
-	  i=inst;
-	}
-      UNLESS(p=PyObject_GetAttr(parent, py_aq_base))
-	{
-	  PyErr_Clear();
-	  Py_INCREF(parent);
-	  p=parent;
-	}
-
-      ir = i==p;
-      Py_DECREF(i);
-      Py_DECREF(p);
-
-      return PyInt_FromLong(ir);
-    }
-
-  /* if roles is None: return 1 */
-  if (__roles__==Py_None)
-    {
-      Py_DECREF(__roles__);
-      return PyInt_FromLong(1);
-    }
-
-  /*    try: 
-	    if md.AUTHENTICATED_USER.hasRole(value, roles):
-		return 1
-	    
-	except AttributeError: pass
-   */
-  if (md=PyObject_GetAttr(md,py_AUTHENTICATED_USER))
-    {
-      ASSIGN(md,PyObject_GetAttr(md,py_hasRole));
-      if (md) ASSIGN(md,PyObject_CallFunction(md,"OO",value,__roles__));
-      if (md)
-	{
-	  if (PyObject_IsTrue(md))
-	    {
-	      Py_DECREF(__roles__);
-	      return md;
-	    }
-	  Py_DECREF(md);
-	}
-      else PyErr_Clear();
-    }
-  else PyErr_Clear();
-
-  /*    for r in self._proxy_roles:
-	    if r in roles: return 1
-   */
-  if (PyObject_IsTrue(__roles__))
-     if ((md=PyObject_GetAttr(self, py__proxy_roles)))
-     {
-       int i,l, isIn;
-       PyObject *role;
-       
-       if ((l=PyObject_Length(md)) < 0) PyErr_Clear();
-       else
-	 {
-	   for (i=0; i < l; i++)
-	     {
-	       UNLESS(role=PySequence_GetItem(md,i))
-		 {
-		   PyErr_Clear();
-		   break;
-		 }
-	       isIn=PySequence_In(__roles__,role);
-	       Py_DECREF(role);
-	       if (isIn < 0)
-		 {
-		   PyErr_Clear();
-		   break;
-		 }
-	       if (isIn)
-		 {
-		   Py_DECREF(md);
-		   return __roles__; /* Any true object would do. */
-		 }
-	     }
-	 }
-       Py_DECREF(md);
-     }
-     else PyErr_Clear();
-
-  Py_DECREF(__roles__);
-
-  /*     raise 'Unauthorized', (
-		'You are not authorized to access <em>%s</em>.' % name)
-   */
-  if (name=PyString_Format(py_Unauthorized_fmt, name))
-    {
-      PyErr_SetObject(py_Unauthorized, name);
-      Py_DECREF(name);
-    }
-  return NULL;
-}
-
-  
-static struct PyMethodDef Document_methods[] = {
-  {"validate", (PyCFunction)validate,	METH_VARARGS,
-   ""},
-  {NULL, (PyCFunction)NULL, 0, NULL}		/* sentinel */
-};
-
-
-/* Initialization function for the module (*must* be called initcDocumentTemplate) */
-
 void
-initcDocumentTemplate()
+initcDocumentTemplate(void)
 {
   PyObject *m, *d;
-  char *rev="$Revision: 1.29.4.1 $";
-  PURE_MIXIN_CLASS(cDocument,
-	"Base class for documents that adds fast validation method",
-	Document_methods);
+  char *rev="$Revision: 1.42 $";
+
+  DictInstanceType.ob_type=&PyType_Type;
+
+  UNLESS (html_quote = PyImport_ImportModule("html_quote")) return;
+  ASSIGN(html_quote, PyObject_GetAttrString(html_quote, "html_quote"));
+  UNLESS (html_quote) return;
 
   UNLESS(py_isDocTemp=PyString_FromString("isDocTemp")) return;
+  UNLESS(py_renderNS=PyString_FromString("__render_with_namespace__")) return;
   UNLESS(py_blocks=PyString_FromString("blocks")) return;
   UNLESS(py_acquire=PyString_FromString("aq_acquire")) return;
   UNLESS(py___call__=PyString_FromString("__call__")) return;
   UNLESS(py___roles__=PyString_FromString("__roles__")) return;
   UNLESS(py__proxy_roles=PyString_FromString("_proxy_roles")) return;
   UNLESS(py_hasRole=PyString_FromString("hasRole")) return;
-  UNLESS(py_validate=PyString_FromString("validate")) return;
+  UNLESS(py_guarded_getattr=PyString_FromString("guarded_getattr")) return;
   UNLESS(py__push=PyString_FromString("_push")) return;
   UNLESS(py__pop=PyString_FromString("_pop")) return;
   UNLESS(py_aq_base=PyString_FromString("aq_base")) return;
   UNLESS(py_Unauthorized=PyString_FromString("Unauthorized")) return;
   UNLESS(py_Unauthorized_fmt=PyString_FromString(
 	 "You are not authorized to access <em>%s</em>.")) return;
+  UNLESS(py___class__=PyString_FromString("__class__")) return;
 
   UNLESS(py_AUTHENTICATED_USER=PyString_FromString("AUTHENTICATED_USER"))
     return;
@@ -1098,13 +907,10 @@ initcDocumentTemplate()
 
   d = PyModule_GetDict(m);
 
-  PyExtensionClass_Export(d,"cDocument",cDocumentType);
   PyExtensionClass_Export(d,"InstanceDict",InstanceDictType);
   PyExtensionClass_Export(d,"TemplateDict",MMtype);
 
   PyDict_SetItemString(d, "__version__",
 		       PyString_FromStringAndSize(rev+11,strlen(rev+11)-2));
 
-  if (PyErr_Occurred())
-    Py_FatalError("can't initialize module cDocumentTemplate");
 }
