@@ -1,96 +1,26 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 __doc__="""Copy interface"""
-__version__='$Revision: 1.45.4.1 $'[11:-2]
+__version__='$Revision: 1.77 $'[11:-2]
 
 import sys, string, Globals, Moniker, tempfile, ExtensionClass
 from marshal import loads, dumps
 from urllib import quote, unquote
 from zlib import compress, decompress
 from App.Dialogs import MessageDialog
-
+from AccessControl import getSecurityManager
+from Acquisition import aq_base, aq_inner, aq_parent
+from zExceptions import Unauthorized
 
 CopyError='Copy Error'
 
@@ -101,7 +31,7 @@ class CopyContainer(ExtensionClass.Base):
     __ac_permissions__=(
         ('View management screens',
          ('manage_cutObjects', 'manage_copyObjects', 'manage_pasteObjects',
-          'manage_renameForm', 'manage_renameObject',)),
+          'manage_renameForm', 'manage_renameObject', 'manage_renameObjects',)),
         )
 
 
@@ -110,17 +40,25 @@ class CopyContainer(ExtensionClass.Base):
     def _setOb(self, id, object): setattr(self, id, object)
     def _delOb(self, id): delattr(self, id)
     def _getOb(self, id, default=_marker):
-        if hasattr(self, 'aq_base'): self=self.aq_base
+        self = aq_base(self)
         if default is _marker: return getattr(self, id)
         try: return getattr(self, id)
         except: return default
 
 
     def manage_CopyContainerFirstItem(self, REQUEST):
-        return self._getOb(REQUEST['ids'][0])        
+        return self._getOb(REQUEST['ids'][0])
 
-    def manage_cutObjects(self, ids, REQUEST=None):
+    def manage_CopyContainerAllItems(self, REQUEST):
+        return map(lambda i, s=self: s._getOb(i), tuple(REQUEST['ids']))
+
+    def manage_cutObjects(self, ids=None, REQUEST=None):
         """Put a reference to the objects named in ids in the clip board"""
+        if ids is None and REQUEST is not None:
+            return eNoItemsSpecified
+        elif ids is None:
+            raise ValueError, 'ids must be specified'
+
         if type(ids) is type(''):
             ids=[ids]
         oblist=[]
@@ -129,17 +67,23 @@ class CopyContainer(ExtensionClass.Base):
             if not ob.cb_isMoveable():
                 raise CopyError, eNotSupported % id
             m=Moniker.Moniker(ob)
-            oblist.append((m.jar, m.ids))
+            oblist.append(m.dump())
         cp=(1, oblist)
         cp=_cb_encode(cp)
         if REQUEST is not None:
             resp=REQUEST['RESPONSE']
-            resp.setCookie('__cp', cp, path='%s' % REQUEST['SCRIPT_NAME'])
-            return self.manage_main(self, REQUEST, cb_dataValid=1)
+            resp.setCookie('__cp', cp, path='%s' % cookie_path(REQUEST))
+            REQUEST['__cp'] = cp
+            return self.manage_main(self, REQUEST)
         return cp
     
-    def manage_copyObjects(self, ids, REQUEST=None, RESPONSE=None):
+    def manage_copyObjects(self, ids=None, REQUEST=None, RESPONSE=None):
         """Put a reference to the objects named in ids in the clip board"""
+        if ids is None and REQUEST is not None:
+            return eNoItemsSpecified
+        elif ids is None:
+            raise ValueError, 'ids must be specified'
+
         if type(ids) is type(''):
             ids=[ids]
         oblist=[]
@@ -148,13 +92,14 @@ class CopyContainer(ExtensionClass.Base):
             if not ob.cb_isCopyable():
                 raise CopyError, eNotSupported % id
             m=Moniker.Moniker(ob)
-            oblist.append((m.jar, m.ids))
+            oblist.append(m.dump())
         cp=(0, oblist)
         cp=_cb_encode(cp)
         if REQUEST is not None:
             resp=REQUEST['RESPONSE']
-            resp.setCookie('__cp', cp, path='%s' % REQUEST['SCRIPT_NAME'])
-            return self.manage_main(self, REQUEST, cb_dataValid=1)
+            resp.setCookie('__cp', cp, path='%s' % cookie_path(REQUEST))
+            REQUEST['__cp'] = cp
+            return self.manage_main(self, REQUEST)
         return cp
 
     def _get_id(self, id):
@@ -189,26 +134,32 @@ class CopyContainer(ExtensionClass.Base):
         except: raise CopyError, eInvalid
 
         oblist=[]
-        m=Moniker.Moniker()
         op=cp[0]
-        for j, d in cp[1]:
-            m.jar=j
-            m.ids=d
-            try: ob=m.bind(self._p_jar)
+        app = self.getPhysicalRoot()
+
+        for mdata in cp[1]:
+            m = Moniker.loadMoniker(mdata)
+            try: ob = m.bind(app)
             except: raise CopyError, eNotFound
-            self._verifyObjectPaste(ob, REQUEST)
+            self._verifyObjectPaste(ob)
             oblist.append(ob)
 
         if op==0:
             # Copy operation
             for ob in oblist:
                 if not ob.cb_isCopyable():
-                     raise CopyError, eNotSupported % absattr(ob.id)
+                    raise CopyError, eNotSupported % ob.getId()
+                try:    ob._notifyOfCopyTo(self, op=0)
+                except: raise CopyError, MessageDialog(
+                    title='Copy Error',
+                    message=sys.exc_info()[1],
+                    action ='manage_main')
                 ob=ob._getCopy(self)
-                ob.manage_afterClone(ob)
-                id=self._get_id(absattr(ob.id))
+                id=self._get_id(ob.getId())
                 ob._setId(id)
                 self._setObject(id, ob)
+                ob = self._getOb(id)
+                ob.manage_afterClone(ob)
 
             if REQUEST is not None:
                 return self.manage_main(self, REQUEST, update_menu=1,
@@ -217,52 +168,79 @@ class CopyContainer(ExtensionClass.Base):
         if op==1:
             # Move operation
             for ob in oblist:
-                id=absattr(ob.id)
+                id=ob.getId()
                 if not ob.cb_isMoveable():
                     raise CopyError, eNotSupported % id
+                try:    ob._notifyOfCopyTo(self, op=1)
+                except: raise CopyError, MessageDialog(
+                    title='Move Error',
+                    message=sys.exc_info()[1],
+                    action ='manage_main')
                 if not sanity_check(self, ob):
                     raise CopyError, 'This object cannot be pasted into itself'
-                ob.aq_parent._delObject(id)
-                if hasattr(ob, 'aq_base'):
-                    ob=ob.aq_base
+
+                # try to make ownership explicit so that it gets carried
+                # along to the new location if needed.
+                ob.manage_changeOwnershipType(explicit=1)
+                
+                aq_parent(aq_inner(ob))._delObject(id)
+                ob = aq_base(ob)
                 id=self._get_id(id)
                 ob._setId(id)
-                self._setObject(id, ob)
+
+                self._setObject(id, ob, set_owner=0)
+
+                # try to make ownership implicit if possible
+                ob=self._getOb(id)
+                ob.manage_changeOwnershipType(explicit=0)
 
             if REQUEST is not None:
-                REQUEST['RESPONSE'].setCookie('cp_', 'deleted',
-                                    path='%s' % REQUEST['SCRIPT_NAME'],
+                REQUEST['RESPONSE'].setCookie('__cp', 'deleted',
+                                    path='%s' % cookie_path(REQUEST),
                                     expires='Wed, 31-Dec-97 23:59:59 GMT')
+                REQUEST['__cp'] = None
                 return self.manage_main(self, REQUEST, update_menu=1,
                                         cb_dataValid=0)
         return ''
 
 
-    manage_renameForm=Globals.HTMLFile('renameForm', globals())
+    manage_renameForm=Globals.DTMLFile('dtml/renameForm', globals())
+
+    def manage_renameObjects(self, ids=[], new_ids=[], REQUEST=None):
+        """Rename several sub-objects"""
+        if len(ids) != len(new_ids):
+            raise 'Bad Request','Please rename each listed object.'
+        for i in range(len(ids)):
+            if ids[i] != new_ids[i]:
+                self.manage_renameObject(ids[i], new_ids[i], REQUEST)
+        if REQUEST is not None:
+            return self.manage_main(self, REQUEST, update_menu=1)
+        return None
 
     def manage_renameObject(self, id, new_id, REQUEST=None):
         """Rename a particular sub-object"""
         try: self._checkId(new_id)
         except: raise CopyError, MessageDialog(
                       title='Invalid Id',
-                      message=sys.exc_value,
+                      message=sys.exc_info()[1],
                       action ='manage_main')
         ob=self._getOb(id)
         if not ob.cb_isMoveable():
             raise CopyError, eNotSupported % id            
-        self._verifyObjectPaste(ob, REQUEST)
+        self._verifyObjectPaste(ob)
         try:    ob._notifyOfCopyTo(self, op=1)
         except: raise CopyError, MessageDialog(
                       title='Rename Error',
-                      message=sys.exc_value,
+                      message=sys.exc_info()[1],
                       action ='manage_main')
         self._delObject(id)
-        if hasattr(ob, 'aq_base'):
-            ob=ob.aq_base
+        ob = aq_base(ob)
         ob._setId(new_id)
-        self._setObject(new_id, ob)
-        #ob=ob.__of__(self)            
-        #ob._postCopy(self, op=1)
+        
+        # Note - because a rename always keeps the same context, we
+        # can just leave the ownership info unchanged.
+        self._setObject(new_id, ob, set_owner=0)
+
         if REQUEST is not None:
             return self.manage_main(self, REQUEST, update_menu=1)
         return None
@@ -273,56 +251,68 @@ class CopyContainer(ExtensionClass.Base):
     # Because it's still a "management" function.
     manage_clone__roles__=None
     def manage_clone(self, ob, id, REQUEST=None):
-        """Clone an object, creating a new object with the given id."""
+        # Clone an object, creating a new object with the given id.
         if not ob.cb_isCopyable():
-            raise CopyError, eNotSupported % absattr(ob.id)            
+            raise CopyError, eNotSupported % ob.getId()
         try: self._checkId(id)
         except: raise CopyError, MessageDialog(
                       title='Invalid Id',
-                      message=sys.exc_value,
+                      message=sys.exc_info()[1],
                       action ='manage_main')
-        self._verifyObjectPaste(ob, REQUEST)
+        self._verifyObjectPaste(ob)
         try:    ob._notifyOfCopyTo(self, op=0)
         except: raise CopyError, MessageDialog(
-                      title='Rename Error',
-                      message=sys.exc_value,
+                      title='Clone Error',
+                      message=sys.exc_info()[1],
                       action ='manage_main')
         ob=ob._getCopy(self)
         ob._setId(id)
         self._setObject(id, ob)
-        #ob=ob.__of__(self)
+        ob=ob.__of__(self)
         #ob._postCopy(self, op=0)
         return ob
 
     def cb_dataValid(self):
-        "Return true if clipboard data seems valid."
+        # Return true if clipboard data seems valid.
         try:    cp=_cb_decode(self.REQUEST['__cp'])
         except: return 0
         return 1
 
     def cb_dataItems(self):
-        "List of objects in the clip board"
+        # List of objects in the clip board
         try:    cp=_cb_decode(self.REQUEST['__cp'])
         except: return []
         oblist=[]
-        m=Moniker.Moniker()
-        op=cp[0]
-        for j, d in cp[1]:
-            m.jar=j
-            m.ids=d
-            oblist.append(m.bind(self._p_jar))
+
+        app = self.getPhysicalRoot()
+        for mdata in cp[1]:
+            m = Moniker.loadMoniker(mdata)
+            oblist.append(m.bind(app))
         return oblist
 
     validClipData=cb_dataValid
 
-    def _verifyObjectPaste(self, ob, REQUEST=None):
-        if not hasattr(ob, 'meta_type'):
+    def _verifyObjectPaste(self, object, validate_src=1):
+        # Verify whether the current user is allowed to paste the
+        # passed object into self. This is determined by checking
+        # to see if the user could create a new object of the same
+        # meta_type of the object passed in and checking that the
+        # user actually is allowed to access the passed in object
+        # in its existing context.
+        #
+        # Passing a false value for the validate_src argument will skip
+        # checking the passed in object in its existing context. This is
+        # mainly useful for situations where the passed in object has no 
+        # existing context, such as checking an object during an import
+        # (the object will not yet have been connected to the acquisition
+        # heirarchy).
+        if not hasattr(object, 'meta_type'):
             raise CopyError, MessageDialog(
                   title='Not Supported',
                   message='The object <EM>%s</EM> does not support this ' \
-                          'operation' % absattr(ob.id),
+                          'operation' % absattr(object.id),
                   action='manage_main')
-        mt=ob.meta_type
+        mt=object.meta_type
         if not hasattr(self, 'all_meta_types'):
             raise CopyError, MessageDialog(
                   title='Not Supported',
@@ -330,50 +320,56 @@ class CopyContainer(ExtensionClass.Base):
                   action='manage_main')
 
         method_name=None
+        mt_permission=None
         meta_types=absattr(self.all_meta_types)
         for d in meta_types:
             if d['name']==mt:
                 method_name=d['action']
+                mt_permission=d.get( 'permission', None )
                 break
 
-        if REQUEST is None:
-            REQUEST=getattr(self, 'REQUEST', None)
-        
-        if method_name is not None:
-
-            meth=None
-            if hasattr(self, method_name):
-                meth=getattr(self, method_name)
+        if mt_permission is not None:
+            if getSecurityManager().checkPermission( mt_permission, self ):
+                if not validate_src:
+                    return
+                # Ensure the user is allowed to access the object on the
+                # clipboard.
+                try:    parent=aq_parent(aq_inner(object))
+                except: parent=None
+                if getSecurityManager().validate(None, parent, None, object):
+                    return
+                raise Unauthorized, absattr(object.id)
             else:
-                # Handle strange names that come from the Product
-                # machinery ;(
-                mn=string.split(method_name, '/')
-                if len(mn) > 1:
-                    pname= mn[1]
-                    product=self.manage_addProduct[pname]
-                    fname=mn[2]
-                    factory=getattr(product, fname)
-                    try: meth=getattr(factory, factory.initial)
-                    except: meth=factory
+                raise Unauthorized(permission=mt_permission)
+        #
+        #   XXX:    Ancient cruft, left here in true co-dependent fashion
+        #           to keep from breaking old products which don't put
+        #           permissions on their metadata registry entries.
+        #
+        if method_name is not None:
+            meth=self.unrestrictedTraverse(method_name)
+            if hasattr(meth, 'im_self'):
+                parent = meth.im_self
+            else:
+                try:    parent=aq_parent(aq_inner(meth))
+                except: parent=None
+            if getSecurityManager().validate(None, parent, None, meth):
+                # Ensure the user is allowed to access the object on the
+                # clipboard.
+                if not validate_src:
+                    return
+                try:    parent=aq_parent(aq_inner(object))
+                except: parent=None
+                if getSecurityManager().validate(None, parent, None, object):
+                    return
+                raise Unauthorized, absattr(object.id)
+            else:
+                raise Unauthorized, method_name
 
-            # if we still have a factory, get the add method
-            try: meth=getattr(meth, meth.initial)
-            except: pass
-
-            if hasattr(meth, '__roles__'):
-                roles=meth.__roles__
-                user=REQUEST.get('AUTHENTICATED_USER', None)
-                if (not hasattr(user, 'has_role') or
-                    not user.has_role(roles, self)):
-                    raise 'Unauthorized', (
-                          """You are not authorized to perform this
-                             operation."""
-                          )
-                return
         raise CopyError, MessageDialog(
               title='Not Supported',
               message='The object <EM>%s</EM> does not support this ' \
-                      'operation' % absattr(ob.id),
+                      'operation' % absattr(object.id),
               action='manage_main')
 
 Globals.default__class_init__(CopyContainer)
@@ -413,7 +409,7 @@ class CopySource:
         self.id=id
 
     def cb_isCopyable(self):
-        """Is object copyable? Returns 0 or 1"""
+        # Is object copyable? Returns 0 or 1
         if not (hasattr(self, '_canCopy') and self._canCopy(0)):
             return 0
         if hasattr(self, '_p_jar') and self._p_jar is None:
@@ -421,12 +417,12 @@ class CopySource:
         return 1
 
     def cb_isMoveable(self):
-        """Is object moveable? Returns 0 or 1"""
+        # Is object moveable? Returns 0 or 1
         if not (hasattr(self, '_canCopy') and self._canCopy(1)):
             return 0
         if hasattr(self, '_p_jar') and self._p_jar is None:
             return 0
-        try:    n=self.aq_parent._reserved_names
+        try:    n=aq_parent(aq_inner(self))._reserved_names
         except: n=()
         if absattr(self.id) in n:
             return 0
@@ -438,26 +434,29 @@ def sanity_check(c, ob):
     # This is called on cut/paste operations to make sure that
     # an object is not cut and pasted into itself or one of its
     # subobjects, which is an undefined situation.
-    ob=getattr(ob, 'aq_base', ob)
+    ob = aq_base(ob)
     while 1:
-        if getattr(c, 'aq_base', c) is ob:
+        if aq_base(c) is ob:
             return 0
-        if not hasattr(c, 'aq_parent'):
+        inner = aq_inner(c)
+        if aq_parent(inner) is None:
             return 1
-        c=c.aq_parent
+        c = aq_parent(inner)
 
 def absattr(attr):
     if callable(attr): return attr()
     return attr
 
-
-
 def _cb_encode(d):
     return quote(compress(dumps(d), 9))
 
 def _cb_decode(s):
     return loads(decompress(unquote(s)))
 
+def cookie_path(request):
+    # Return a "path" value for use in a cookie that refers
+    # to the root of the Zope object space.
+    return request['BASEPATH1'] or "/"
 
 
 
@@ -519,3 +518,10 @@ eNotSupported=fMessageDialog(
               title='Not Supported',
               message='The item <EM>%s</EM> does not support this operation.',
               action ='manage_main',)
+
+eNoItemsSpecified=MessageDialog(
+                  title='No items specified',
+                  message='You must select one or more items to perform ' \
+                  'this operation.',
+                  action ='manage_main'
+                  )
