@@ -1,104 +1,38 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 
 """Property sheets"""
-__version__='$Revision: 1.43 $'[11:-2]
+__version__='$Revision: 1.81.4.1 $'[11:-2]
 
 import time, string, App.Management, Globals
+from webdav.WriteLockInterface import WriteLockInterface
 from ZPublisher.Converters import type_converters
-from DocumentTemplate.DT_Util import html_quote
-from Globals import HTMLFile, MessageDialog
+from Globals import DTMLFile, MessageDialog
 from string import find,join,lower,split,rfind
 from Acquisition import Implicit, Explicit
-from App.Common import rfc1123_date
+from App.Common import rfc1123_date, iso8601_date
 from webdav.common import urlbase
 from ExtensionClass import Base
 from Globals import Persistent
+from Traversable import Traversable
+from Acquisition import aq_base
+from AccessControl import getSecurityManager
 
 class View(App.Management.Tabs, Base):
     """A view of an object, typically used for management purposes
+
+    This class provides bits of management framework needed by propertysheets
+    to be used as a view on an object.
     """
 
     def manage_workspace(self, URL1, RESPONSE):
@@ -106,7 +40,7 @@ class View(App.Management.Tabs, Base):
         '''
         RESPONSE.redirect(URL1+'/manage')
 
-    def tpURL(self): return self.id
+    def tpURL(self): return self.getId()
         
     def manage_options(self):
         """Return a manage option data structure for me instance
@@ -126,9 +60,13 @@ class View(App.Management.Tabs, Base):
         r=[]
         for d in self.aq_parent.aq_parent.manage_options:
             path=d['action']
-            r.append({'label': d['label'],
+            option={'label': d['label'],
                       'action': pre+path,
-                      'path': '../../'+path})
+                      'path': '../../'+path}
+            help=d.get('help')
+            if help is not None:
+                option['help']=help
+            r.append(option)
         return r
 
     def tabs_path_info(self, script, path):
@@ -145,7 +83,7 @@ class View(App.Management.Tabs, Base):
         except: return ''
 
 
-class PropertySheet(Persistent, Implicit):
+class PropertySheet(Traversable, Persistent, Implicit):
     """A PropertySheet is a container for a set of related properties and
        metadata describing those properties. PropertySheets may or may not
        provide a web interface for managing its properties."""
@@ -153,16 +91,44 @@ class PropertySheet(Persistent, Implicit):
     _properties=()
     _extensible=1
     icon='p_/Properties_icon'
-    
-    def property_extensible_schema__(self): return self._extensible
+
+    __ac_permissions__=(
+        ('Manage properties', ('manage_addProperty',
+                               'manage_editProperties',
+                               'manage_delProperties',
+                               'manage_changeProperties',
+                               'manage_propertiesForm',
+                               )),
+        ('Access contents information',
+         ('xml_namespace', 'hasProperty', 'getProperty', 'getPropertyType',
+          'propertyIds', 'propertyValues','propertyItems', 'propertyInfo',
+          'propertyMap', ''),
+         ('Anonymous', 'Manager'),
+         ),
+        )
+
+    __reserved_ids= ('values','items') 
+
+    def property_extensible_schema__(self):
+        """Return a flag indicating whether new properties may be
+        added or removed."""
+        return self._extensible
 
     def __init__(self, id, md=None):
         # Create a new property set, using the given id and namespace
         # string. The namespace string should be usable as an xml name-
         # space identifier.
+
+        if id in self.__reserved_ids:
+            raise ValueError, "'%s' is a reserved Id (forbidden Ids are: %s)" % \
+                    (id, self.__reserved_ids)
+
         self.id=id
         self._md=md or {}
-        
+
+    def getId(self):
+        return self.id
+
     def xml_namespace(self):
         # Return a namespace string usable as an xml namespace
         # for this property set.
@@ -175,17 +141,14 @@ class PropertySheet(Persistent, Implicit):
         return self.v_self()
 
     def valid_property_id(self, id):
-        # Return a true value if the given id is valid to use as 
-        # a property id. Note that this method does not consider 
-        # factors other than the actual value of the id, such as 
-        # whether the given id is already in use.
-        if not id or (id[:1]=='_') or (' ' in id):
+        if not id or id[:1]=='_' or (id[:3]=='aq_') \
+           or (' ' in id):
             return 0
         return 1
 
     def hasProperty(self, id):
         # Return a true value if a property exists with the given id.
-        for prop in self.propertyMap():
+        for prop in self._propertyMap():
             if id==prop['id']:
                 return 1
         return 0
@@ -197,12 +160,29 @@ class PropertySheet(Persistent, Implicit):
             return getattr(self.v_self(), id)
         return default
 
+    def getPropertyType(self, id):
+        """Get the type of property 'id', returning None if no
+           such property exists"""
+        pself=self.p_self()
+        for md in pself._properties:
+            if md['id']==id:
+                return md.get('type', 'string')
+        return None
+
+    def _wrapperCheck(self, object):
+        # Raise an error if an object is wrapped.
+        if hasattr(object, 'aq_base'):
+            raise ValueError, 'Invalid property value: wrapped object'
+        return
+
     def _setProperty(self, id, value, type='string', meta=None):
         # Set a new property with the given id, value and optional type.
         # Note that different property sets may support different typing
         # systems.
+        self._wrapperCheck(value)
         if not self.valid_property_id(id):
             raise 'Bad Request', 'Invalid property id, %s.' % id
+
         if not self.property_extensible_schema__():
             raise 'Bad Request', (
                 'Properties cannot be added to this property sheet')
@@ -230,6 +210,7 @@ class PropertySheet(Persistent, Implicit):
         # an attempt will be made to convert the value to the type of the
         # existing property. If a mapping containing meta-data is passed,
         # it will used to _replace_ the properties meta data.
+        self._wrapperCheck(value)
         if not self.hasProperty(id):
             raise 'Bad Request', 'The property %s does not exist.' % id
         propinfo=self.propertyInfo(id)
@@ -266,36 +247,37 @@ class PropertySheet(Persistent, Implicit):
 
     def propertyIds(self):
         # Return a list of property ids.
-        return map(lambda i: i['id'], self.propertyMap())
+        return map(lambda i: i['id'], self._propertyMap())
 
     def propertyValues(self):
         # Return a list of property values.
         return map(lambda i, s=self: s.getProperty(i['id']),
-                   self.propertyMap())
+                   self._propertyMap())
 
     def propertyItems(self):
         # Return a list of (id, property) tuples.
         return map(lambda i, s=self: (i['id'], s.getProperty(i['id'])), 
-                   self.propertyMap())
+                   self._propertyMap())
 
     def propertyInfo(self, id):
         # Return a mapping containing property meta-data
-        for p in self.propertyMap():
+        for p in self._propertyMap():
             if p['id']==id: return p
         raise ValueError, 'The property %s does not exist.' % id
 
-    def propertyMap(self):
+    def _propertyMap(self):
         # Return a tuple of mappings, giving meta-data for properties.
         # Some ZClass instances dont seem to have an _properties, so
         # we have to fake it...
         return self.p_self()._properties
 
-
-
+    def propertyMap(self):
+        # Returns a secure copy of the property definitions.
+        return tuple(map(lambda dict: dict.copy(), self._propertyMap()))
 
     def _propdict(self):
         dict={}
-        for p in self.propertyMap():
+        for p in self._propertyMap():
             dict[p['id']]=p
         return dict
 
@@ -314,13 +296,14 @@ class PropertySheet(Persistent, Implicit):
         # DAV helper method - return one or more propstat elements
         # indicating property names and values for all properties.
         result=[]
-        for item in self.propertyMap():
+        for item in self._propertyMap():
             name, type=item['id'], item.get('type','string')
             value=self.getProperty(name)
+
             if type=='tokens':
-                value=join(value, ' ')
+                value=join(str(value), ' ')
             elif type=='lines':
-                value=join(value, '\n')
+                value=join(str(value), '\n')
             # check for xml property
             attrs=item.get('meta', {}).get('__xml_attrs__', None)
             if attrs is not None:
@@ -329,10 +312,16 @@ class PropertySheet(Persistent, Implicit):
             else:
                 # Quote non-xml items here?
                 attrs='' 
-            prop='  <n:%s%s>%s</n:%s>' % (name, attrs, value, name)
+
+            if hasattr(self,"dav__"+name):
+                prop='  <n:%s%s>%s</n:%s>' % (name, attrs, value, name)
+            else:
+                prop='  <n:%s%s>%s</n:%s>' % (name, attrs, xml_escape(value), name)
+
             result.append(prop)
         if not result: return ''
         result=join(result, '\n')
+        
         return propstat % (self.xml_namespace(), result, '200 OK', '')
 
     def dav__propnames(self, propstat=propstat, join=string.join):
@@ -365,9 +354,9 @@ class PropertySheet(Persistent, Implicit):
             name, type=item['id'], item.get('type','string')
             value=self.getProperty(name)
             if type=='tokens':
-                value=join(value, ' ')
+                value=join(str(value), ' ')
             elif type=='lines':
-                value=join(value, '\n')
+                value=join(str(value), '\n')
             # allow for xml properties
             attrs=item.get('meta', {}).get('__xml_attrs__', None)
             if attrs is not None:
@@ -391,11 +380,11 @@ class PropertySheet(Persistent, Implicit):
 
     # Web interface
     
-    manage=HTMLFile('properties', globals())
+    manage=DTMLFile('dtml/properties', globals())
     def manage_propertiesForm(self, URL1):
         " "
         raise 'Redirect', URL1+'/manage'
-    
+
     def manage_addProperty(self, id, value, type, REQUEST=None):
         """Add a new property via the web. Sets a new property with
         the given id, type, and value."""
@@ -405,11 +394,24 @@ class PropertySheet(Persistent, Implicit):
         if REQUEST is not None:
             return self.manage(self, REQUEST)
 
+    def manage_editProperties(self, REQUEST):
+        """Edit object properties via the web."""
+        for prop in self._propertyMap():
+            name=prop['id']
+            if 'w' in prop.get('mode', 'wd'):
+                value=REQUEST.get(name, '')
+                self._updateProperty(name, value)
+        return MessageDialog(
+               title  ='Success!',
+               message='Your changes have been saved',
+               action ='manage')
+
     def manage_changeProperties(self, REQUEST=None, **kw):
         """Change existing object properties by passing either a mapping
            object of name:value pairs {'foo':6} or passing name=value
            parameters."""
-        if REQUEST is None: props={}
+        if REQUEST is None:
+            props={}
         else: props=REQUEST
         if kw:
             for name, value in kw.items():
@@ -417,6 +419,8 @@ class PropertySheet(Persistent, Implicit):
         propdict=self._propdict()
         for name, value in props.items():
             if self.hasProperty(name):
+                if not 'w' in propdict[name].get('mode', 'wd'):
+                    raise 'BadRequest', '%s cannot be changed' % name
                 self._updateProperty(name, value)
         if REQUEST is not None:
             return MessageDialog(
@@ -424,17 +428,6 @@ class PropertySheet(Persistent, Implicit):
                 message='Your changes have been saved.',
                 action ='manage')
 
-    def manage_editProperties(self, REQUEST):
-        """Edit object properties via the web."""
-        for p in self.propertyMap():
-            n=p['id']
-            self._updateProperty(n, REQUEST.get(n, ''))
-
-        return MessageDialog(
-               title  ='Success!',
-               message='Your changes have been saved',
-               action ='manage')
-
     def manage_delProperties(self, ids=None, REQUEST=None):
         """Delete one or more properties specified by 'ids'."""
         if ids is None:
@@ -447,7 +440,11 @@ class PropertySheet(Persistent, Implicit):
         if REQUEST is not None:
             return self.manage(self, REQUEST)
 
+Globals.default__class_init__(PropertySheet)
+
+
 class Virtual:
+    """A virtual propertysheet stores it's properties in it's instance."""
 
     def __init__(self):
         pass
@@ -463,6 +460,8 @@ class DefaultProperties(Virtual, PropertySheet, View):
     id='default'
     _md={'xmlns': 'http://www.zope.org/propsets/default'}
 
+Globals.default__class_init__(DefaultProperties)
+
 
 class DAVProperties(Virtual, PropertySheet, View):
     """WebDAV properties"""
@@ -472,12 +471,13 @@ class DAVProperties(Virtual, PropertySheet, View):
     pm=({'id':'creationdate',     'mode':'r'},
         {'id':'displayname',      'mode':'r'},
         {'id':'resourcetype',     'mode':'r'},
-        {'id':'getlastmodified',  'mode':'r'},
         {'id':'getcontenttype',   'mode':'r'},
         {'id':'getcontentlength', 'mode':'r'},
         {'id':'source',           'mode':'r'},
+        {'id':'supportedlock',    'mode':'r'},
+        {'id':'lockdiscovery',    'mode':'r'},
         )
-
+     
     def getProperty(self, id, default=None):
         method='dav__%s' % id
         if not hasattr(self, method):
@@ -493,32 +493,37 @@ class DAVProperties(Virtual, PropertySheet, View):
     def _delProperty(self, id):
         raise ValueError, '%s cannot be deleted.' % id
 
-    def propertyMap(self):
+    def _propertyMap(self):
+        # Only use getlastmodified if returns a value
+        if hasattr(self.v_self(), '_p_mtime'):
+            return self.pm + ({'id':'getlastmodified',  'mode':'r'},)
         return self.pm
     
+    def propertyMap(self):
+        return map(lambda dict: dict.copy(), self._propertyMap())
+    
     def dav__creationdate(self):
-        return ''
+        return iso8601_date(43200.0)
 
     def dav__displayname(self):
         return absattr(self.v_self().id)
 
     def dav__resourcetype(self):
         vself=self.v_self()
-        if hasattr(aq_base(vself), 'isAnObjectManager') and \
-           vself.isAnObjectManager:
+        if (hasattr(vself, '__dav_collection__') or
+            getattr(aq_base(vself), 'isAnObjectManager', None)):
             return '<n:collection/>'
         return ''
 
     def dav__getlastmodified(self):
-        vself=self.v_self()
-        if hasattr(vself, '_p_mtime'):
-            return rfc1123_date(vself._p_mtime)
-        return ''
+        return rfc1123_date(self.v_self()._p_mtime)
 
     def dav__getcontenttype(self):
         vself=self.v_self()
         if hasattr(vself, 'content_type'):
-            return vself.content_type
+            return absattr(vself.content_type)
+        if hasattr(vself, 'default_content_type'):
+            return absattr(vself.default_content_type)
         return ''
 
     def dav__getcontentlength(self):
@@ -543,13 +548,50 @@ class DAVProperties(Virtual, PropertySheet, View):
                '  <d:locktype><d:write/></d:locktype>\n' \
                '  </n:lockentry>\n  '
 
+    def dav__lockdiscovery(self):
+        security = getSecurityManager()
+        user = security.getUser().getUserName()
+        
 
-class PropertySheets(Implicit, App.Management.Tabs):
+        vself = self.v_self()
+        out = '\n'
+        if WriteLockInterface.isImplementedBy(vself):
+            locks = vself.wl_lockValues(killinvalids=1)
+            for lock in locks:
+
+                creator = lock.getCreator()[-1]
+                if creator == user: fake=0
+                else:               fake=1
+                    
+                out = '%s\n%s' % (out, lock.asLockDiscoveryProperty('n',fake=fake))
+                
+            out = '%s\n' % out
+
+        return out
+
+
+Globals.default__class_init__(DAVProperties)
+
+
+class PropertySheets(Traversable, Implicit, App.Management.Tabs):
     """A tricky container to keep property sets from polluting
        an object's direct attribute namespace."""
     
     id='propertysheets'
 
+    __ac_permissions__=(
+        ('Manage properties', ('manage_addPropertySheet',
+                               'addPropertySheet',
+                               'delPropertySheet'
+                               )),
+        ('Access contents information',
+         ('items', 'values', 'get', ''),
+         ('Anonymous', 'Manager'),
+         ),
+        ('View management screens', ('manage',)),
+        )
+
+
     webdav =DAVProperties()
     def _get_defaults(self):
         return (self.webdav,)
@@ -561,7 +603,7 @@ class PropertySheets(Implicit, App.Management.Tabs):
 
     def __bobo_traverse__(self, REQUEST, name=None):
         for propset in self.__propsets__():
-            if propset.id==name:
+            if propset.getId()==name:
                 return propset.__of__(self)
         return getattr(self, name)
 
@@ -584,7 +626,8 @@ class PropertySheets(Implicit, App.Management.Tabs):
         
     def get(self, name, default=None):
         for propset in self.__propsets__():
-            if propset.id==name or propset.xml_namespace()==name:
+            if propset.id==name or (hasattr(propset, 'xml_namespace') and \
+                                    propset.xml_namespace()==name):
                 return propset.__of__(self)
         return default
 
@@ -603,21 +646,20 @@ class PropertySheets(Implicit, App.Management.Tabs):
     def delPropertySheet(self, name):
         result=[]
         for propset in self.aq_parent.__propsets__:
-            if propset.id != name and  propset.xml_namespace() != name:
+            if propset.getId() != name and  propset.xml_namespace() != name:
                 result.append(propset)
         self.parent.__propsets__=tuple(result)
-
-    # Why?
-    #def __del__(self):
-    #    self.parent=None
         
     def __len__(self):
         return len(self.__propsets__())
 
+    def getId(self):
+        return self.id
+
 
     # Management interface:
 
-    manage=Globals.HTMLFile('propertysheets', globals())
+    manage=Globals.DTMLFile('dtml/propertysheets', globals())
 
     def manage_options(self):
         """Return a manage option data structure for me instance
@@ -645,6 +687,8 @@ class PropertySheets(Implicit, App.Management.Tabs):
         return PropertySheets.inheritedAttribute('tabs_path_info')(
             self, script, path)
 
+Globals.default__class_init__(PropertySheets)
+
 
 class DefaultPropertySheets(PropertySheets):
     """A PropertySheets container that contains a default property
@@ -655,37 +699,54 @@ class DefaultPropertySheets(PropertySheets):
     def _get_defaults(self):
         return (self.default, self.webdav)
 
+Globals.default__class_init__(DefaultPropertySheets)
+
 
 class FixedSchema(PropertySheet):
+    """A Fixed-schema property sheet has no control over it's schema
+
+    It gets its schema from another proprtysheet but has control over
+    its value storage.
+
+    This mix-in is used for ZClass instance proprtysheets, which store
+    their data in instances, but get their schema from the
+    proprtysheet managed in the ZClass.
+    """
 
     def __init__(self, id, base, md=None):
         FixedSchema.inheritedAttribute('__init__')(self, id, md)
         self._base=base
 
-    def propertyMap(self):
+    def _propertyMap(self):
         # Return a tuple of mappings, giving meta-data for properties.
-        r=[]
-        for d in self._base.propertyMap():
-            mode=d.get('mode', 'wd')
+        r = []
+        for d in self._base._propertyMap():
+            d = d.copy()
+            mode = d.get('mode', 'wd')
             if 'd' in mode:
-                dd={}
-                dd.update(d)
-                d=dd
                 d['mode']=filter(lambda c: c != 'd', mode)
             r.append(d)
             
         return tuple(r)
 
+    def propertyMap(self):
+        return self._propertyMap()
+
     def property_extensible_schema__(self):
         return 0
         return self._base._extensible
+
+Globals.default__class_init__(FixedSchema)
     
 
 
 class vps(Base):
-    # The vps object implements a "computed attribute" - it ensures
-    # that a PropertySheets instance is returned when the propertysheets
-    # attribute of a PropertyManager is accessed.
+    """Virtual Propertysheets
+
+    The vps object implements a "computed attribute" - it ensures
+    that a PropertySheets instance is returned when the propertysheets
+    attribute of a PropertyManager is accessed.
+    """
     def __init__(self, c=PropertySheets):
         self.c=c
         
@@ -697,7 +758,15 @@ def absattr(attr):
         return attr()
     return attr
 
-def aq_base(ob):
-    if hasattr(ob, 'aq_base'):
-        return ob.aq_base
-    return ob
+
+def xml_escape(v):
+    """ convert any content from ISO-8859-1 to UTF-8
+    The main use is to escape non-US object property values
+    (e.g. containing accented characters). Also we convert "<" and ">"
+    to entities to keep the properties XML compliant. 
+    """
+    v = str(v)
+    v = v.replace('&', '&amp;')
+    v = v.replace('<', '&lt;')
+    v = v.replace('>', '&gt;')
+    return  unicode(v,"latin-1").encode("utf-8")
