@@ -1,104 +1,29 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
-__doc__='''Generic Database adapter
+__doc__='''Generic Database adapter'''
 
 
-$Id: DA.py,v 1.75.4.3 2000/03/20 16:03:28 brian Exp $'''
-__version__='$Revision: 1.75.4.3 $'[11:-2]
+__version__='$Revision: 1.104 $'[11:-2]
 
-import OFS.SimpleItem, Aqueduct, RDB
+import OFS.SimpleItem, Aqueduct, RDB, re
 import DocumentTemplate, marshal, md5, base64, Acquisition, os
 from Aqueduct import decodestring, parse
 from Aqueduct import custom_default_report, default_input_form
-from Globals import HTMLFile, MessageDialog
+from Globals import DTMLFile, MessageDialog
 from cStringIO import StringIO
 import sys, Globals, OFS.SimpleItem, AccessControl.Role
-from string import atoi, find, join, split
+from string import atoi, find, join, split, rstrip
 import DocumentTemplate, sqlvar, sqltest, sqlgroup
-from AccessControl.User import verify_watermark
-from DocumentTemplate.DT_Util import cDocument
 from time import time
 from zlib import compress, decompress
 from DateTime.DateTime import DateTime
@@ -108,8 +33,10 @@ import DocumentTemplate.DT_Util
 from cPickle import dumps, loads
 from Results import Results
 from App.Extensions import getBrain
-
-
+from AccessControl import getSecurityManager
+from AccessControl.DTML import RestrictedDTML
+from webdav.Resource import Resource
+from webdav.Lockable import ResourceLockedError
 try: from IOBTree import Bucket
 except: Bucket=lambda:{}
 
@@ -125,17 +52,17 @@ class nvSQL(DocumentTemplate.HTML):
     _proxy_roles=()
 
 
-class SQL(cDocument, nvSQL):
+class SQL(RestrictedDTML, ExtensionClass.Base, nvSQL):
     # Validating SQL template for Zope SQL Methods.
     pass
 
 
-
 class DA(
     Aqueduct.BaseQuery,Acquisition.Implicit,
     Globals.Persistent,
     AccessControl.Role.RoleManager,
     OFS.SimpleItem.Item,
+    Resource
     ):
     'Database Adapter'
 
@@ -144,14 +71,21 @@ class DA(
     cache_time_=0
     max_cache_=100
     class_name_=class_file_=''
+    _zclass=None
     allow_simple_one_argument_traversal=None
     template_class=SQL
     
     manage_options=(
-        {'label':'Edit', 'action':'manage_main'},
-        {'label':'Test', 'action':'manage_testForm'},
-        {'label':'Advanced', 'action':'manage_advancedForm'},
-        {'label':'Security', 'action':'manage_access'},
+        (
+        {'label':'Edit', 'action':'manage_main',
+         'help':('ZSQLMethods','Z-SQL-Method_Edit.stx')},
+        {'label':'Test', 'action':'manage_testForm',
+         'help':('ZSQLMethods','Z-SQL-Method_Test.stx')},
+        {'label':'Advanced', 'action':'manage_advancedForm',
+         'help':('ZSQLMethods','Z-SQL-Method_Advanced.stx')},
+        )
+        +AccessControl.Role.RoleManager.manage_options
+        +OFS.SimpleItem.Item.manage_options
         )
  
     # Specify how individual operations add up to "permissions":
@@ -159,19 +93,20 @@ class DA(
         ('View management screens',
          (
         'manage_main', 'index_html',
-        'manage_advancedForm', 'PrincipiaSearchSource'
+        'manage_advancedForm', 'PrincipiaSearchSource', 'document_src'
         )),
         ('Change Database Methods',
-         ('manage_edit','manage_advanced', 'manage_testForm','manage_test')),
+         ('manage_edit','manage_advanced', 'manage_testForm','manage_test',
+          'manage_product_zclass_info', 'PUT')),
         ('Use Database Methods', ('__call__',''), ('Anonymous','Manager')),
         )
    
 
     def __init__(self, id, title, connection_id, arguments, template):
-        self.id=id
+        self.id=str(id)
         self.manage_edit(title, connection_id, arguments, template)
     
-    manage_advancedForm=HTMLFile('advanced', globals())
+    manage_advancedForm=DTMLFile('dtml/advanced', globals())
 
     test_url___roles__=None
     def test_url_(self):
@@ -225,10 +160,15 @@ class DA(
             return self._er(title,connection_id,arguments,template,
                             SUBMIT,sql_pref__cols,sql_pref__rows,REQUEST)
 
-        self.title=title
-        self.connection_id=connection_id
+        if self.wl_isLocked():
+            raise ResourceLockedError, 'SQL Method is locked via WebDAV'
+        
+        self.title=str(title)
+        self.connection_id=str(connection_id)
+        arguments=str(arguments)
         self.arguments_src=arguments
         self._arg=parse(arguments)
+        template=str(template)
         self.src=template
         self.template=t=self.template_class(template)
         t.cook()
@@ -243,7 +183,7 @@ class DA(
 
     def manage_advanced(self, max_rows, max_cache, cache_time,
                         class_name, class_file, direct=None,
-                        REQUEST=None):
+                        REQUEST=None, zclass=''):
         """Change advanced properties
 
         The arguments are:
@@ -273,12 +213,30 @@ class DA(
         used.
   
         """
+        # paranoid type checking
+        if type(max_rows) is not type(1):
+            max_rows=atoi(max_rows)
+        if type(max_cache) is not type(1):
+            max_cache=atoi(max_cache)
+        if type(cache_time) is not type(1):
+            cache_time=atoi(cache_time)            
+        class_name=str(class_name)
+        class_file=str(class_file)
+        
         self.max_rows_ = max_rows
         self.max_cache_, self.cache_time_ = max_cache, cache_time
         self._v_cache={}, Bucket()
         self.class_name_, self.class_file_ = class_name, class_file
         self._v_brain=getBrain(self.class_file_, self.class_name_, 1)
         self.allow_simple_one_argument_traversal=direct
+
+        if zclass:
+            for d in self.aq_acquire('_getProductRegistryData')('zclasses'):
+                if ("%s/%s" % (d.get('product'),d.get('id'))) == zclass:
+                    self._zclass=d['meta_class']
+                    break
+
+        
         if REQUEST is not None:
             m="ZSQL Method advanced settings have been set"
             return self.manage_advancedForm(self,REQUEST,manage_tabs_message=m)
@@ -291,7 +249,42 @@ class DA(
     def PrincipiaSearchSource(self):
         """Return content for use by the Find machinery."""
         return '%s\n%s' % (self.arguments_src, self.src)
-    
+
+
+    # WebDAV / FTP support
+
+    default_content_type = 'text/plain'
+
+    def document_src(self, REQUEST=None, RESPONSE=None):
+        """Return unprocessed document source."""
+        if RESPONSE is not None:
+            RESPONSE.setHeader('Content-Type', 'text/plain')
+        return '<params>%s</params>\n%s' % (self.arguments_src, self.src)
+
+    def manage_FTPget(self):
+        """Get source for FTP download"""
+        self.REQUEST.RESPONSE.setHeader('Content-Type', 'text/plain')
+        return '<params>%s</params>\n%s' % (self.arguments_src, self.src)
+
+    def PUT(self, REQUEST, RESPONSE):
+        """Handle put requests"""
+        self.dav__init(REQUEST, RESPONSE)
+        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
+        body = REQUEST.get('BODY', '')
+        m = re.match('\s*<params>(.*)</params>\s*\n', body, re.I | re.S)
+        if m:
+            self.arguments_src = m.group(1)
+            self._arg=parse(self.arguments_src)
+            body = body[m.end():]
+        template = body
+        self.src = template
+        self.template=t=self.template_class(template)
+        t.cook()
+        self._v_cache={}, Bucket()
+        RESPONSE.setStatus(204)
+        return RESPONSE
+
+
     def manage_testForm(self, REQUEST):
         " "
         input_src=default_input_form(self.title_or_id(),
@@ -300,26 +293,30 @@ class DA(
         return DocumentTemplate.HTML(input_src)(self, REQUEST, HTTP_REFERER='')
 
     def manage_test(self, REQUEST):
-        ' '
-
-        src="Could not render the query template!"
+        """Test an SQL method."""
+        # Try to render the query template first so that the rendered
+        # source will be available for the error message in case some
+        # error occurs...
+        try:    src=self(REQUEST, src__=1)
+        except: src="Could not render the query template!"
         result=()
         t=v=tb=None
         try:
             try:
-                src=self(REQUEST, src__=1)
-                if find(src,'\0'): src=join(split(src,'\0'),'\n'+'-'*60+'\n')
-                result=self(REQUEST, test__=1)
+                src, result=self(REQUEST, test__=1)
+                if find(src,'\0'):
+                    src=join(split(src,'\0'),'\n'+'-'*60+'\n')
                 if result._searchable_result_columns():
                     r=custom_default_report(self.id, result)
                 else:
-                    r='This was not a query.'
+                    r='This statement returned no results.'
             except:
                 t, v, tb = sys.exc_info()
                 r='<strong>Error, <em>%s</em>:</strong> %s' % (t, v)
 
             report=DocumentTemplate.HTML(
-                '<html><BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">\n'
+                '<html>\n'
+                '<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">\n'
                 '<dtml-var manage_tabs>\n<hr>\n%s\n\n'
                 '<hr><strong>SQL used:</strong><br>\n<pre>\n%s\n</pre>\n<hr>\n'
                 '</body></html>'
@@ -335,9 +332,9 @@ class DA(
         
         finally: tb=None
 
-    def index_html(self, URL1):
-        " "
-        raise 'Redirect', ("%s/manage_testForm" % URL1)
+    def index_html(self, REQUEST):
+        """ """
+        REQUEST.RESPONSE.redirect("%s/manage_testForm" % REQUEST['URL1'])
 
     def _searchable_arguments(self): return self._arg
 
@@ -359,7 +356,8 @@ class DA(
                 key=keys[-1]
                 q=tcache[key]
                 del tcache[key]
-                del cache[q]
+                if int(cache[q][0]) == key:
+                    del cache[q]
                 del keys[-1]
                 
         if cache.has_key(query):
@@ -403,25 +401,18 @@ class DA(
         
         if hasattr(self, 'aq_parent'):
             p=self.aq_parent
+            if self._isBeingAccessedAsZClassDefinedInstanceMethod():
+                p=p.aq_parent
         else: p=None
 
         argdata=self._argdata(REQUEST)
         argdata['sql_delimiter']='\0'
         argdata['sql_quote__']=dbc.sql_quote__
 
-        # Also need the authenticated user.
-        auth_user=REQUEST.get('AUTHENTICATED_USER', None)
-        if auth_user is None:
-            auth_user=getattr(self, 'REQUEST', None)
-            if auth_user is not None:
-                try: auth_user=auth_user.get('AUTHENTICATED_USER', None)
-                except: auth_user=None
-                
-        if auth_user is not None:
-            verify_watermark(auth_user)
-            argdata['AUTHENTICATED_USER']=auth_user
-
-        query=apply(self.template, (p, argdata))
+        security=getSecurityManager()
+        security.addContext(self)
+        try:     query=apply(self.template, (p,), argdata)
+        finally: security.removeContext(self)
 
         if src__: return query
 
@@ -433,15 +424,23 @@ class DA(
         else:
             brain=self._v_brain=getBrain(self.class_file_, self.class_name_)
 
+        zc=self._zclass
+        if zc is not None: zc=zc._zclass_
+
         if type(result) is type(''):
             f=StringIO()
             f.write(result)
             f.seek(0)
-            result=RDB.File(f,brain,p)
+            result=RDB.File(f,brain,p, zc)
         else:
-            result=Results(result, brain, p)
+            result=Results(result, brain, p, zc)
         columns=result._searchable_result_columns()
         if test__ and columns != self._col: self._col=columns
+
+        # If run in test mode, return both the query and results so
+        # that the template doesn't have to be rendered twice!
+        if test__: return query, result
+
         return result
 
     def da_has_single_argument(self): return len(self._arg)==1
@@ -467,6 +466,26 @@ class DA(
     def connected(self):
         return getattr(getattr(self, self.connection_id), 'connected')()
 
+
+    def manage_product_zclass_info(self):
+        r=[]
+        Z=self._zclass
+        Z=getattr(Z, 'aq_self', Z)
+        for d in self.aq_acquire('_getProductRegistryData')('zclasses'):
+            z=d['meta_class']
+            if hasattr(z._zclass_,'_p_deactivate'):
+                # Eek, persistent
+                continue
+            x={}
+            x.update(d)
+            x['selected'] = (z is Z) and 'selected' or ''
+            del x['meta_class']
+            r.append(x)
+            
+        return r 
+
+
+
 Globals.default__class_init__(DA)
 
 
