@@ -1,85 +1,13 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 """Provide conversion between Python pickles and XML
@@ -93,38 +21,30 @@ from string import replace
 import struct
 import base64
 import string
-import regex
 import pickle
 import tempfile
 import marshal
 import xyap
 
+mdumps = marshal.dumps
+mloads = marshal.loads
 
 xyap=xyap.xyap
 
 ListType=type([])
-#  Create a list of all the characters
-L = map(chr,range(256))
 
-#  Create an empty dictionary
-d = {}     
-
-#  Create a dictionary d that maps each character to its
-#  repr form
-
-for c in L:
-    d[c] = repr(c)[1:-1]
-    
-#  Modify values in the dictionary
-d2={}
-d2['<'] = "\\074"
-d2['>'] = "\\076"
-d2['&'] = "\\046"
-d['\n'] = "\\n\n"
-d['\t'] = "\\t"
-d['\\'] = "\\"
-d['\r'] = "\\r"
-d["'"] = "\\'"
+# Create repr mappong
+reprs = {}
+for c in map(chr,range(256)): reprs[c] = repr(c)[1:-1]
+reprs['\n'] = "\\n\n"
+reprs['\t'] = "\\t"
+reprs['\\'] = "\\\\"
+reprs['\r'] = "\\r"
+reprs["'"] = "\\'"
+reprs2={}
+reprs2['<'] = "\\074"
+reprs2['>'] = "\\076"
+reprs2['&'] = "\\046"
 
 
 #  Function convert takes a string and converts it to either
@@ -133,7 +53,7 @@ d["'"] = "\\'"
 def convert(S, find=string.find):
     new = ''
     encoding = 'repr'
-    new = string.join(map(lambda s: d[s], S), '')
+    new = string.join(map(reprs.get, S), '')
     if len(new) > (1.4*len(S)):
         encoding = 'base64'
         new = base64.encodestring(S)[:-1]
@@ -142,7 +62,7 @@ def convert(S, find=string.find):
             new='<![CDATA[\n\n'+new+'\n\n]]>'
             encoding='cdata'
         else:
-            new=string.join(map(lambda s: d2[s], new), '')
+            new=string.join(map(lambda s: reprs2.get(s,s), new), '')
     return encoding, new
 
 #  Function unconvert takes a encoding and a string and 
@@ -168,23 +88,12 @@ def p64(v, pack=struct.pack):
 
 def u64(v, unpack=struct.unpack):
     h, v = unpack(">ii", v)
-    if v < 0: v=t32-v
+    if v < 0: v=t32+v
     if h:
-        if h < 0: h=t32-h
+        if h < 0: h=t32+h
         v=h*t32+v
     return v
 
-def cp(f1, f2, l):
-    read=f1.read
-    write=f2.write
-    n=8192
-    
-    while l > 0:
-        if n > l: n=l
-        d=read(n)
-        write(d)
-        l = l - len(d)
-
 class Global:
 
     def __init__(self, module, name): 
@@ -222,7 +131,12 @@ def xmlstr(v):
 
 class Int(Scalar): pass
 class Long(Scalar): 
-    def value(self): return str(self._v)[:-1]
+    def value(self):
+        result = str(self._v)
+        if result[-1:] == 'L':
+            return result[:-1]
+        return result
+
 class Float(Scalar): pass
 class String(Scalar):
     def __init__(self, v, encoding=''):
@@ -275,7 +189,7 @@ class Value(Wrapper): pass
 class Dictionary(Collection):
     def __init__(self): self._d=[]
     def __len__(self): return len(self._d)
-    def __setitem__(self, k, v): self._d.append(k,v)
+    def __setitem__(self, k, v): self._d.append((k,v))
     def value(self, indent):
         return string.join(
             map(lambda i, ind=' '*indent, indent=indent+4:
@@ -527,12 +441,6 @@ def end_string(self, tag, data):
     if a.has_key('id'): self._pickleids[a['id']]=v
     return v
 
-def end_none(self,tag,data):
-    return None
-
-def end_reference(self, tag, data):
-    return self._pickleids[data[1]['id']]
-
 def end_list(self, tag, data):
     v=data[2:]
     a=data[1]
@@ -552,14 +460,13 @@ def end_dictionary(self, tag, data):
     if a.has_key('id'): self._pickleids[a['id']]=D
     return D
 
-def end_item(self, tag, data):
-    v=data[2:]
-    return v
-
 class xmlUnpickler(NoBlanks, xyap):
     start_handlers={'pickle': start_pickle}
     end_handlers={
         'int':
+        lambda self,tag,data,atoi=string.atoi,name=name:
+            atoi(name(self, tag, data)),
+        'long':
         lambda self,tag,data,atoi=string.atoi,name=name:
             atoi(name(self, tag, data)),
         'boolean':
@@ -572,20 +479,17 @@ class xmlUnpickler(NoBlanks, xyap):
         'float':
         lambda self,tag,data,atof=string.atof,name=name:
             atof(name(self, tag, data)),
-        'none': end_none,
+        'none': lambda self, tag, data: None,
         'list': end_list,
         'tuple': end_tuple,
         'dictionary': end_dictionary,
         'key': lambda self, tag, data: data[2],
         'value': lambda self, tag, data: data[2],
-        'item': end_item,
-        'reference': end_reference,
+        'item': lambda self, tag, data: data[2:],
+        'reference': lambda self, tag, data: self._pickleids[data[1]['id']],
         'state': lambda self, tag, data: data[2],
         'klass': lambda self, tag, data: data[2],
         }
-
-def save_none(self, tag, data):
-    return 'N'
     
 def save_int(self, tag, data):
     binary=self.binary
@@ -609,6 +513,27 @@ def save_float(self, tag, data):
     else: v='F'+name(self, tag, data)+'\012'
     return v
 
+def save_put(self, v, attrs):
+    id=attrs.get('id','')
+    if id:
+        prefix=string.rfind(id,'.')
+        if prefix >= 0: id=id[prefix+1:]
+        elif id[0]=='i': id=id[1:]
+        if self.binary:
+            id=string.atoi(id)
+            s=mdumps(id)[1:]
+            if (id < 256):
+                id=s[0]
+                put='q'
+            else:
+                id=s
+                put='r'
+            id=put+id
+        else:
+            id="p"+id+"\012"
+        return v+id
+    return v
+
 def save_string(self, tag, data):
     binary=self.binary
     v=''
@@ -629,132 +554,37 @@ def save_string(self, tag, data):
             v='T'+s+v
         put='q'
     else: v="S'"+v+"'\012"
-    if a.has_key('id'):
-        id=a['id']
-        prefix=string.rfind(id,'.')
-        if prefix>=0: id=id[prefix+1:]
-        if binary:
-            id=string.atoi(id)
-            s=mdumps(id)[1:]
-            if (id < 256):
-                id=s[0]
-                put='q'
-            else:
-                id=s
-                put='r'
-            v=v+put+id
-        else: v=v+put+id+"\012"
-    return v
+    return save_put(self, v, a)
 
 def save_tuple(self, tag, data):
-    binary=self.binary
     T=data[2:]
-    a=data[1]
-    v=''
-    put='p'
-    for x in T:
-        v=v+x
-    if v is '': return ')'
-    v='('+v+'t'
-    if a.has_key('id'):
-        id=a['id']
-        prefix=string.rfind(id,'.')
-        if prefix>=0: id=id[prefix+1:]
-        if binary:
-            id=string.atoi(id)
-            s=mdumps(id)[1:]
-            if (id < 256):
-                id=s[0]
-                put='q'
-            else:
-                id=s
-                put='r'
-            v=v+put+id
-        else:
-            v=v+put+id+'\012'
-    return v
+    if not T: return ')'
+    return save_put(self, '('+string.join(T,'')+'t', data[1])
 
 def save_list(self, tag, data):
-    binary=self.binary
     L=data[2:]
     a=data[1]
-    v=''
-    x=0
-    if a.has_key('id'):
-        id=a['id']
-        prefix=string.rfind(id,'.')
-        if prefix>=0: id=id[prefix+1:]
-    if binary:
-        while x<len(L):
-            v=v+L[x]
-            x=x+1
-        if id:
-            id=string.atoi(id)
-            s=mdumps(id)[1:]
-            if (id < 256):
-                id=s[0]
-                put='q'
-            else:
-                id=s
-                put='r'
-            if v is not '':
-                v=']'+put+id+'('+v+'e'
-            else: v=']'+put+id
+    if self.binary:
+        v=save_put(self, ']', a)
+        if L: v=v+'('+string.join(L,'')+'e'
     else:
-        while x<len(L):
-            v=v+L[x]+'a'
-            x=x+1
-        if id:
-            v='(lp'+id+'\012'+v
-    if v=='': v=']'
-    
+        v=save_put(self, '(l', a)
+        if L: v=string.join(L,'a')+'a'
     return v
 
 def save_dict(self, tag, data):
-    binary=self.binary
     D=data[2:]
-    a=data[1]
-    id=a['id']
-    prefix=string.rfind(id,'.')
-    if prefix>=0: id=id[prefix+1:]
-    if binary:
-        id=string.atoi(id)
-        s=mdumps(id)[1:]
-        if (id < 256):
-            id=s[0]
-            put='q'
-        else:
-            id=s
-            put='r'
-        v='}'+put+id
-        if len(D)>0: v=v+'('
-        x=0
-        while x < len(D):
-            v=v+D[x]
-            x=x+1
-        if len(D)>0:
-            v=v+'u'
+    if self.binary:
+        v=save_put(self, '}', data[1])
+        if D: v=v+'('+string.join(D,'')+'u'
     else:
-        v='(dp'+id+'\012'
-        x=0
-        while x<len(D):
-            v=v+D[x]+'s'
-            x=x+1
-    return v
-
-def save_item(self, tag, data):
-    v=''
-    for x in data[2:]:
-        v=v+x
-    return v        
-
-def save_pickle(self, tag, data):
-    v=data[2]+'.'
+        v=save_put(self, '(d', data[1])
+        if D: v=v+string.join(D,'s')+'s'
     return v
 
 def save_reference(self, tag, data):
     binary=self.binary
-    a=data[1]
+    a=data[1]    
     id=a['id']
     prefix=string.rfind(id,'.')
     if prefix>=0: id=id[prefix+1:]
@@ -774,93 +604,42 @@ def save_reference(self, tag, data):
     return v
 
 def save_object(self, tag, data):
-    binary=self.binary
-    a=data[1]
-    v='('
-    j=0
-    id=a['id']
-    prefix=string.rfind(id,'.')
-    if prefix>=0: id=id[prefix+1:]
-    put='p'
-    if binary:
-        id=string.atoi(id)
-        s=mdumps(id)[1:]
-        if (id < 256):
-            id=s[0]
-            put='q'
-        else:
-            id=s
-            put='r'
-        for x in data[2:]:
-            if j==0: v=v + x
-            elif j==1:
-                x=x[1:]
-                stop=string.rfind(x,'t')
-                if stop>=0: x=x[:stop]
-                v=v+x+'o'+put+id
-            elif j==2: v=v+x
-            j=j+1
-    else:
-        for x in data[2:]:
-            if j==0: v=v+x
-            elif j==1:
-                x=x[1:]
-                stop=string.rfind(x,'t')
-                if stop>=0: x=x[:stop]
-                v=v+x+'o'+put+id+'\012'
-            elif j==2: v=v+x
-            j=j+1
-    v=v+'b'
-    if a.has_key('id'): self._pickleids[a['id']]=v
+    v='('+data[2]
+    x=data[3][1:]
+    stop=string.rfind(x,'t')  # This seems
+    if stop>=0: x=x[:stop]    # wrong!
+    v=save_put(self, v+x+'o', data[1])
+    v=v+data[4]+'b' # state
     return v
 
 def save_global(self, tag, data):
-    binary=self.binary
     a=data[1]
-    if a.has_key('id'):
-        id=a['id']
-        prefix=string.rfind(id,'.')
-        if prefix>=0: id=id[prefix+1:]
-        put='p'
-        if binary:
-            id=string.atoi(id)
-            s=mdumps(id)[1:]
-            if (id < 256):
-                id=s[0]
-                put='q'
-            else:
-                id=s
-                put='r'
-            v='c'+a['module']+'\012'+a['name']+'\012'+put+id
-        else:
-            v=a['module']+'\012'+a['name']+'\012'+put+id+'\012'
-        self._pickleids[a['id']]=v
-    else: v=a['module']+'\012'+a['name']+'\012'
-    return v
+    return save_put(self, 'c'+a['module']+'\012'+a['name']+'\012', a)
 
 def save_persis(self, tag, data):
-    binary=self.binary
     v=data[2]
-    if  binary:
+    if  self.binary:
         v=v+'Q'
     else:
         v='P'+v
     return v
 
-
 class xmlPickler(NoBlanks, xyap):
-    start_handlers={'pickle': start_pickle,}
+    start_handlers={
+        'pickle': lambda self, tag, attrs: [tag, attrs],
+        }
     end_handlers={
-        'pickle': save_pickle,
-        'none': save_none,
+        'pickle': lambda self, tag, data: data[2]+'.',
+        'none': lambda self, tag, data: 'N',
         'int': save_int,
+        'long': lambda self, tag, data: 'L'+data[2]+'L\012',
         'float': save_float,
         'string': save_string,
         'reference': save_reference,
         'tuple': save_tuple,
         'list': save_list,
         'dictionary': save_dict,
-        'item': save_item,
+        'item': lambda self, tag, data, j=string.join: j(data[2:],''),
         'value': lambda self, tag, data: data[2],
         'key' : lambda self, tag, data: data[2],
         'object': save_object,
@@ -889,9 +668,11 @@ def test():
     t=()
     l=[]
     s=''
+    L = long('999999999999')
     x.append(t)
     x.append(l)
     x.append(s)
+    x.append(L)
     x.append(55555)
     x.append(13)
     r=[x]
@@ -966,9 +747,11 @@ def test2():
     t=()
     l=[]
     s=''
+    L = long('999999999999')
     x.append(t)
     x.append(l)
     x.append(s)
+    x.append(L)
     x.append(5)
     x.append(13)
     print x, '\012'
