@@ -1,99 +1,28 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 """Rendering object hierarchies as Trees
 """
-__rcs_id__='$Id: TreeTag.py,v 1.39 1999/11/12 16:25:49 brian Exp $'
-__version__='$Revision: 1.39 $'[11:-2]
+__rcs_id__='$Id: TreeTag.py,v 1.49 2001/11/28 15:51:17 matt Exp $'
+__version__='$Revision: 1.49 $'[11:-2]
 
 from DocumentTemplate.DT_Util import *
 from DocumentTemplate.DT_String import String
 
-from string import join, split, rfind, find, translate
+from string import join, split, rfind, find, translate, replace
 from urllib import quote, unquote
 from zlib import compress, decompress
 from binascii import b2a_base64, a2b_base64
+import re
 
 tbl=join(map(chr, range(256)),'')
 tplus=tbl[:ord('+')]+'-'+tbl[ord('+')+1:]
@@ -115,33 +44,37 @@ class Tree:
                           # opened_decoration=None,
                           # closed_decoration=None,
                           # childless_decoration=None,
-                          assume_children=1)
+                          assume_children=1,
+                          urlparam=None, prefix=None)
         has_key=args.has_key
 
-        if has_key('name'): name=args['name']
-        elif has_key(''): name=args['name']=args['']
+        if has_key('') or has_key('name') or has_key('expr'):
+            name,expr=name_param(args,'tree',1)
+
+            if expr is not None: args['expr']=expr
+            elif has_key(''): args['name']=name
         else: name='a tree tag'
 
         if has_key('branches_expr'):
             if has_key('branches'):
                 raise ParseError, _tm(
                     'branches and  and branches_expr given', 'tree')
-            args['branches_expr']=VSEval.Eval(
-                args['branches_expr'], expr_globals).eval
+            args['branches_expr']=Eval(args['branches_expr']).eval
         elif not has_key('branches'): args['branches']='tpValues'
 
         if not has_key('id'): args['id']='tpId'
         if not has_key('url'): args['url']='tpURL'
         if not has_key('childless_decoration'):
             args['childless_decoration']=''
+
+        prefix = args.get('prefix')
+        if prefix and not simple_name(prefix):
+            raise ParseError, _tm(
+                'prefix is not a simple name', 'tree')
         
         self.__name__ = name
         self.section=section.blocks
         self.args=args
-        if args.has_key('expr'):
-            if args.has_key('name'):
-                raise ParseError, _tm('name and expr given', 'tree')
-            args['expr']=VSEval.Eval(args['expr'], expr_globals)
             
 
     def render(self,md):
@@ -207,7 +140,21 @@ def tpRender(self, md, section, args,
         if md.has_key('collapse_all'):
             state=[id,[]],
         elif md.has_key('expand_all'):
-            state=[id, tpValuesIds(self, args['branches'], args)],
+            have_arg=args.has_key
+            if have_arg('branches'):
+                def get_items(node, branches=args['branches'], md=md):
+                    get = md.guarded_getattr
+                    if get is None:
+                        get = getattr
+                    items = get(node, branches)
+                    return items()
+            elif have_arg('branches_expr'):
+                def get_items(node, branches_expr=args['branches_expr'], md=md):
+                    md._push(InstanceDict(node, md))
+                    items=branches_expr(md)
+                    md._pop()
+                    return items
+            state=[id, tpValuesIds(self, get_items, args)],
         else:
             if md.has_key('tree-s'):
                 state=md['tree-s']
@@ -237,6 +184,11 @@ def tpRender(self, md, section, args,
     treeData={'tree-root-url': root,
               'tree-colspan': colspan,
               'tree-state': state }
+
+    prefix = args.get('prefix')
+    if prefix:
+        for k, v in treeData.items():
+            treeData[prefix + replace(k[4:], '-', '_')] = v
     
     md._push(InstanceDict(self, md))
     md._push(treeData)
@@ -269,9 +221,10 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
             url = (url and ('%s/%s' % (url, tpUrl))) or tpUrl
             root_url = root_url or tpUrl
 
-    treeData['tree-item-url']=url
-    treeData['tree-level']=level
-    treeData['tree-item-expanded']=0
+    ptreeData = add_with_prefix(treeData, 'tree', args.get('prefix'))
+    ptreeData['tree-item-url']=url
+    ptreeData['tree-level']=level
+    ptreeData['tree-item-expanded']=0
     idattr=args['id']
 
     output=data.append
@@ -288,14 +241,14 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
                 break
         if not exp: items=1
 
+    get=md.guarded_getattr
+    if get is None:
+        get = getattr
+
     if items is None:
-        validate=md.validate
         if have_arg('branches') and hasattr(self, args['branches']):
-            if validate is None or not hasattr(self, 'aq_acquire'):
-                items=getattr(self, args['branches'])
-            else:
-                items=self.aq_acquire(args['branches'],validate,md)
-            items=items()
+            items = get(self, args['branches'])
+            items = items()
         elif have_arg('branches_expr'):
             items=args['branches_expr'](md)
 
@@ -303,20 +256,19 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
 
     if items and items != 1:
 
-        if validate is not None:
+        getitem = getattr(md, 'guarded_getitem', None)
+        if getitem is not None:
             unauth=[]
-            index=0
-            for i in items:
-                try: v=validate(items,items,index,i,md)
-                except: v=0
-                if not v: unauth.append(index)
-                index=index+1
-
+            for index in range(len(items)):
+                try:
+                    getitem(items, index)
+                except ValidationError:
+                    unauth.append(index)
             if unauth:
                 if have_arg('skip_unauthorized') and args['skip_unauthorized']:
                     items=list(items)
                     unauth.reverse()
-                    for i in unauth: del items[i]
+                    for index in unauth: del items[index]
                 else:
                     raise ValidationError, unauth
 
@@ -381,15 +333,23 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
             ####################################
 
             script=md['SCRIPT_NAME']
-            if exp:
-                treeData['tree-item-expanded']=1
-                output('<A NAME="%s" HREF="%s?tree-c=%s#%s">'
-                       '<IMG SRC="%s/p_/mi" ALT="-" BORDER=0></A>' %
-                       (id, root_url, s, id, script))
+
+            # Propagate extra args through tree.
+            if args.has_key( 'urlparam' ):
+                param = args['urlparam']
+                param = "%s&" % param
             else:
-                output('<A NAME="%s" HREF="%s?tree-e=%s#%s">'
+                param = ""
+
+            if exp:
+                ptreeData['tree-item-expanded']=1
+                output('<A NAME="%s" HREF="%s?%stree-c=%s#%s">'
+                       '<IMG SRC="%s/p_/mi" ALT="-" BORDER=0></A>' %
+                       (id, root_url, param, s, id, script))
+            else:
+                output('<A NAME="%s" HREF="%s?%stree-e=%s#%s">'
                        '<IMG SRC="%s/p_/pl" ALT="+" BORDER=0></A>' %
-                       (id, root_url, s, id, script))
+                       (id, root_url, param, s, id, script))
             output('</TD>\n')
 
         else:
@@ -413,8 +373,8 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
 
         level=level+1
         dataspan=colspan-level
-        if level > 3:   h=_td_colspan % (level-1)
-        elif level > 1: h=_td_single  * (level-1)
+        if level > 2:   h=_td_colspan % level
+        elif level > 0: h=_td_single  * level
         else: h=''
 
         if have_arg('header'):
@@ -441,7 +401,7 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
                 else: doc=None
                 if doc is not None:
                     treeData['-tree-substate-']=sub
-                    treeData['tree-level']=level
+                    ptreeData['tree-level']=level
                     md._push(treeData)
                     try: output(doc(
                         None,md,
@@ -460,7 +420,7 @@ def tpRenderTABLE(self, id, root_url, url, state, substate, diff, data,
             else: doc=None
             if doc is not None:
                 treeData['-tree-substate-']=sub
-                treeData['tree-level']=level
+                ptreeData['tree-level']=level
                 md._push(treeData)
                 try: output(doc(
                     None,md,
@@ -528,6 +488,8 @@ def apply_diff(state, diff, expand):
         del diff[-1]
         if len(s)==1: s.append([])
         s=s[1]
+        if type(s)==type(()):
+            s=list(s)
         loc=-1
         for i in range(len(s)):
             if s[i][0]==id:
@@ -623,20 +585,22 @@ def tpStateLevel(state, level=0):
         else: level=max(level,1)
     return level
 
-def tpValuesIds(self, branches, args,
+def tpValuesIds(self, get_items, args,
                 simple_type={type(''):0, type(1):0, type(1.0):0}.has_key,
                 ):
+    # get_item(node) is a function that returns the subitems of node
+
     # This should build the ids of subitems which are
     # expandable (non-empty). Leaves should never be
     # in the state - it will screw the colspan counting.
     r=[]
     idattr=args['id']
     try:
-        try: items=getattr(self, branches)()
+        try: items=get_items(self)
         except AttributeError: items=()
         for item in items:
             try:
-                if getattr(item, branches)():
+                if get_items(item):
 
                     if hasattr(item, idattr):
                         id=getattr(item, idattr)
@@ -644,7 +608,7 @@ def tpValuesIds(self, branches, args,
                     elif hasattr(item, '_p_oid'): id=oid(item)
                     else: id=pyid(item)
 
-                    e=tpValuesIds(item, branches, args)
+                    e=tpValuesIds(item, get_items, args)
                     if e: id=[id,e]
                     else: id=[id]
                     r.append(id)
