@@ -1,107 +1,48 @@
 ##############################################################################
+#
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py,v 1.28 1999/10/07 22:41:01 jim Exp $"""
-__version__='$Revision: 1.28 $'[11:-2]
+$Id: Connection.py,v 1.61.16.2 2002/02/20 15:05:06 fdrake Exp $"""
+__version__='$Revision: 1.61.16.2 $'[11:-2]
 
 from cPickleCache import PickleCache
-from POSException import ConflictError, ExportError
-from cStringIO import StringIO
-from cPickle import Unpickler, Pickler
+from POSException import ConflictError, ReadConflictError
 from ExtensionClass import Base
-from time import time
-import Transaction, string, ExportImport, sys, traceback, TmpStore
-from zLOG import LOG, ERROR
+import ExportImport, TmpStore
+from zLOG import LOG, ERROR, BLATHER
 from coptimizations import new_persistent_id
+from ConflictResolution import ResolvedSerial
+
+from cPickle import Unpickler, Pickler
+from cStringIO import StringIO
+import sys
+from time import time
+from types import StringType, ClassType
+
+global_code_timestamp = 0
+
+def updateCodeTimestamp():
+    '''
+    Called after changes are made to persistence-based classes.
+    Causes all connection caches to be re-created as the
+    connections are reopened.
+    '''
+    global global_code_timestamp
+    global_code_timestamp = time()
 
 ExtensionKlass=Base.__class__
 
-class HelperClass: pass
-ClassType=type(HelperClass)
-
 class Connection(ExportImport.ExportImport):
     """Object managers for individual object space.
 
@@ -114,6 +55,10 @@ class Connection(ExportImport.ExportImport):
     _tmp=None
     _debug_info=()
     _opened=None
+    _code_timestamp = 0
+
+    # Experimental. Other connections can register to be closed
+    # when we close by putting something here.
 
     def __init__(self, version='', cache_size=400,
                  cache_deactivate_after=60):
@@ -124,6 +69,7 @@ class Connection(ExportImport.ExportImport):
         self._invalidated=d={}
         self._invalid=d.has_key
         self._committed=[]
+        self._code_timestamp = global_code_timestamp
 
     def _breakcr(self):
         try: del self._cache
@@ -134,17 +80,22 @@ class Connection(ExportImport.ExportImport):
         except: pass
 
     def __getitem__(self, oid,
-                    tt=type(()), ct=type(HelperClass)):
+                    tt=type(())):
         cache=self._cache
         if cache.has_key(oid): return cache[oid]
 
-        __traceback_info__=oid
+        __traceback_info__ = (oid)
         p, serial = self._storage.load(oid, self._version)
+        __traceback_info__ = (oid, p)
         file=StringIO(p)
         unpickler=Unpickler(file)
         unpickler.persistent_load=self._persistent_load
 
-        object = unpickler.load()
+        try:
+            object = unpickler.load()
+        except:
+            raise "Could not load oid %s, pickled data in traceback info may\
+            contain clues" % (oid)
 
         klass, args = object
 
@@ -170,8 +121,7 @@ class Connection(ExportImport.ExportImport):
         return object
 
     def _persistent_load(self,oid,
-                        d={'__builtins__':{}},
-                        tt=type(()), st=type(''), ct=type(HelperClass)):
+                        tt=type(())):
 
         __traceback_info__=oid
 
@@ -208,42 +158,98 @@ class Connection(ExportImport.ExportImport):
         """Begin a new transaction.
 
         Any objects modified since the last transaction are invalidated.
-        """     
+        """
         self._db=odb
         self._storage=s=odb._storage
         self.new_oid=s.new_oid
-        self._cache.invalidate(self._invalidated)
+        if self._code_timestamp != global_code_timestamp:
+            # New code is in place.  Start a new cache.
+            self._resetCache()
+        else:
+            self._cache.invalidate(self._invalidated)
         self._opened=time()
 
         return self
 
+    def _resetCache(self):
+        '''
+        Creates a new cache, discarding the old.
+        '''
+        self._code_timestamp = global_code_timestamp
+        self._invalidated.clear()
+        orig_cache = self._cache
+        self._cache = PickleCache(self, orig_cache.cache_size,
+                                  orig_cache.cache_age)
+
     def abort(self, object, transaction):
         """Abort the object in the transaction.
 
         This just deactivates the thing.
         """
-        self._cache.invalidate(object._p_oid)
+        if object is self:
+            self._cache.invalidate(self._invalidated)
+        else:
+            self._cache.invalidate(object._p_oid)
 
     def cacheFullSweep(self, dt=0): self._cache.full_sweep(dt)
     def cacheMinimize(self, dt=0): self._cache.minimize(dt)
 
+    __onCloseCallbacks = None
+    
+    def onCloseCallback(self, f):
+        if self.__onCloseCallbacks is None:
+            self.__onCloseCallbacks = []
+        self.__onCloseCallbacks.append(f)
+
     def close(self):
         self._incrgc() # This is a good time to do some GC
         db=self._db
+
+        # Call the close callbacks.
+        if self.__onCloseCallbacks is not None:
+            for f in self.__onCloseCallbacks:
+                try: f()
+                except:
+                    f=getattr(f, 'im_self', f)
+                    LOG('ZODB',ERROR, 'Close callback failed for %s' % f,
+                        error=sys.exc_info())
+            self.__onCloseCallbacks = None
         self._db=self._storage=self._tmp=self.new_oid=self._opened=None
         self._debug_info=()
+        # Return the connection to the pool.
         db._closeConnection(self)
+                        
+    __onCommitActions = None
+    
+    def onCommitAction(self, method_name, *args, **kw):
+        if self.__onCommitActions is None:
+            self.__onCommitActions = []
+        self.__onCommitActions.append((method_name, args, kw))
+        get_transaction().register(self)
 
     def commit(self, object, transaction):
+        if object is self:
+            # We registered ourself.  Execute a commit action, if any.
+            if self.__onCommitActions is not None:
+                method_name, args, kw = self.__onCommitActions.pop(0)
+                apply(getattr(self, method_name), (transaction,) + args, kw)
+            return
         oid=object._p_oid
         invalid=self._invalid
         if oid is None or object._p_jar is not self:
+            # new object
             oid = self.new_oid()
             object._p_jar=self
             object._p_oid=oid
+            self._creating.append(oid)
 
         elif object._p_changed:
-            if invalid(oid) or invalid(None): raise ConflictError, oid
+            if (
+                (invalid(oid) and not hasattr(object, '_p_resolveConflict'))
+                or
+                invalid(None)
+                ):
+                raise ConflictError(object=object)
             self._invalidating.append(oid)
 
         else:
@@ -287,9 +293,11 @@ class Connection(ExportImport.ExportImport):
         dbstore=self._storage.store
         file=file.getvalue
         cache=self._cache
+        get=cache.get
         dump=pickler.dump
         clear_memo=pickler.clear_memo
 
+
         version=self._version
         
         while stack:
@@ -297,7 +305,20 @@ class Connection(ExportImport.ExportImport):
             del stack[-1]
             oid=object._p_oid
             serial=getattr(object, '_p_serial', '\0\0\0\0\0\0\0\0')
-            if invalid(oid): raise ConflictError, oid
+            if serial == '\0\0\0\0\0\0\0\0':
+                # new object
+                self._creating.append(oid)
+            else:
+                #XXX We should never get here
+                if (
+                    (invalid(oid) and
+                     not hasattr(object, '_p_resolveConflict'))
+                    or
+                    invalid(None)
+                    ):
+                    raise ConflictError(object=object)
+                self._invalidating.append(oid)
+                
             klass = object.__class__
         
             if klass is ExtensionKlass:
@@ -324,18 +345,29 @@ class Connection(ExportImport.ExportImport):
             dump((klass,args))
             dump(state)
             p=file(1)
-            object._p_serial=dbstore(oid,serial,p,version,transaction)
-            object._p_changed=0
+            s=dbstore(oid,serial,p,version,transaction)
+            # Put the object in the cache before handling the
+            # response, just in case the response contains the
+            # serial number for a newly created object
             try: cache[oid]=object
             except:
                 # Dang, I bet its wrapped:
                 if hasattr(object, 'aq_base'):
                     cache[oid]=object.aq_base
+                else:
+                    raise
+
+            self._handle_serial(s, oid)
 
     def commit_sub(self, t):
+        """Commit all work done in subtransactions"""
         tmp=self._tmp
         if tmp is None: return
         src=self._storage
+
+        LOG('ZODB', BLATHER,
+            'Commiting subtransaction of size %s' % src.getSize())
+        
         self._storage=tmp
         self._tmp=None
 
@@ -346,22 +378,47 @@ class Connection(ExportImport.ExportImport):
         dest=self._version
         get=self._cache.get
         oids=src._index.keys()
+
+        # Copy invalidating and creating info from temporary storage:
         invalidating=self._invalidating
         invalidating[len(invalidating):]=oids
+        creating=self._creating
+        creating[len(creating):]=src._creating
         
         for oid in oids:
             data, serial = load(oid, src)
             s=store(oid, serial, data, dest, t)
-            o=get(oid, None)
-            if o is not None: o._p_serial=s
+            self._handle_serial(s, oid, change=0)
 
     def abort_sub(self, t):
+        """Abort work done in subtransactions"""
         tmp=self._tmp
         if tmp is None: return
         src=self._storage
         self._tmp=None
         self._storage=tmp
+
+        
         self._cache.invalidate(src._index.keys())
+        self._invalidate_creating(src._creating)
+
+    def _invalidate_creating(self, creating=None):
+        """Dissown any objects newly saved in an uncommitted transaction.
+        """
+        if creating is None:
+            creating=self._creating
+            self._creating=[]
+
+        cache=self._cache
+        cache_get=cache.get
+        for oid in creating:
+            o=cache_get(oid, None)
+            if o is not None:
+                del o._p_jar
+                del o._p_oid
+                del cache[oid]
+
+    #XXX
 
     def db(self): return self._db
 
@@ -383,12 +440,31 @@ class Connection(ExportImport.ExportImport):
 
     def root(self): return self['\0\0\0\0\0\0\0\0']
 
-    def setstate(self,object):
+    def setstate(self, object):
         try:
             oid=object._p_oid
-            invalid=self._invalid
-            if invalid(oid) or invalid(None): raise ConflictError, oid
+             
             p, serial = self._storage.load(oid, self._version)
+
+            # XXX this is quite conservative!
+            # We need, however, to avoid reading data from a transaction
+            # that committed after the current "session" started, as
+            # that might lead to mixing of cached data from earlier
+            # transactions and new inconsistent data.
+            #
+            # Note that we (carefully) wait until after we call the
+            # storage to make sure that we don't miss an invaildation
+            # notifications between the time we check and the time we
+            # read.
+            invalid=self._invalid
+            if invalid(oid) or invalid(None):
+                if not hasattr(object.__class__, '_p_independent'):
+                    get_transaction().register(self)
+                    raise ReadConflictError(object=object)
+                invalid=1
+            else:
+                invalid=0
+
             file=StringIO(p)
             unpickler=Unpickler(file)
             unpickler.persistent_load=self._persistent_load
@@ -403,14 +479,31 @@ class Connection(ExportImport.ExportImport):
 
             object._p_serial=serial
 
+            if invalid:
+                if object._p_independent():
+                    try: del self._invalidated[oid]
+                    except KeyError: pass
+                else:
+                    get_transaction().register(self)
+                    raise ConflictError(object=object)
+
+        except ConflictError:
+            raise
         except:
-            t, v =sys.exc_info()[:2]
             LOG('ZODB',ERROR, "Couldn't load state for %s" % `oid`,
                 error=sys.exc_info())
             raise
 
-    def setklassstate(self, object,
-                      tt=type(()), ct=type(HelperClass)):
+    def oldstate(self, object, serial):
+        oid=object._p_oid
+        p = self._storage.loadSerial(oid, serial)
+        file=StringIO(p)
+        unpickler=Unpickler(file)
+        unpickler.persistent_load=self._persistent_load
+        unpickler.load()
+        return  unpickler.load()
+
+    def setklassstate(self, object):
         try:
             oid=object._p_oid
             __traceback_info__=oid
@@ -438,19 +531,23 @@ class Connection(ExportImport.ExportImport):
             object._p_changed=0
             object._p_serial=serial
         except:
-            LOG('ZODB',ERROR, 'setklassstate failed', error=sys.exc_info)
+            LOG('ZODB',ERROR, 'setklassstate failed', error=sys.exc_info())
             raise
 
     def tpc_abort(self, transaction):
+        if self.__onCommitActions is not None:
+            del self.__onCommitActions
         self._storage.tpc_abort(transaction)
         cache=self._cache
         cache.invalidate(self._invalidated)
         cache.invalidate(self._invalidating)
+        self._invalidate_creating()
 
     def tpc_begin(self, transaction, sub=None):
         if self._invalid(None): # Some nitwit invalidated everything!
-            raise ConflictError, oid 
+            raise ConflictError("transaction already invalidated")
         self._invalidating=[]
+        self._creating=[]
 
         if sub:
             # Sub-transaction!
@@ -463,17 +560,91 @@ class Connection(ExportImport.ExportImport):
 
         self._storage.tpc_begin(transaction)
 
+    def tpc_vote(self, transaction):
+        if self.__onCommitActions is not None:
+            del self.__onCommitActions
+        try:
+            vote=self._storage.tpc_vote
+        except AttributeError:
+            return
+        s = vote(transaction)
+        self._handle_serial(s)
+
+    def _handle_serial(self, store_return, oid=None, change=1):
+        """Handle the returns from store() and tpc_vote() calls."""
+
+        # These calls can return different types depending on whether
+        # ZEO is used.  ZEO uses asynchronous returns that may be
+        # returned in batches by the ClientStorage.  ZEO1 can also
+        # return an exception object and expect that the Connection
+        # will raise the exception.
+
+        # When commit_sub() exceutes a store, there is no need to
+        # update the _p_changed flag, because the subtransaction
+        # tpc_vote() calls already did this.  The change=1 argument
+        # exists to allow commit_sub() to avoid setting the flag
+        # again. 
+        if not store_return:
+            return
+        if isinstance(store_return, StringType):
+            assert oid is not None
+            serial = store_return
+            obj = self._cache.get(oid, None)
+            if obj is None:
+                return
+            if serial == ResolvedSerial:
+                obj._p_changed = None
+            else:
+                if change:
+                    obj._p_changed = 0
+                obj._p_serial = serial
+        else:
+            for oid, serial in store_return:
+                if not isinstance(serial, StringType):
+                    raise serial
+                obj = self._cache.get(oid, None)
+                if obj is None:
+                    continue
+                if serial == ResolvedSerial:
+                    obj._p_changed = None
+                else:
+                    if change:
+                        obj._p_changed = 0
+                    obj._p_serial = serial
+
+
     def tpc_finish(self, transaction):
-        self._storage.tpc_finish(transaction, self.tpc_finish_)
+
+        # It's important that the storage call the function we pass
+        # (self._invalidate_invalidating) while it still has it's
+        # lock.  We don't want another thread to be able to read any
+        # updated data until we've had a chance to send an
+        # invalidation message to all of the other connections!
+
+        if self._tmp is not None:
+            # Commiting a subtransaction!
+            # There is no need to invalidate anything.
+            self._storage.tpc_finish(transaction)
+            self._storage._creating[:0]=self._creating
+            del self._creating[:]
+        else:
+            self._db.begin_invalidation()
+            self._storage.tpc_finish(transaction,
+                                     self._invalidate_invalidating)
+
         self._cache.invalidate(self._invalidated)
         self._incrgc() # This is a good time to do some GC
 
-    def tpc_finish_(self):
+    def _invalidate_invalidating(self):
         invalidate=self._db.invalidate
-        for oid in self._invalidating: invalidate(oid, self)
+        for oid in self._invalidating:
+            invalidate(oid, self)
+        self._db.finish_invalidation()
 
     def sync(self):
         get_transaction().abort()
+        sync=getattr(self._storage, 'sync', 0)
+        if sync != 0: sync()
         self._cache.invalidate(self._invalidated)
         self._incrgc() # This is a good time to do some GC
 
