@@ -1,108 +1,39 @@
 ##############################################################################
+#
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 """Database objects
 
-$Id: DB.py,v 1.18.4.1 2000/01/11 17:42:05 jim Exp $"""
-__version__='$Revision: 1.18.4.1 $'[11:-2]
+$Id: DB.py,v 1.38.2.1 2002/02/13 16:37:04 gvanrossum Exp $"""
+__version__='$Revision: 1.38.2.1 $'[11:-2]
 
-import cPickle, cStringIO, sys, POSException
+import cPickle, cStringIO, sys, POSException, UndoLogCompatible
 from Connection import Connection
 from bpthread import allocate_lock
 from Transaction import Transaction
 from referencesf import referencesf
 from time import time, ctime
+from zLOG import LOG, ERROR
 
-StringType=type('')
+from types import StringType
 
-class DB:
+class DB(UndoLogCompatible.UndoLogCompatible):
     """The Object Database
 
     The Object database coordinates access to and interaction of one
     or more connections, which manage object spaces.  Most of the actual work
     of managing objects is done by the connections.
     """
+    klass = Connection
 
     def __init__(self, storage,
                  pool_size=7,
@@ -120,29 +51,15 @@ class DB:
         pool_size -- The size of the pool of object spaces.
 
         """
-        self._storage=storage
-        storage.registerDB(self, None)
-        try: storage.load('\0\0\0\0\0\0\0\0','')
-        except:
-            import PersistentMapping
-            file=cStringIO.StringIO()
-            p=cPickle.Pickler(file,1)
-            p.dump((PersistentMapping.PersistentMapping,None))
-            p.dump({'_container': {}})
-            t=Transaction()
-            t.description='initial database creation'
-            storage.tpc_begin(t)
-            storage.store('\0\0\0\0\0\0\0\0', None, file.getvalue(), '', t)
-            storage.tpc_finish(t)
 
         # Allocate locks:
         l=allocate_lock()
         self._a=l.acquire
         self._r=l.release
 
+        # Setup connection pools and cache info
         self._pools={},[]
         self._temps=[]
-
         self._pool_size=pool_size
         self._cache_size=cache_size
         self._cache_deactivate_after=cache_deactivate_after
@@ -152,17 +69,44 @@ class DB:
 
         self._miv_cache={}
 
+        # Setup storage
+        self._storage=storage
+        storage.registerDB(self, None)
+        if not hasattr(storage,'tpc_vote'): storage.tpc_vote=lambda *args: None
+        try:
+            storage.load('\0\0\0\0\0\0\0\0','')
+        except KeyError:
+            # Create the database's root in the storage if it doesn't exist
+            import PersistentMapping
+            root = PersistentMapping.PersistentMapping()
+            # Manually create a pickle for the root to put in the storage.
+            # The pickle must be in the special ZODB format.
+            file = cStringIO.StringIO()
+            p = cPickle.Pickler(file, 1)
+            p.dump((root.__class__, None))
+            p.dump(root.__getstate__())
+            t = Transaction()
+            t.description = 'initial database creation'
+            storage.tpc_begin(t)
+            storage.store('\0\0\0\0\0\0\0\0', None, file.getvalue(), '', t)
+            storage.tpc_vote(t)
+            storage.tpc_finish(t)
+
         # Pass through methods:
         for m in ('history',
                   'supportsUndo', 'supportsVersions', 'undoLog',
                   'versionEmpty', 'versions'):
             setattr(self, m, getattr(storage, m))
-        
+
+        if hasattr(storage, 'undoInfo'):
+            self.undoInfo=storage.undoInfo
+            
 
     def _cacheMean(self, attr):
+        # XXX this method doesn't work
         m=[0,0]
-        def f(con, m=m):
-            t=getattr(con._cache,attr)
+        def f(con, m=m, attr=attr):
+            t=getattr(con._cache, attr)
             m[0]=m[0]+t
             m[1]=m[1]+1
 
@@ -216,7 +160,9 @@ class DB:
         detail={}
         def f(con,detail=detail,have_detail=detail.has_key):
             for oid, ob in con._cache.items():
-                c="%s.%s" % (ob.__class__.__module__, ob.__class__.__name__)
+                module = getattr(ob.__class__, '__module__', '')
+                module = module and '%s.' % module or ''
+                c="%s%s" % (module, ob.__class__.__name__)
                 if have_detail(c): detail[c]=detail[c]+1
                 else: detail[c]=1
         
@@ -227,22 +173,30 @@ class DB:
 
     def cacheExtremeDetail(self):
         detail=[]
-        def f(con, detail=detail, rc=sys.getrefcount):
+        conn_no = [0]  # A mutable reference to a counter
+        def f(con, detail=detail, rc=sys.getrefcount, conn_no=conn_no):
+            conn_no[0] = conn_no[0] + 1
+            cn = conn_no[0]
             for oid, ob in con._cache.items():
-                id=oid
+                id=''
                 if hasattr(ob,'__dict__'):
                     d=ob.__dict__
                     if d.has_key('id'):
-                        id="%s (%s)" % (oid, d['id'])
+                        id=d['id']
                     elif d.has_key('__name__'):
-                        id="%s (%s)" % (oid, d['__name__'])
+                        id=d['__name__']
+
+                module = getattr(ob.__class__, '__module__', '')
+                module = module and '%s.' % module or ''
     
                 detail.append({
-                    'oid': id,
-                    'klass': "%s.%s" % (ob.__class__.__module__,
-                                        ob.__class__.__name__),
+                    'conn_no': cn,
+                    'oid': oid,
+                    'id': id,
+                    'klass': "%s%s" % (module, ob.__class__.__name__),
                     'rc': rc(ob)-4,
-                    'references': con.references(oid),
+                    'state': ob._p_changed,
+                    #'references': con.references(oid),
                     })
 
         self._connectionMap(f)
@@ -301,6 +255,16 @@ class DB:
     def importFile(self, file):
         raise 'Not yet implemented'
 
+    def begin_invalidation(self):
+        # Must be called before first call to invalidate and before
+        # the storage lock is held.
+        self._a()
+
+    def finish_invalidation(self):
+        # Must be called after begin_invalidation() and after final
+        # invalidate() call.
+        self._r()
+
     def invalidate(self, oid, connection=None, version='',
                    rc=sys.getrefcount):
         """Invalidate references to a given oid.
@@ -310,42 +274,33 @@ class DB:
         passed in to prevent useless (but harmless) messages to the
         connection.
         """
-        if connection is not None: version=connection._version
-        self._a()
-        try:
+        if connection is not None:
+            version=connection._version
+        # Update modified in version cache
+        h=hash(oid)%131
+        o=self._miv_cache.get(h, None)
+        if o is not None and o[0]==oid: del self._miv_cache[h]
 
-            # Update modified in version cache
-            h=hash(oid)%131
-            cache=self._miv_cache
-            o=cache.get(h, None)
-            if o and o[0]==oid: del cache[h]
+        # Notify connections
+        for pool, allocated in self._pools[1]:
+            for cc in allocated:
+                if (cc is not connection and
+                    (not version or cc._version==version)):
+                    if rc(cc) <= 3:
+                        cc.close()
+                    cc.invalidate(oid)
 
-            # Notify connections
-            pools,pooll=self._pools
-            for pool, allocated in pooll:
-                for cc in allocated:
+        temps=self._temps
+        if temps:
+            t=[]
+            for cc in temps:
+                if rc(cc) > 3:
                     if (cc is not connection and
                         (not version or cc._version==version)):
-                        if rc(cc) <= 3:
-                            cc.close()
                         cc.invalidate(oid)
-
-            temps=self._temps
-            if temps:
-                t=[]
-                for cc in temps:
-                    if rc(cc) > 3:
-                        if cc is not connection and cc._version==version:
-                            cc.invalidate(oid)
-                        t.append(cc)
-                    else: cc.close()
-                self._temps=t
-        finally: self._r()
-
-    def invalidateMany(self, oids=None, version=''):
-        if oids is None: self.invalidate(None, version=version)
-        else:
-            for oid in oids: self.invalidate(oid, version=version)
+                    t.append(cc)
+                else: cc.close()
+            self._temps=t
 
     def modifiedInVersion(self, oid):
         h=hash(oid)%131
@@ -394,7 +349,7 @@ class DB:
                 # This is a temporary connection.
                 # We won't bother with the pools.  This will be
                 # a one-use connection.
-                c=Connection(
+                c=self.klass(
                     version=version,
                     cache_size=self._version_cache_size,
                     cache_deactivate_after=
@@ -425,7 +380,7 @@ class DB:
 
             # Pool locks are tricky.  Basically, the lock needs to be
             # set whenever the pool becomes empty so that threads are
-            # forced to wait until the pool gets a connection it it.
+            # forced to wait until the pool gets a connection in it.
             # The lock is acquired when the (empty) pool is
             # created. The The lock is acquired just prior to removing
             # the last connection from the pool and just after adding
@@ -445,7 +400,7 @@ class DB:
                 c=None
                 if version:
                     if self._version_pool_size > len(allocated) or force:
-                        c=Connection(
+                        c=self.klass(
                             version=version,
                             cache_size=self._version_cache_size,
                             cache_deactivate_after=
@@ -453,7 +408,7 @@ class DB:
                         allocated.append(c)
                         pool.append(c)
                 elif self._pool_size > len(allocated) or force:
-                    c=Connection(
+                    c=self.klass(
                         version=version,
                         cache_size=self._cache_size,
                         cache_deactivate_after=
@@ -518,8 +473,13 @@ class DB:
                     })
         return r
         
-    def pack(self, t):
-        self._storage.pack(t,referencesf)
+    def pack(self, t=None, days=0):
+        if t is None: t=time()
+        t=t-(days*86400)
+        try: self._storage.pack(t,referencesf)
+        except:
+            LOG("ZODB", ERROR, "packing", error=sys.exc_info())
+            raise
                            
     def setCacheDeactivateAfter(self, v):
         self._cache_deactivate_after=v
@@ -555,8 +515,20 @@ class DB:
     def cacheStatistics(self): return () # :(
 
     def undo(self, id):
-        for oid in self._storage.undo(id):
-            self.invalidate(oid)
+        storage=self._storage
+        try: supportsTransactionalUndo = storage.supportsTransactionalUndo
+        except AttributeError:
+            supportsTransactionalUndo=0
+        else:
+            supportsTransactionalUndo=supportsTransactionalUndo()
+
+        if supportsTransactionalUndo:
+            # new style undo
+            TransactionalUndo(self, id)
+        else:
+            # fall back to old undo
+            for oid in storage.undo(id):
+                self.invalidate(oid)
 
     def versionEmpty(self, version):
         return self._storage.versionEmpty(version)
@@ -573,6 +545,7 @@ class CommitVersion:
         self._dest=dest
         self.tpc_abort=s.tpc_abort
         self.tpc_begin=s.tpc_begin
+        self.tpc_vote=s.tpc_vote
         self.tpc_finish=s.tpc_finish
         get_transaction().register(self)
 
@@ -581,8 +554,12 @@ class CommitVersion:
     def commit(self, reallyme, t):
         db=self._db
         dest=self._dest
-        for oid in db._storage.commitVersion(self._version, dest, t):
-            db.invalidate(oid, version=dest)
+        oids=db._storage.commitVersion(self._version, dest, t)
+        for oid in oids: db.invalidate(oid, version=dest)
+        if dest:
+            # the code above just invalidated the dest version.
+            # now we need to invalidate the source!
+            for oid in oids: db.invalidate(oid, version=self._version)
     
 class AbortVersion(CommitVersion):
     """An object that will see to version abortion
@@ -593,5 +570,23 @@ class AbortVersion(CommitVersion):
     def commit(self, reallyme, t):
         db=self._db
         version=self._version
-        for oid in db._storage.abortVersion(version, t):
+        oids = db._storage.abortVersion(version, t)
+        for oid in oids:
             db.invalidate(oid, version=version)
+
+
+class TransactionalUndo(CommitVersion):
+    """An object that will see to transactional undo
+
+    in cooperation with a transaction manager.
+    """
+    
+    # I'm lazy. I'm reusing __init__ and abort and reusing the
+    # version attr for the transavtion id. There's such a strong
+    # similarity of rythm, that I think it's justified.
+
+    def commit(self, reallyme, t):
+        db=self._db
+        oids=db._storage.transactionalUndo(self._version, t)
+        for oid in oids:
+            db.invalidate(oid)
