@@ -1,85 +1,14 @@
 ##############################################################################
+#
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 """Demo ZODB storage
@@ -100,7 +29,7 @@ There are three main data structures:
   _data -- Transaction logging information necessary for undo
 
       This is a mapping from transaction id to transaction, where
-      a transaction is simply a 4-tuple:
+      a transaction is simply a 5-tuple:
 
         packed, user, description, extension_data, records
 
@@ -137,7 +66,7 @@ The pickle data will be None for a record for an object created in
 an aborted version.
 
 It is instructive to watch what happens to the internal data structures
-as changes are made.  Foe example, in Zope, you can create an external
+as changes are made.  For example, in Zope, you can create an external
 method::
 
   import Zope
@@ -147,14 +76,16 @@ method::
 
       return Zope.DB._storage._splat()
 
-and call it to minotor the storage.
+and call it to monitor the storage.
 
 """
-__version__='$Revision: 1.4 $'[11:-2]
+__version__='$Revision: 1.8.16.2 $'[11:-2]
 
-import base64, POSException, BTree, BaseStorage, time, string, utils
+import base64, time, string
+from ZODB import POSException, BaseStorage, utils
 from TimeStamp import TimeStamp
 from cPickle import loads
+from BTrees import OOBTree
 
 class DemoStorage(BaseStorage.BaseStorage):
 
@@ -163,7 +94,7 @@ class DemoStorage(BaseStorage.BaseStorage):
         BaseStorage.BaseStorage.__init__(self, name, base)
 
         # We use a BTree because the items are sorted!
-        self._data=BTree.BTree()
+        self._data=OOBTree.OOBTree()
         self._index={}
         self._vindex={}
         self._base=base
@@ -198,6 +129,8 @@ class DemoStorage(BaseStorage.BaseStorage):
     def abortVersion(self, src, transaction):
         if transaction is not self._transaction:
             raise POSException.StorageTransactionError(self, transaction)
+        if not src:
+            raise POSException.VersionCommitError("Invalid version")
         
         self._lock_acquire()
         try:
@@ -224,6 +157,12 @@ class DemoStorage(BaseStorage.BaseStorage):
         if transaction is not self._transaction:
             raise POSException.StorageTransactionError(self, transaction)
         
+        if not src:
+            raise POSException.VersionCommitError("Invalid source version")
+        if src == dest:
+            raise POSException.VersionCommitError(
+                "Can't commit to same version: %s" % repr(src))
+
         self._lock_acquire()
         try:
             v=self._vindex.get(src, None)
@@ -233,8 +172,14 @@ class DemoStorage(BaseStorage.BaseStorage):
             oids=[]
             for r in v.values():
                 oid, serial, pre, vdata, p = r
+                assert vdata is not None
                 oids.append(oid)
-                tindex.append([oid, serial, r, None, p])
+                if dest:
+                    new_vdata = dest, vdata[1]
+                else:
+                    new_vdata = None
+                tindex.append([oid, serial, r, new_vdata, p])
+                
 
             return oids
 
@@ -243,18 +188,23 @@ class DemoStorage(BaseStorage.BaseStorage):
     def load(self, oid, version):
         self._lock_acquire()
         try:
-            try: oid, serial, pre, vdata, p = self._index[oid]
-            except:
-                if self._base: return self._base.load(oid, '')
+            try:
+                oid, serial, pre, vdata, p = self._index[oid]
+            except KeyError:
+                if self._base:
+                    return self._base.load(oid, '')
                 raise KeyError, oid
 
             if vdata:
                 oversion, nv = vdata
                 if oversion != version:
-                    if nv: oid, serial, pre, vdata, p = nv
-                    else: raise KeyError, oid
+                    if nv:
+                        oid, serial, pre, vdata, p = nv
+                    else:
+                        raise KeyError, oid
 
-            if p is None: raise KeyError, oid
+            if p is None:
+                raise KeyError, oid
             
             return p, serial
         finally: self._lock_release()
@@ -295,17 +245,18 @@ class DemoStorage(BaseStorage.BaseStorage):
                 else:
                     nv=old
 
-                if serial != oserial: raise POSException.ConflictError
+                if serial != oserial:
+                    raise POSException.ConflictError(serials=(oserial, serial))
                 
             serial=self._serial
-            r=[oid, serial, old, version and (version,nv) or None, data]
+            r=[oid, serial, old, version and (version, nv) or None, data]
             self._tindex.append(r)
 
             s=self._tsize
             s=s+72+(data and (16+len(data)) or 4)
             if version: s=s+32+len(version)
 
-            if s > self._quota:
+            if self._quota is not None and s > self._quota:
                 raise POSException.StorageError, (
                     '''<b>Quota Exceeded</b><br>
                     The maximum quota for this demonstration storage
@@ -364,7 +315,7 @@ class DemoStorage(BaseStorage.BaseStorage):
             vindex_get=self._vindex.get
             for r in t:
                 if index[r[0]] is not r:
-                    raise POSException.UndoError, 'Undoable transaction'
+                    raise POSException.UndoError, 'non-undoable transaction'
 
             oids=[]
             for r in t:
@@ -435,13 +386,13 @@ class DemoStorage(BaseStorage.BaseStorage):
         return not self._vindex.get(version, None)
 
     def versions(self, max=None):
-        r=[]
-        a=r.append
-        for version in self._vindex.keys()[:max]:
-            if self.versionEmpty(version): continue
-            a(version)
-            if max and len(r) >= max: return r
-
+        r = []
+        for v in self._vindex.keys():
+            if self.versionEmpty(v):
+                continue
+            r.append(v)
+            if max is not None and len(r) >= max:
+                break
         return r
 
     def _build_indexes(self, stop='\377\377\377\377\377\377\377\377'):
