@@ -1,88 +1,20 @@
 /*****************************************************************************
+
+  Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+  All Rights Reserved.
   
-  Zope Public License (ZPL) Version 1.0
-  -------------------------------------
-  
-  Copyright (c) Digital Creations.  All rights reserved.
-  
-  This license has been certified as Open Source(tm).
-  
-  Redistribution and use in source and binary forms, with or without
-  modification, are permitted provided that the following conditions are
-  met:
-  
-  1. Redistributions in source code must retain the above copyright
-     notice, this list of conditions, and the following disclaimer.
-  
-  2. Redistributions in binary form must reproduce the above copyright
-     notice, this list of conditions, and the following disclaimer in
-     the documentation and/or other materials provided with the
-     distribution.
-  
-  3. Digital Creations requests that attribution be given to Zope
-     in any manner possible. Zope includes a "Powered by Zope"
-     button that is installed by default. While it is not a license
-     violation to remove this button, it is requested that the
-     attribution remain. A significant investment has been put
-     into Zope, and this effort will continue if the Zope community
-     continues to grow. This is one way to assure that growth.
-  
-  4. All advertising materials and documentation mentioning
-     features derived from or use of this software must display
-     the following acknowledgement:
-  
-       "This product includes software developed by Digital Creations
-       for use in the Z Object Publishing Environment
-       (http://www.zope.org/)."
-  
-     In the event that the product being advertised includes an
-     intact Zope distribution (with copyright and license included)
-     then this clause is waived.
-  
-  5. Names associated with Zope or Digital Creations must not be used to
-     endorse or promote products derived from this software without
-     prior written permission from Digital Creations.
-  
-  6. Modified redistributions of any form whatsoever must retain
-     the following acknowledgment:
-  
-       "This product includes software developed by Digital Creations
-       for use in the Z Object Publishing Environment
-       (http://www.zope.org/)."
-  
-     Intact (re-)distributions of any official Zope release do not
-     require an external acknowledgement.
-  
-  7. Modifications are encouraged but must be packaged separately as
-     patches to official Zope releases.  Distributions that do not
-     clearly separate the patches from the original work must be clearly
-     labeled as unofficial distributions.  Modifications which do not
-     carry the name Zope may be packaged in any form, as long as they
-     conform to all of the clauses above.
-  
-  
-  Disclaimer
-  
-    THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-    EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-    PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-    USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-    SUCH DAMAGE.
-  
-  
-  This software consists of contributions made by Digital Creations and
-  many individuals on behalf of Digital Creations.  Specific
-  attributions are listed in the accompanying credits file.
+  This software is subject to the provisions of the Zope Public License,
+  Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+  WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+  WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+  FOR A PARTICULAR PURPOSE
   
  ****************************************************************************/
-static char *what_string = "$Id: cPersistence.c,v 1.35 1999/10/18 10:36:50 jim Exp $";
+static char cPersistence_doc_string[] = 
+"Defines Persistent mixin class for persistent objects.\n"
+"\n"
+"$Id: cPersistence.c,v 1.47.16.3 2002/03/08 19:13:02 jeremy Exp $\n";
 
 #include <string.h>
 #include "cPersistence.h"
@@ -94,6 +26,7 @@ static char *what_string = "$Id: cPersistence.c,v 1.35 1999/10/18 10:36:50 jim E
 
 static PyObject *py_keys, *py_setstate, *py___dict__, *py_timeTime;
 static PyObject *py__p_changed, *py__p_deactivate;
+static PyObject *py___getattr__, *py___setattr__, *py___delattr__;
 
 static PyObject *TimeStamp;
 
@@ -116,23 +49,42 @@ call_debug(char *event, cPersistentObject *self)
 }
 #endif
 
-static void
-init_strings()
+static int
+init_strings(void)
 {
-#define INIT_STRING(S) py_ ## S = PyString_FromString(#S)
+#define INIT_STRING(S) if (! (py_ ## S = PyString_FromString(#S))) return -1;
   INIT_STRING(keys);
   INIT_STRING(setstate);
   INIT_STRING(timeTime);
   INIT_STRING(__dict__);
   INIT_STRING(_p_changed);
   INIT_STRING(_p_deactivate);
+  INIT_STRING(__getattr__);
+  INIT_STRING(__setattr__);
+  INIT_STRING(__delattr__);
 #undef INIT_STRING
+  return 0;
+}
+
+static int
+checknoargs(PyObject *args)
+{
+  if (!PyTuple_Check(args))
+    return 0;
+  if (PyTuple_GET_SIZE(args) != 0) {
+    PyErr_Format(PyExc_TypeError, 
+		 "function takes exactly 0 arguments (%d given)",
+		 PyTuple_GET_SIZE(args));
+    return 0;
+  }
+  return 1;
 }
 
 static PyObject *
 callmethod(PyObject *self, PyObject *name)
 {
-  if(self=PyObject_GetAttr(self,name))
+  self=PyObject_GetAttr(self,name);
+  if(self)
     ASSIGN(self,PyObject_CallObject(self,NULL));
   return self;
 }
@@ -140,46 +92,18 @@ callmethod(PyObject *self, PyObject *name)
 static PyObject *
 callmethod1(PyObject *self, PyObject *name, PyObject *arg)
 {
-  if((self=PyObject_GetAttr(self,name)) && (name=PyTuple_New(1)))
+  self = PyObject_GetAttr(self, name);
+  UNLESS(self) return NULL;
+  name = PyTuple_New(1);
+  UNLESS(name) 
     {
-      PyTuple_SET_ITEM(name, 0, arg);
-      ASSIGN(self,PyObject_CallObject(self,name));
-      PyTuple_SET_ITEM(name, 0, NULL);
-      Py_DECREF(name);
-    }
-  return self;
-}
-
-static PyObject *
-callmethod2(PyObject *self, PyObject *name, PyObject *arg, PyObject *arg2)
-{
-  if((self=PyObject_GetAttr(self,name)) && (name=PyTuple_New(2)))
-    {
-      PyTuple_SET_ITEM(name, 0, arg);
-      PyTuple_SET_ITEM(name, 1, arg2);
-      ASSIGN(self,PyObject_CallObject(self,name));
-      PyTuple_SET_ITEM(name, 0, NULL);
-      PyTuple_SET_ITEM(name, 1, NULL);
-      Py_DECREF(name);
-    }
-  return self;
-}
-
-static PyObject *
-callmethod3(PyObject *self, PyObject *name,
-	    PyObject *arg, PyObject *arg2, PyObject *arg3)
-{
-  if((self=PyObject_GetAttr(self,name)) && (name=PyTuple_New(3)))
-    {
-      PyTuple_SET_ITEM(name, 0, arg);
-      PyTuple_SET_ITEM(name, 1, arg2);
-      PyTuple_SET_ITEM(name, 2, arg3);
-      ASSIGN(self,PyObject_CallObject(self,name));
-      PyTuple_SET_ITEM(name, 0, NULL);
-      PyTuple_SET_ITEM(name, 1, NULL);
-      PyTuple_SET_ITEM(name, 2, NULL);
-      Py_DECREF(name);
+      Py_DECREF(self);
+      return NULL;
     }
+  PyTuple_SET_ITEM(name, 0, arg);
+  ASSIGN(self, PyObject_CallObject(self, name));
+  PyTuple_SET_ITEM(name, 0, NULL);
+  Py_DECREF(name);
   return self;
 }
 
@@ -199,43 +123,9 @@ if(self->state < 0 && self->jar)                                 \
 }
 
 
-static PyObject *
-#ifdef HAVE_STDARG_PROTOTYPES
-/* VARARGS 2 */
-PyString_BuildFormat(char *stringformat, char *format, ...)
-#else
-/* VARARGS */
-PyString_BuildFormat(va_alist) va_dcl
-#endif
-{
-  va_list va;
-  PyObject *args=0, *retval=0, *v=0;
-#ifdef HAVE_STDARG_PROTOTYPES
-  va_start(va, format);
-#else
-  PyObject *ErrType;
-  char *stringformat, *format;
-  va_start(va);
-  ErrType = va_arg(va, PyObject *);
-  stringformat   = va_arg(va, char *);
-  format   = va_arg(va, char *);
-#endif
-  
-  args = Py_VaBuildValue(format, va);
-  va_end(va);
-  if(! args) return NULL;
-  if(!(retval=PyString_FromString(stringformat))) return NULL;
-
-  v=PyString_Format(retval, args);
-  Py_DECREF(retval);
-  Py_DECREF(args);
-  return v;
-}
-
 /****************************************************************************/
 
 staticforward PyExtensionClass Pertype;
-staticforward PyExtensionClass TPertype;
 
 static int
 changed(cPersistentObject *self)
@@ -295,13 +185,14 @@ Per___changed__(cPersistentObject *self, PyObject *args)
 static PyObject *
 Per__p_deactivate(cPersistentObject *self, PyObject *args)
 {
-  PyObject *init=0, *dict;
+  PyObject *dict;
 
 #ifdef DEBUG_LOG
   if (idebug_log < 0) call_debug("reinit",self);
 #endif
 
-  if (args && ! PyArg_ParseTuple(args,"")) return NULL;
+  if (args && !checknoargs(args))
+    return NULL;
 
   if (self->state==cPersistent_UPTODATE_STATE && self->jar &&
       HasInstDict(self) && (dict=INSTANCE_DICT(self)))
@@ -327,13 +218,11 @@ Per_setstate(cPersistentObject *self)
 }
 
 static PyObject *
-Per__getstate__(self,args)
-     cPersistentObject *self;
-     PyObject *args;
+Per__getstate__(cPersistentObject *self, PyObject *args)
 {
   PyObject *__dict__, *d=0;
 
-  UNLESS(PyArg_ParseTuple(args, "")) return NULL;
+  if (!checknoargs(args)) return NULL;
 
 #ifdef DEBUG_LOG
   if(idebug_log < 0) call_debug("get",self);
@@ -349,12 +238,12 @@ Per__getstate__(self,args)
 	  
       for(pos=0; PyDict_Next(__dict__, &pos, &k, &v); )
 	{
-	  if(PyString_Check(k) && (ck=PyString_AsString(k)) &&
+	  if(PyString_Check(k) && (ck=PyString_AS_STRING(k)) &&
 	     (*ck=='_' && ck[1]=='v' && ck[2]=='_'))
 	    {
 	      UNLESS(d=PyDict_New()) goto err;
 	      for(pos=0; PyDict_Next(__dict__, &pos, &k, &v); )
-		UNLESS(PyString_Check(k) && (ck=PyString_AsString(k)) &&
+		UNLESS(PyString_Check(k) && (ck=PyString_AS_STRING(k)) &&
 		       (*ck=='_' && ck[1]=='v' && ck[2]=='_'))
 		  if(PyDict_SetItem(d,k,v) < 0) goto err;
 	      return d;
@@ -368,15 +257,12 @@ Per__getstate__(self,args)
   return __dict__;
 
 err:
-  Py_DECREF(__dict__);
   Py_XDECREF(d);
   return NULL;
 }  
 
 static PyObject *
-Per__setstate__(self,args)
-     cPersistentObject *self;
-     PyObject *args;
+Per__setstate__(cPersistentObject *self, PyObject *args)
 {
   PyObject *__dict__, *v, *keys=0, *key=0, *e=0;
   int l, i;
@@ -394,8 +280,8 @@ Per__setstate__(self,args)
 	   
 	   if(PyDict_Check(v))
 	     {
-	       for(i=0; PyDict_Next(v,&i,&key,&e);)
-		 if(PyObject_SetItem(__dict__,key,e) < 0)
+	       for(i=0; PyDict_Next(v, &i, &key, &e);)
+		 if(PyDict_SetItem(__dict__, key, e) < 0)
 		   return NULL;
 	     }
 	   else
@@ -407,7 +293,7 @@ Per__setstate__(self,args)
 		 {
 		   UNLESS_ASSIGN(key,PySequence_GetItem(keys,i)) goto err;
 		   UNLESS_ASSIGN(e,PyObject_GetItem(v,key)) goto err;
-		   UNLESS(-1 != PyObject_SetItem(__dict__,key,e)) goto err;
+		   UNLESS(-1 != PyDict_SetItem(__dict__,key,e)) goto err;
 		 }
 	       
 	       Py_XDECREF(key);
@@ -442,8 +328,7 @@ static struct PyMethodDef Per_methods[] = {
 /* ---------- */
 
 static void
-Per_dealloc(self)
-	cPersistentObject *self;
+Per_dealloc(cPersistentObject *self)
 {
 #ifdef DEBUG_LOG
   if(idebug_log < 0) call_debug("del",self);
@@ -451,7 +336,7 @@ Per_dealloc(self)
   Py_XDECREF(self->jar);
   Py_XDECREF(self->oid);
   Py_DECREF(self->ob_type);
-  PyMem_DEL(self);
+  PyObject_DEL(self);
 }
 
 static PyObject *
@@ -544,18 +429,28 @@ static PyObject*
 Per_getattro(cPersistentObject *self, PyObject *name)
 {
   char *s=NULL;
+  PyObject *r;
 
   if (PyString_Check(name))
-    UNLESS(s=PyString_AsString(name)) return NULL;
-  return Per_getattr(self, name, s, PyExtensionClassCAPI->getattro);
-}
+    UNLESS(s=PyString_AS_STRING(name)) return NULL;
 
-static int 
-bad_delattr()
-{
-  PyErr_SetString(PyExc_AttributeError,
-		  "delete undeletable attribute");
-  return -1;
+  r = Per_getattr(self, name, s, PyExtensionClassCAPI->getattro);
+  if (! r && self->state != cPersistent_GHOST_STATE &&
+      (((PyExtensionClass*)(self->ob_type))->class_flags 
+       & EXTENSIONCLASS_USERGETATTR_FLAG)
+      )
+    {
+      PyErr_Clear();
+
+      r=PyObject_GetAttr(OBJECT(self), py___getattr__);
+      if (r) 
+	{
+	  ASSIGN(r, PyObject_CallFunction(r, "O", name));
+	}
+      else PyErr_SetObject(PyExc_AttributeError, name);
+    }
+ 
+  return r;  
 }
 
 static int
@@ -565,7 +460,7 @@ _setattro(cPersistentObject *self, PyObject *oname, PyObject *v,
   char *name="";
 
   UNLESS(oname) return -1;
-  if(PyString_Check(oname)) UNLESS(name=PyString_AsString(oname)) return -1;
+  if(PyString_Check(oname)) UNLESS(name=PyString_AS_STRING(oname)) return -1;
 	
   if(*name=='_' && name[1]=='p' && name[2]=='_')
     {
@@ -612,8 +507,8 @@ _setattro(cPersistentObject *self, PyObject *oname, PyObject *v,
 	  if (v==Py_None)
 	    {
 	      v=PyObject_GetAttr(OBJECT(self), py__p_deactivate);
-	      if (v) ASSIGN(v, PyObject_CallObject(v, NULL));
-	      if (v) Py_DECREF(v);
+	      if (v) { ASSIGN(v, PyObject_CallObject(v, NULL)); }
+	      if (v) { Py_DECREF(v); }
 	      self->state=cPersistent_GHOST_STATE;
 	      return 0;
 	    }
@@ -629,18 +524,63 @@ _setattro(cPersistentObject *self, PyObject *oname, PyObject *v,
       /* Record access times */
       self->atime=((long)(time(NULL)/3))%65536;
 
-      if(! (*name=='_' && name[1]=='v' && name[2]=='_')
-	 && self->state != cPersistent_CHANGED_STATE && self->jar)
+      if((! (*name=='_' && name[1]=='v' && name[2]=='_'))
+	 && (self->state != cPersistent_CHANGED_STATE && self->jar)
+	 && setattrf
+	 )
 	if(changed(self) < 0) return -1;
     }
 
-  return setattrf((PyObject*)self,oname,v);
+  if (setattrf)
+    return setattrf((PyObject*)self,oname,v);
+  
+  return 1;			/* Ready for user setattr */
 }
 
 static int
 Per_setattro(cPersistentObject *self, PyObject *oname, PyObject *v)
 {
-  return _setattro(self,oname, v, PyExtensionClassCAPI->setattro);
+  int r;
+  PyObject *m;
+
+  if (v && (((PyExtensionClass*)self->ob_type)->class_flags 
+	    & EXTENSIONCLASS_USERSETATTR_FLAG)
+      )
+    {
+      r=_setattro(self,oname, v, NULL);
+      if (r < 1) return r;
+
+      m=PyObject_GetAttr(OBJECT(self), py___setattr__);
+      if (m) 
+	{
+	  ASSIGN(m, PyObject_CallFunction(m, "OO", oname, v));
+	}
+      else PyErr_SetObject(PyExc_AttributeError, oname);
+    }
+  else if (!v && (((PyExtensionClass*)self->ob_type)->class_flags 
+		  & EXTENSIONCLASS_USERDELATTR_FLAG)
+	   )
+    {
+      r=_setattro(self,oname, v, NULL);
+      if (r < 1) return r;
+
+      m=PyObject_GetAttr(OBJECT(self), py___delattr__);
+      if (m) 
+      {
+	ASSIGN(m, PyObject_CallFunction(m, "O", oname));
+      }
+      else PyErr_SetObject(PyExc_AttributeError, oname);
+    }
+  else
+    return _setattro(self,oname, v, PyExtensionClassCAPI->setattro);
+
+  if (m) 
+    {
+      Py_DECREF(m);
+      return 0;
+    }
+  
+  return -1;
 }
 
 static PyExtensionClass Pertype = {
@@ -665,9 +605,39 @@ static PyExtensionClass Pertype = {
 	(getattrofunc)Per_getattro,	/*tp_getattr with object key*/
 	(setattrofunc)Per_setattro,	/*tp_setattr with object key*/
 	/* Space for future expansion */
-	0L,0L,"",
+	0L,0L,
+        "Base class for objects that are stored in their own database records",
 	METHOD_CHAIN(Per_methods),
-	EXTENSIONCLASS_BASICNEW_FLAG | PERSISTENT_TYPE_FLAG,
+	PERSISTENCE_FLAGS,
+};
+
+static PyExtensionClass Overridable = {
+	PyObject_HEAD_INIT(NULL)
+	0,				/*ob_size*/
+	"Overridable",			/*tp_name*/
+	sizeof(cPersistentObject),	/*tp_basicsize*/
+	0,				/*tp_itemsize*/
+	/* methods */
+	(destructor)Per_dealloc,	/*tp_dealloc*/
+	(printfunc)0,			/*tp_print*/
+	(getattrfunc)0,			/*tp_getattr*/
+	(setattrfunc)0,	       		/*tp_setattr*/
+	(cmpfunc)0,			/*tp_compare*/
+	(reprfunc)0,			/*tp_repr*/
+	0,				/*tp_as_number*/
+	0,				/*tp_as_sequence*/
+	0,				/*tp_as_mapping*/
+	(hashfunc)0,			/*tp_hash*/
+	(ternaryfunc)0,			/*tp_call*/
+	(reprfunc)0,			/*tp_str*/
+	(getattrofunc)Per_getattro,	/*tp_getattr with object key*/
+	(setattrofunc)Per_setattro,	/*tp_setattr with object key*/
+	/* Space for future expansion */
+	0L,0L,
+        "Hacked base class used in Zope's App.Uninstalled.BrokenClass\n\n"
+        "Not sure if this is still needed",
+	METHOD_CHAIN(Per_methods),
+	EXTENSIONCLASS_BASICNEW_FLAG | PERSISTENT_TYPE_FLAG
 };
 
 /* End of code for Persistent objects */
@@ -716,33 +686,39 @@ truecPersistenceCAPI = {
 };
 
 void
-initcPersistence()
+initcPersistence(void)
 {
-  PyObject *m, *d;
-  char *rev="$Revision: 1.35 $";
+  PyObject *m, *d, *s;
+
+  s = PyString_FromString("TimeStamp");
+  if (s == NULL)
+      return;
+  m = PyImport_Import(s);
+  if (m == NULL) {
+      Py_DECREF(s);
+      return;
+  }
+  TimeStamp = PyObject_GetAttr(m, s);
+  Py_DECREF(m);
+  Py_DECREF(s);
+
+  if (init_strings() < 0) return;
+
+  m = Py_InitModule4("cPersistence", cP_methods, cPersistence_doc_string,
+		     (PyObject*)NULL, PYTHON_API_VERSION);
 
-  TimeStamp=PyString_FromString("TimeStamp");
-  if (! TimeStamp) return;
-  ASSIGN(TimeStamp, PyImport_Import(TimeStamp));
-  if (! TimeStamp) return;
-  ASSIGN(TimeStamp, PyObject_GetAttrString(TimeStamp, "TimeStamp"));
-  if (! TimeStamp) return;
   
-  m = Py_InitModule4("cPersistence", cP_methods,
-		     "",
-		     (PyObject*)NULL,PYTHON_API_VERSION);
-
-  init_strings();
-
   d = PyModule_GetDict(m);
-  PyDict_SetItemString(d,"__version__",
-		       PyString_FromStringAndSize(rev+11,strlen(rev+11)-2));
-  PyExtensionClass_Export(d,"Persistent",Pertype);
+
+  PyExtensionClass_Export(d, "Persistent",  Pertype);
+  PyExtensionClass_Export(d, "Overridable", Overridable);
 
   cPersistenceCAPI=&truecPersistenceCAPI;
-  PyDict_SetItemString(d, "CAPI",
-		       PyCObject_FromVoidPtr(cPersistenceCAPI,NULL));
+  s = PyCObject_FromVoidPtr(cPersistenceCAPI,NULL);
+  PyDict_SetItemString(d, "CAPI", s);
+  Py_XDECREF(s);
 
+  /* Check for errors */
   if (PyErr_Occurred())
-    Py_FatalError("can't initialize module cDocumentTemplate");
+    Py_FatalError("can't initialize module cPersistence");
 }
