@@ -4,87 +4,15 @@ exec python $0 ${1+"$@"}
 """
 #"
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 """Bobo call interface
@@ -103,9 +31,9 @@ that allows one to simply make a single web request.
 The module also provides a command-line interface for calling objects.
 
 """
-__version__='$Revision: 1.33 $'[11:-2]
+__version__='$Revision: 1.44 $'[11:-2]
 
-import sys, regex, socket, mimetools
+import sys, re, socket, mimetools
 from httplib import HTTP
 from os import getpid
 from time import time
@@ -113,7 +41,7 @@ from random import random
 from base64 import encodestring
 from urllib import urlopen, quote
 from types import FileType, ListType, DictType, TupleType
-from string import strip, split, atoi, join, rfind, translate, maketrans, replace
+from string import translate, maketrans
 from urlparse import urlparse
 
 class Function:
@@ -130,8 +58,8 @@ class Function:
         self.url=url
         self.headers=headers
         if not headers.has_key('Host') and not headers.has_key('host'):
-            headers['Host']=split(urlparse(url)[1],':')[0]
-        self.func_name=url[rfind(url,'/')+1:]
+            headers['Host']=urlparse(url)[1]
+        self.func_name=url[url.rfind('/')+1:]
         self.__dict__['__name__']=self.func_name
         self.func_defaults=()
         
@@ -142,9 +70,10 @@ class Function:
         if password is not None: self.password=password
         if timeout is not None: self.timeout=timeout
 
-        if urlregex.match(url) >= 0:
-            host,port,rurl=urlregex.group(1,2,3)
-            if port: port=atoi(port[1:])
+        mo = urlregex.match(url)
+        if mo is not None:
+            host,port,rurl=mo.group(1,2,3)
+            if port: port=int(port[1:])
             else: port=80
             self.host=host
             self.port=port
@@ -188,7 +117,7 @@ class Function:
 
         url=self.rurl
         if query:
-            query=join(query,'&')
+            query='&'.join(query)
             method=method or 'POST'
             if method == 'PUT':
                 headers['Content-Length']=str(len(query))
@@ -204,8 +133,9 @@ class Function:
             not headers.has_key('Authorization')):
             headers['Authorization']=(
                 "Basic %s" %
-                replace(encodestring('%s:%s' % (self.username,self.password)),
-				     '\012',''))
+                encodestring('%s:%s' % (self.username,self.password)).replace(
+				     '\012','')
+                )
 	    
         try:
             h=HTTP()
@@ -219,9 +149,9 @@ class Function:
             response     =h.getfile().read()
         except:
             raise NotAvailable, RemoteException(
-                NotAvailable,sys.exc_value,self.url,query)
-
-        if ec==200: return (headers,response)
+                NotAvailable,sys.exc_info()[1],self.url,query)
+        if (ec - (ec % 100)) == 200:
+            return (headers,response)
         self.handleError(query, ec, em, headers, response)
 
 
@@ -267,34 +197,34 @@ class Function:
         for n,v in self.headers.items():
             rq.append('%s: %s' % (n,v))
         if self.username and self.password:
-            c=replace(encodestring('%s:%s' % (self.username,self.password)),'\012','')
+            c=encodestring('%s:%s' % (self.username,self.password)).replace('\012','')
             rq.append('Authorization: Basic %s' % c)
         rq.append(MultiPart(d).render())
-        rq=join(rq,'\r\n')   
+        rq='\r\n'.join(rq)   
 
         try:
             sock=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
-            sock.connect(self.host,self.port)
+            sock.connect((self.host,self.port))
             sock.send(rq)
             reply=sock.makefile('rb')
             sock=None
             line=reply.readline()
 
             try:
-                [ver, ec, em] = split(line, None, 2)
+                [ver, ec, em] = line.split(None, 2)
             except ValueError:
                 raise 'BadReply','Bad reply from server: '+line
             if ver[:5] != 'HTTP/':
                 raise 'BadReply','Bad reply from server: '+line
 
-            ec=atoi(ec)
-            em=strip(em)
+            ec=int(ec)
+            em=em.strip()
             headers=mimetools.Message(reply,0)
             response=reply.read()
         finally:
           if 0:
             raise NotAvailable, (
-                RemoteException(NotAvailable,sys.exc_value,
+                RemoteException(NotAvailable,sys.exc_info()[1],
                                 self.url,'<MultiPart Form>'))
                 
         if ec==200: return (headers,response)
@@ -315,7 +245,7 @@ class Object:
         self.url=url
         self.headers=headers
         if not headers.has_key('Host') and not headers.has_key('host'):
-            headers['Host']=split(urlparse(url)[1],':')[0]
+            headers['Host']=urlparse(url)[1]
         if method is not None: self.method=method
         if username is not None: self.username=username
         if password is not None: self.password=password
@@ -346,23 +276,17 @@ def call(url,username=None, password=None, **kw):
 ##############################################################################
 # Implementation details below here
 
-urlregex=regex.compile('http://\([^:/]+\)\(:[0-9]+\)?\(/.+\)?', regex.casefold)
+urlregex=re.compile(r'http://([^:/]+)(:[0-9]+)?(/.+)?', re.I)
 
 dashtrans=maketrans('_','-')
 
 def marshal_float(n,f): return '%s:float=%s' % (n,f)
 def marshal_int(n,f):   return '%s:int=%s' % (n,f)
-def marshal_long(n,f):  return ('%s:long=%s' % (n,f))[:-1]
-
-sample_regex=regex.compile('')
-def marshal_regex(n,r):
-    if r.translate is sample_regex.translate:
-        t='Regex'
-    elif r.translate is regex.casefold:
-        t='regex'
-    else:
-        raise ValueError, 'regular expression used unsupported translation'
-    return "%s:%s=%s" % (n,t,quote(r.givenpat))
+def marshal_long(n,f):
+    value = '%s:long=%s' % (n, f)
+    if value[-1] == 'L':
+        value = value[:-1]
+    return value
 
 def marshal_list(n,l,tname='list', lt=type([]), tt=type(())):
     r=[]
@@ -372,7 +296,7 @@ def marshal_list(n,l,tname='list', lt=type([]), tt=type(())):
             raise TypeError, 'Invalid recursion in data to be marshaled.'
         r.append(marshal_whatever("%s:%s" % (n,tname) ,v))
     
-    return join(r,'&')
+    return '&'.join(r)
 
 def marshal_tuple(n,l):
     return marshal_list(n,l,'tuple')
@@ -381,7 +305,6 @@ type2marshal={
     type(1.0):                  marshal_float,
     type(1):                    marshal_int,
     type(1L):                   marshal_long,
-    type(regex.compile('')):    marshal_regex,
     type([]):                   marshal_list,
     type(()):                   marshal_tuple,
     }
@@ -395,7 +318,7 @@ def querify(items):
     query=[]
     for k,v in items: query.append(marshal_whatever(k,v))
 
-    return query and join(query,'&') or ''
+    return query and '&'.join(query) or ''
 
 NotFound     ='bci.NotFound'
 InternalError='bci.InternalError'
@@ -478,11 +401,13 @@ class MultiPart:
 
         elif dt==FileType or hasattr(val,'read'):
             if hasattr(val,'name'):
-                fn=replace(val.name, '\\', '/')
-                fn=fn[(rfind(fn,'/')+1):]
-                ex=fn[(rfind(fn,'.')+1):]
-                if self._extmap.has_key(ex): ct=self._extmap[ex]
-                else: ct=self._extmap['']
+                fn=val.name.replace( '\\', '/')
+                fn=fn[(fn.rfind('/')+1):]
+                ex=(fn[(fn.rfind('.')+1):]).lower()
+                if self._extmap.has_key(ex):
+                    ct=self._extmap[ex]
+                else:
+                    ct=self._extmap['']
             else:
                 fn=''
                 ct=self._extmap[None]
@@ -530,12 +455,12 @@ class MultiPart:
             b=self._boundary
             for d in self._data: p.append(d.render())
             t.append('--%s\n' % b)
-            t.append(join(p,'\n--%s\n' % b))
+            t.append(('\n--%s\n' % b).join(p))
             t.append('\n--%s--\n' % b)
-            t=join(t,'')
+            t=''.join(t)
             s.append('Content-Length: %s\r\n\r\n' % len(t))
             s.append(t)
-            return join(s,'')
+            return ''.join(s)
 
         else:
             for n,v in h.items():
@@ -551,11 +476,11 @@ class MultiPart:
                 b=self._boundary
                 for d in self._data: p.append(d.render())
                 s.append('--%s\n' % b)
-                s.append(join(p,'\n--%s\n' % b))
+                s.append(('\n--%s\n' % b).join(p))
                 s.append('\n--%s--\n' % b)
-                return join(s,'')
+                return ''.join(s)
             else:
-                return join(s+self._data,'')
+                return ''.join(s+self._data)
 
 
     _extmap={'':     'text/plain',
@@ -602,7 +527,6 @@ The headers of the response are written to standard error.
 
 def main():
     import getopt
-    from string import split
 
     user=None
 
@@ -611,15 +535,15 @@ def main():
         url=args[0]
         u =filter(lambda o: o[0]=='-u', optlist)
         if u:
-            [user, pw] = split(u[0][1],':')
+            [user, pw] = u[0][1].split(':')
 
         kw={}
         for arg in args[1:]:
-            [name,v]=split(arg,'=')
+            [name,v]=arg.split('=')
             if name[-5:]==':file':
                 name=name[:-5]
                 if v=='-': v=sys.stdin
-                else: v=open(v)
+                else: v=open(v, 'rb')
             kw[name]=v
 
     except:
@@ -630,7 +554,7 @@ def main():
     f=Function(url)
     if user: f.username, f.password = user, pw
     headers, body = apply(f,(),kw)
-    sys.stderr.write(join(map(lambda h: "%s: %s\n" % h, headers.items()),"")
+    sys.stderr.write(''.join(map(lambda h: "%s: %s\n" % h, headers.items()))
                      +"\n\n")
     print body
 
