@@ -1,100 +1,31 @@
 ##############################################################################
+#
+# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
 # 
-# Zope Public License (ZPL) Version 1.0
-# -------------------------------------
-# 
-# Copyright (c) Digital Creations.  All rights reserved.
-# 
-# This license has been certified as Open Source(tm).
-# 
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are
-# met:
-# 
-# 1. Redistributions in source code must retain the above copyright
-#    notice, this list of conditions, and the following disclaimer.
-# 
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions, and the following disclaimer in
-#    the documentation and/or other materials provided with the
-#    distribution.
-# 
-# 3. Digital Creations requests that attribution be given to Zope
-#    in any manner possible. Zope includes a "Powered by Zope"
-#    button that is installed by default. While it is not a license
-#    violation to remove this button, it is requested that the
-#    attribution remain. A significant investment has been put
-#    into Zope, and this effort will continue if the Zope community
-#    continues to grow. This is one way to assure that growth.
-# 
-# 4. All advertising materials and documentation mentioning
-#    features derived from or use of this software must display
-#    the following acknowledgement:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    In the event that the product being advertised includes an
-#    intact Zope distribution (with copyright and license included)
-#    then this clause is waived.
-# 
-# 5. Names associated with Zope or Digital Creations must not be used to
-#    endorse or promote products derived from this software without
-#    prior written permission from Digital Creations.
-# 
-# 6. Modified redistributions of any form whatsoever must retain
-#    the following acknowledgment:
-# 
-#      "This product includes software developed by Digital Creations
-#      for use in the Z Object Publishing Environment
-#      (http://www.zope.org/)."
-# 
-#    Intact (re-)distributions of any official Zope release do not
-#    require an external acknowledgement.
-# 
-# 7. Modifications are encouraged but must be packaged separately as
-#    patches to official Zope releases.  Distributions that do not
-#    clearly separate the patches from the original work must be clearly
-#    labeled as unofficial distributions.  Modifications which do not
-#    carry the name Zope may be packaged in any form, as long as they
-#    conform to all of the clauses above.
-# 
-# 
-# Disclaimer
-# 
-#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
-#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
-#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
-#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
-#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-#   SUCH DAMAGE.
-# 
-# 
-# This software consists of contributions made by Digital Creations and
-# many individuals on behalf of Digital Creations.  Specific
-# attributions are listed in the accompanying credits file.
+# This software is subject to the provisions of the Zope Public License,
+# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
+# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
+# FOR A PARTICULAR PURPOSE
 # 
 ##############################################################################
 
-__version__='$Revision: 1.26.4.1 $'[11:-2]
+__version__='$Revision: 1.61.6.2 $'[11:-2]
 
-import regex, sys, os, string
-from string import lower, atoi, rfind, split, strip, join, upper, find
+import re, sys, os,  urllib, time, whrandom, cgi
 from BaseRequest import BaseRequest
 from HTTPResponse import HTTPResponse
-from cgi import FieldStorage
-from urllib import quote, unquote
+from cgi import FieldStorage, escape
+from urllib import quote, unquote, splittype, splitport
 from Converters import get_converter
 from maybe_lock import allocate_lock
 xmlrpc=None # Placeholder for module that we'll import if we have to.
 
+#cgi hotfix:
+if not hasattr(cgi, 'valid_boundary'):
+    import cgi_hotfix
+
 isCGI_NAME = {
         'SERVER_SOFTWARE' : 1, 
         'SERVER_NAME' : 1, 
@@ -120,6 +51,8 @@ hide_key={'HTTP_AUTHORIZATION':1,
           'HTTP_CGI_AUTHORIZATION': 1,
           }.has_key
 
+default_port={'http': '80', 'https': '443'}
+
 _marker=[]
 class HTTPRequest(BaseRequest):
     """\
@@ -133,7 +66,7 @@ class HTTPRequest(BaseRequest):
 
     The request object is a mapping object that represents a
     collection of variable to value mappings.  In addition, variables
-    are divided into four categories:
+    are divided into five categories:
 
       - Environment variables
 
@@ -153,6 +86,12 @@ class HTTPRequest(BaseRequest):
 
         These are the cookie data, if present.
 
+      - Lazy Data
+
+        These are callables which are deferred until explicitly
+        referenced, at which point they are resolved and stored as
+        application data.
+
       - Other
 
         Data that may be set by an application object.
@@ -170,9 +109,13 @@ class HTTPRequest(BaseRequest):
     _hacked_path=None
     args=()
     _file=None
+    _urls = ()
 
     retry_max_count=3
-    def supports_retry(self): return self.retry_count < self.retry_max_count
+    def supports_retry(self):
+        if self.retry_count < self.retry_max_count:
+            time.sleep(whrandom.uniform(0, 2**(self.retry_count)))
+            return 1
 
     def retry(self):
         self.retry_count=self.retry_count+1
@@ -181,8 +124,103 @@ class HTTPRequest(BaseRequest):
                          environ=self._orig_env,
                          response=self.response.retry()
                          )
+        r.retry_count=self.retry_count
         return r
 
+    def close(self):
+        # we want to clear the lazy dict here because BaseRequests don't have
+        # one.  Without this, there's the possibility of memory leaking
+        # after every request.
+        self._lazies = {}
+        BaseRequest.close(self)
+
+    def setServerURL(self, protocol=None, hostname=None, port=None):
+        """ Set the parts of generated URLs. """
+        other = self.other
+        server_url = other.get('SERVER_URL', '')
+        if protocol is None and hostname is None and port is None:
+            return server_url
+        oldprotocol, oldhost = splittype(server_url)
+        oldhostname, oldport = splitport(oldhost[2:])
+        if protocol is None: protocol = oldprotocol
+        if hostname is None: hostname = oldhostname
+        if port is None: port = oldport
+        
+        if (port is None or default_port[protocol] == port):
+            host = hostname
+        else:
+            host = hostname + ':' + port
+        server_url = other['SERVER_URL'] = '%s://%s' % (protocol, host)
+        self._resetURLS()
+        return server_url
+
+    def setVirtualRoot(self, path, hard=0):
+        """ Treat the current publishing object as a VirtualRoot """
+        other = self.other
+        if type(path) is type(''):
+            path = path.split( '/')
+        self._script[:] = map(quote, filter(None, path))
+        del self._steps[:]
+        parents = other['PARENTS']
+        if hard:
+            del parents[:-1]
+        other['VirtualRootPhysicalPath'] = parents[-1].getPhysicalPath()
+        self._resetURLS()
+
+    def physicalPathToVirtualPath(self, path):
+        """ Remove the path to the VirtualRoot from a physical path """
+        if type(path) is type(''):
+            path = path.split( '/')
+        rpp = self.other.get('VirtualRootPhysicalPath', ('',))
+        i = 0
+        for name in rpp[:len(path)]:
+            if path[i] == name:
+                i = i + 1
+            else:
+                break
+        return path[i:]
+
+    def physicalPathToURL(self, path, relative=0):
+        """ Convert a physical path into a URL in the current context """
+        path = self._script + map(quote, self.physicalPathToVirtualPath(path))
+        if relative:
+            path.insert(0, '')
+        else:
+            path.insert(0, self['SERVER_URL'])
+        return '/'.join(path)
+
+    def physicalPathFromURL(self, URL):
+        """ Convert a URL into a physical path in the current context.
+            If the URL makes no sense in light of the current virtual
+            hosting context, a ValueError is raised."""
+        other = self.other
+        bad_server_url = 0
+        path = filter(None, URL.split( '/'))
+
+        if URL.find( '://') >= 0:
+            path = path[2:]
+
+        # Check the path against BASEPATH1
+        vhbase = self._script
+        vhbl = len(vhbase)
+        bad_basepath = 0
+        if path[:vhbl] == vhbase:
+            path = path[vhbl:]
+        else:
+            raise ValueError, (
+                'Url does not match virtual hosting context'
+                )
+        vrpp = other.get('VirtualRootPhysicalPath', ('',))
+        return list(vrpp) + map(unquote, path)
+
+    def _resetURLS(self):
+        other = self.other
+        other['URL'] = '/'.join([other['SERVER_URL']] + self._script +
+                            self._steps)
+        for x in self._urls:
+            del self.other[x]
+        self._urls = ()
+
     def __init__(self, stdin, environ, response, clean=0):
         self._orig_env=environ
         # Avoid the overhead of scrubbing the environment in the
@@ -195,7 +233,7 @@ class HTTPRequest(BaseRequest):
             self._auth=environ['HTTP_AUTHORIZATION']
             response._auth=1
             del environ['HTTP_AUTHORIZATION']
-        
+
         self.stdin=stdin
         self.environ=environ
         have_env=environ.has_key
@@ -204,36 +242,55 @@ class HTTPRequest(BaseRequest):
         other=self.other={'RESPONSE': response}
         self.form={}
         self.steps=[]
+        self._steps=[]
+        self._lazies={}
 
         ################################################################
         # Get base info first. This isn't likely to cause
         # errors and might be useful to error handlers.
-        b=script=strip(get_env('SCRIPT_NAME',''))
+        b=script=get_env('SCRIPT_NAME','').strip()
+
+        # _script and the other _names are meant for URL construction
+        self._script = map(quote, filter(None, script.split( '/')))
+        
         while b and b[-1]=='/': b=b[:-1]
-        p = rfind(b,'/')
+        p = b.rfind('/')
         if p >= 0: b=b[:p+1]
         else: b=''
         while b and b[0]=='/': b=b[1:]
 
         server_url=get_env('SERVER_URL',None)
         if server_url is not None:
-             server_url=strip(server_url)
+             other['SERVER_URL'] = server_url = server_url.strip()
         else:
              if have_env('HTTPS') and (
                  environ['HTTPS'] == "on" or environ['HTTPS'] == "ON"):
-                 server_url='https://'
+                 protocol = 'https'
              elif (have_env('SERVER_PORT_SECURE') and 
                    environ['SERVER_PORT_SECURE'] == "1"):
-                 server_url='https://'
-             else: server_url='http://'
+                 protocol = 'https'
+             else: protocol = 'http'
 
              if have_env('HTTP_HOST'):
-                 server_url=server_url+strip(environ['HTTP_HOST'])
+                 host = environ['HTTP_HOST'].strip()
+                 hostname, port = splitport(host)
+
+                 # NOTE: some (DAV) clients manage to forget the port. This
+                 # can be fixed with the commented code below - the problem
+                 # is that it causes problems for virtual hosting. I've left
+                 # the commented code here in case we care enough to come
+                 # back and do anything with it later.
+                 #
+                 # if port is None and environ.has_key('SERVER_PORT'):
+                 #     s_port=environ['SERVER_PORT']
+                 #     if s_port not in ('80', '443'):
+                 #         port=s_port
+
              else:
-                 server_url=server_url+strip(environ['SERVER_NAME'])
-                 server_port=environ['SERVER_PORT']
-                 if server_port!='80': server_url=server_url+':'+server_port
-             other['SERVER_URL']=server_url
+                 hostname = environ['SERVER_NAME'].strip()
+                 port = environ['SERVER_PORT']
+             self.setServerURL(protocol=protocol, hostname=hostname, port=port)
+             server_url = other['SERVER_URL']
              
         if server_url[-1:]=='/': server_url=server_url[:-1]
                         
@@ -270,8 +327,7 @@ class HTTPRequest(BaseRequest):
         hasattr=hasattr,
         getattr=getattr,
         setattr=setattr,
-        search_type=regex.compile(':[a-zA-Z][a-zA-Z0-9_]+$').search,
-        rfind=string.rfind,
+        search_type=re.compile('(:[a-zA-Z][a-zA-Z0-9_]+|\\.[xy])$').search,
         ):
         """Process request inputs
 
@@ -302,6 +358,7 @@ class HTTPRequest(BaseRequest):
                 response=xmlrpc.response(response)
                 other['RESPONSE']=self.response=response
                 other['REQUEST_METHOD']='' # We don't want index_html!
+                self.maybe_webdav_client = 0
             else:
                 self._file=fs.file
         else:
@@ -334,15 +391,20 @@ class HTTPRequest(BaseRequest):
                 # We'll search from the back to the front.
                 # We'll do the search in two steps.  First, we'll
                 # do a string search, and then we'll check it with
-                # a regex search.
+                # a re search.
                 
-                l=rfind(key,':')
+
+                l=key.rfind(':')
                 if l >= 0:
-                    l=search_type(key,l)
+                    mo = search_type(key,l)
+                    if mo: l=mo.start(0)
+                    else:  l=-1
+
                     while l >= 0:
                         type_name=key[l+1:]
                         key=key[:l]
                         c=get_converter(type_name, None)
+                        
                         if c is not None: 
                             converter=c
                             flags=flags|CONVERTED
@@ -353,10 +415,11 @@ class HTTPRequest(BaseRequest):
                             seqf=tuple
                             tuple_items[key]=1
                             flags=flags|SEQUENCE
-                        elif type_name == 'method':
+                        elif (type_name == 'method' or type_name == 'action'):
                             if l: meth=key
                             else: meth=item
-                        elif type_name == 'default_method':
+                        elif (type_name == 'default_method' or type_name == \
+                              'default_action'):
                             if not meth:
                                 if l: meth=key
                                 else: meth=item
@@ -369,9 +432,13 @@ class HTTPRequest(BaseRequest):
                         elif type_name == 'ignore_empty':
                             if not item: flags=flags|EMPTY
     
-                        l=rfind(key,':')
+                        l=key.rfind(':')
                         if l < 0: break
-                        l=search_type(key,l)
+                        mo=search_type(key,l)
+                        if mo: l = mo.start(0)
+                        else:  l = -1
+
+
              
                 # Filter out special names from form:
                 if CGI_name(key) or key[:5]=='HTTP_': continue
@@ -383,8 +450,8 @@ class HTTPRequest(BaseRequest):
 
                     #Split the key and its attribute
                     if flags&REC:
-                        key=split(key,".")
-                        key, attr=join(key[:-1],"."), key[-1]
+                        key=key.split(".")
+                        key, attr=".".join(key[:-1]), key[-1]
                        
                     # defer conversion
                     if flags&CONVERTED:
@@ -420,7 +487,7 @@ class HTTPRequest(BaseRequest):
                            reclist.reverse()
                            if not hasattr(x,attr):
                                #If the attribute does not
-                               #exist, set it
+                               #exist, setit
                                if flags&SEQUENCE: item=[item]
                                reclist.remove(x)
                                setattr(x,attr,item)
@@ -563,13 +630,13 @@ class HTTPRequest(BaseRequest):
             if tuple_items:
                 for key in tuple_items.keys():
                    # Split the key and get the attr
-                   k=split(key, ".")
-                   k,attr=join(k[:-1], "."), k[-1]
+                   k=key.split( ".")
+                   k,attr='.'.join(k[:-1]), k[-1]
                    a = attr
                    # remove any type_names in the attr
                    while not a=='':
-                      a=split(a, ":")
-                      a,new=join(a[:-1], ":"), a[-1]
+                      a=a.split( ":")
+                      a,new=':'.join(a[:-1]), a[-1]
                    attr = new
                    if form.has_key(k):
                       # If the form has the split key get its value
@@ -619,7 +686,7 @@ class HTTPRequest(BaseRequest):
         # namespace (e.g. the host, port or script name dont
         # match that of the current request), a ValueError will
         # be raised.
-        if find(url, self.script) != 0:
+        if url.find(self.script) != 0:
             raise ValueError, 'Different namespace.'
         path=url[len(self.script):]
         while path and path[0]=='/':  path=path[1:]
@@ -635,10 +702,10 @@ class HTTPRequest(BaseRequest):
         # (which is usually the default when the exception
         # method is called on the response).
         try: object=req.traverse(path)
-        except: rsp.exception(abort=0)
+        except: rsp.exception()
         if object is None:
             req.close()
-            raise rsp.errmsg, sys.exc_value
+            raise rsp.errmsg, sys.exc_info()[1]
 
         # The traversal machinery may return a "default object"
         # like an index_html document. This is not appropriate
@@ -677,7 +744,7 @@ class HTTPRequest(BaseRequest):
         should all return the Content-Type header, if available.
         """
         environ=self.environ
-        name=upper(join(split(name,"-"),"_"))
+        name=('_'.join(name.split("-"))).upper()
         val=environ.get(name, None)
         if val is not None:
             return val
@@ -685,12 +752,10 @@ class HTTPRequest(BaseRequest):
             name='HTTP_%s' % name
         return environ.get(name, default)
 
-
-    def __getitem__(self,key,
-                    default=_marker, # Any special internal marker will do
-                    URLmatch=regex.compile('URL[0-9]+$').match,
-                    BASEmatch=regex.compile('BASE[0-9]+$').match,
-                    ):
+    def get(self, key, default=None,
+            URLmatch=re.compile('URL(PATH)?([0-9]+)$').match,
+            BASEmatch=re.compile('BASE(PATH)?([0-9]+)$').match,
+            ):
         """Get a variable value
 
         Return a value for the required variable name.
@@ -704,16 +769,21 @@ class HTTPRequest(BaseRequest):
             if key=='REQUEST': return self
             return other[key]
 
-        if key[:1]=='U' and URLmatch(key) >= 0:
-            n=atoi(key[3:])
-            URL=other['URL']
-            for i in range(0,n):
-                l=rfind(URL,'/')
-                if l >= 0: URL=URL[:l]
-                else: raise KeyError, key
-                if len(URL) < len(self.base) and n > 1: raise KeyError, key
-            other[key]=URL
-            return URL
+        if key[:1]=='U':
+            match = URLmatch(key)
+            if match is not None:
+                pathonly, n = match.groups()
+                path = self._script + self._steps
+                n = len(path) - int(n)
+                if n < 0:
+                    raise KeyError, key
+                if pathonly:
+                    path = [''] + path[:n]
+                else:
+                    path = [other['SERVER_URL']] + path[:n]
+                other[key] = URL = '/'.join(path)
+                self._urls = self._urls + (key,)
+                return URL
 
         if isCGI_NAME(key) or key[:5] == 'HTTP_':
             environ=self.environ
@@ -724,22 +794,26 @@ class HTTPRequest(BaseRequest):
         if key=='REQUEST': return self
 
         if key[:1]=='B':
-            if BASEmatch(key) >= 0:
-                n=atoi(key[4:])
+            match = BASEmatch(key)
+            if match is not None:
+                pathonly, n = match.groups()
+                path = self._steps
+                n = int(n)
                 if n:
-                    n=n-1
-                    
-                    if len(self.steps) < n:
+                    n = n - 1
+                    if len(path) < n:
                         raise KeyError, key
 
-                    v=self.script
-                    while v[-1:]=='/': v=v[:-1]
-                    v=join([v]+self.steps[:n],'/')
+                    v = self._script + path[:n]
                 else:
-                    v=self.base
-                    while v[-1:]=='/': v=v[:-1]
-                other[key]=v
-                return v
+                    v = self._script[:-1]
+                if pathonly:
+                    v.insert(0, '')
+                else:
+                    v.insert(0, other['SERVER_URL'])
+                other[key] = URL = '/'.join(v)
+                self._urls = self._urls + (key,)
+                return URL
 
             if key=='BODY' and self._file is not None:
                 p=self._file.tell()
@@ -754,15 +828,33 @@ class HTTPRequest(BaseRequest):
                 self.other[key]=v
                 return v
 
-        v=self.common.get(key, default)
+        v=self.common.get(key, _marker)
         if v is not _marker: return v
 
-        raise KeyError, key
+        if self._lazies:
+            v = self._lazies.get(key, _marker)
+            if v is not _marker:
+                if callable(v): v = v()
+                self[key] = v                   # Promote lazy value
+                del self._lazies[key]
+                return v
 
-    __getattr__=__getitem__
+        return default
 
-    def get(self, key, default=None):
-        return self.__getitem__(key, default)
+    def __getitem__(self, key, default=_marker):
+        v = self.get(key, default)
+        if v is _marker:
+            raise KeyError, key
+        return v
+
+    def __getattr__(self, key, default=_marker):
+        v = self.get(key, default)
+        if v is _marker:
+            raise AttributeError, key
+        return v
+
+    def set_lazy(self, key, callable):
+        self._lazies[key] = callable
 
     def has_key(self, key):
         try: self[key]
@@ -772,6 +864,7 @@ class HTTPRequest(BaseRequest):
     def keys(self):
         keys = {}
         keys.update(self.common)
+        keys.update(self._lazies)
 
         for key in self.environ.keys():
             if (isCGI_NAME(key) or key[:5] == 'HTTP_') and \
@@ -801,40 +894,74 @@ class HTTPRequest(BaseRequest):
         result="<h3>form</h3><table>"
         row='<tr valign="top" align="left"><th>%s</th><td>%s</td></tr>'
         for k,v in self.form.items():
-            result=result + row % (k,v)
+            result=result + row % (escape(k), escape(repr(v)))
         result=result+"</table><h3>cookies</h3><table>"
         for k,v in self.cookies.items():
-            result=result + row % (k,v)
+            result=result + row % (escape(k), escape(repr(v)))
+        result=result+"</table><h3>lazy items</h3><table>"
+        for k,v in self._lazies.items():
+            result=result + row % (escape(k), escape(repr(v)))
         result=result+"</table><h3>other</h3><table>"
         for k,v in self.other.items():
             if k in ('PARENTS','RESPONSE'): continue
-            result=result + row % (k,v)
+            result=result + row % (escape(k), escape(repr(v)))
     
         for n in "0123456789":
             key = "URL%s"%n
-            try: result=result + row % (key,self[key]) 
+            try: result=result + row % (key, escape(self[key])) 
             except KeyError: pass
         for n in "0123456789":
             key = "BASE%s"%n
-            try: result=result + row % (key,self[key]) 
+            try: result=result + row % (key, escape(self[key])) 
             except KeyError: pass
 
         result=result+"</table><h3>environ</h3><table>"
         for k,v in self.environ.items():
             if not hide_key(k):
-                result=result + row % (k,v)
+                result=result + row % (escape(k), escape(repr(v)))
         return result+"</table>"
 
     __repr__=__str__
 
+    def text(self):
+        result="FORM\n\n"
+        row='%-20s %s\n'
+        for k,v in self.form.items():
+            result=result + row % (k, repr(v))
+        result=result+"\nCOOKIES\n\n"
+        for k,v in self.cookies.items():
+            result=result + row % (k, repr(v))
+        result=result+"\nLAZY ITEMS\n\n"
+        for k,v in self._lazies.items():
+            result=result + row % (k, repr(v))
+        result=result+"\nOTHER\n\n"
+        for k,v in self.other.items():
+            if k in ('PARENTS','RESPONSE'): continue
+            result=result + row % (k, repr(v))
+    
+        for n in "0123456789":
+            key = "URL%s"%n
+            try: result=result + row % (key, self[key]) 
+            except KeyError: pass
+        for n in "0123456789":
+            key = "BASE%s"%n
+            try: result=result + row % (key, self[key]) 
+            except KeyError: pass
+
+        result=result+"\nENVIRON\n\n"
+        for k,v in self.environ.items():
+            if not hide_key(k):
+                result=result + row % (k, v)
+        return result
+
     def _authUserPW(self):
         global base64
         auth=self._auth
         if auth:
-            if lower(auth[:6]) == 'basic ':
+            if auth[:6].lower() == 'basic ':
                 if base64 is None: import base64
-                [name,password] = split(
-                    base64.decodestring(split(auth)[-1]), ':')
+                [name,password] = \
+                    base64.decodestring(auth.split()[-1]).split(':')
                 return name, password
 
 
@@ -858,7 +985,6 @@ def sane_environment(env):
     return dict
 
 
-
 def str_field(v):
     if type(v) is ListType:
         return map(str_field,v)
@@ -884,6 +1010,10 @@ class FileUpload:
     containing the name of the uploaded file.
     '''
 
+    # Allow access to attributes such as headers and filename so
+    # that ZClass authors can use DTML to work with FileUploads.
+    __allow_access_to_unprotected_subobjects__=1
+
     def __init__(self, aFieldStorage):
 
         file=aFieldStorage.file
@@ -898,21 +1028,21 @@ class FileUpload:
 
         self.headers=aFieldStorage.headers
         self.filename=aFieldStorage.filename
-    
+
+        # Add an assertion to the rfc822.Message object that implements
+        # self.headers so that managed code can access them.
+        try:    self.headers.__allow_access_to_unprotected_subobjects__ = 1
+        except: pass
+
 
 parse_cookie_lock=allocate_lock()
 def parse_cookie(text,
                  result=None,
-                 qparmre=regex.compile(
-                     '\([\0- ]*'
-                     '\([^\0- ;,=\"]+\)="\([^"]*\)\"'
-                     '\([\0- ]*[;,]\)?[\0- ]*\)'
-                     ),
-                 parmre=regex.compile(
-                     '\([\0- ]*'
-                     '\([^\0- ;,=\"]+\)=\([^\0- ;,\"]*\)'
-                     '\([\0- ]*[;,]\)?[\0- ]*\)'
-                     ),
+                 qparmre=re.compile(
+                    '([\x00- ]*([^\x00- ;,="]+)="([^"]*)"([\x00- ]*[;,])?[\x00- ]*)'),
+                 parmre=re.compile(
+                    '([\x00- ]*([^\x00- ;,="]+)=([^\x00- ;,"]*)([\x00- ]*[;,])?[\x00- ]*)'),
+
                  acquire=parse_cookie_lock.acquire,
                  release=parse_cookie_lock.release,
                  ):
@@ -922,22 +1052,29 @@ def parse_cookie(text,
 
     acquire()
     try:
-        if qparmre.match(text) >= 0:
+
+        mo_q = qparmre.match(text)
+
+        if mo_q:
             # Match quoted correct cookies
-            name=qparmre.group(2)
-            value=qparmre.group(3)
-            l=len(qparmre.group(1))
-        elif parmre.match(text) >= 0:
-            # Match evil MSIE cookies ;)
-            name=parmre.group(2)
-            value=parmre.group(3)
-            l=len(parmre.group(1))
+
+            l     = len(mo_q.group(1))
+            name  = mo_q.group(2)
+            value = mo_q.group(3)
+
         else:
-            # this may be an invalid cookie.
-            # We'll simply bail without raising an error
-            # if the cookie is invalid.
-            return result
-            
+            # Match evil MSIE cookies ;)
+
+            mo_p = parmre.match(text)
+
+            if mo_p:
+                l     = len(mo_p.group(1))
+                name  = mo_p.group(2)
+                value = mo_p.group(3)
+
+            else: 
+                return result 
+
     finally: release()
 
     if not already_have(name): result[name]=value
@@ -946,12 +1083,29 @@ def parse_cookie(text,
 
 # add class
 class record:
+
+    # Allow access to record methods and values from DTML
+    __allow_access_to_unprotected_subobjects__=1
+
+    def __getattr__(self, key, default=None):
+        if key in ('get', 'keys', 'items', 'values', 'copy', 'has_key'):
+            return getattr(self.__dict__, key)
+        raise AttributeError, key
+
+    def __getitem__(self, key):
+        return self.__dict__[key]
+            
     def __str__(self):
         L1 = self.__dict__.items()
         L1.sort()
-        return join(map(lambda item: "%s: %s" %item, L1), ", ") 
+        return ", ".join(map(lambda item: "%s: %s" % item, L1)) 
 
-    __repr__ = __str__
+    def __repr__(self):
+        #return repr( self.__dict__ )
+        L1 = self.__dict__.items()
+        L1.sort()
+        return '{%s}' % ', '.join(
+            map(lambda item: "'%s': %s" % (item[0], repr(item[1])), L1))
 
 # Flags
 SEQUENCE=1
