@@ -217,7 +217,7 @@ if sgmlop:
             try:
                 self.parser.close()
             finally:
-                self.parser = None # nuke circular reference
+                self.parser = self.feed = None # nuke circular reference
 
         def handle_entityref(self, entity):
             # <string> entity
@@ -609,7 +609,7 @@ class Server:
 
         # get the url
         type, uri = urllib.splittype(uri)
-        if type != "http":
+        if type not in ("http", "https"):
             raise IOError, "unsupported XML-RPC protocol"
         self.__host, self.__handler = urllib.splithost(uri)
         if not self.__handler:
